/************************************************/
/*						*/
/*	CLMAILC:				*/
/*	Program to send email from STDIN to a	*/
/*	remote SMTP server			*/
/*						*/
/************************************************/
#include	<getopt.h>
#include	<string.h>
#include	<stdlib.h>
#include	<syslog.h>
#include	<string.h>
#include	<utlgac.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unidns.h"
#include	"unidev.h"
#include	"unijrl.h"
#include	"comeml.h"
#include	"gesclm.h"
#include	"lvleml.h"

#define	MAILC	"clmailc"
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",appname,clm_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-F from] "
		      "[-H hostname[:port]] "
		      "[-v ] "
		      "[-h ] "
		      "email [email ...]\n",appname);
(void) fprintf(stderr,"\t\t-b base              : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug             : debug level\n");
(void) fprintf(stderr,"\t\t-v                   : verbose debug\n");
(void) fprintf(stderr,"\t\t-H hostname[:port]   : remote SMTP host and connection port\n");
(void) fprintf(stderr,"\t\t-F from              : email from\n");
(void) fprintf(stderr,"\t\t-h                   : help\n");
(void) exit(-1);
}
/*

*/
/************************************************/
/*						*/
/*	transmitting email to remote server	*/
/*						*/
/************************************************/
static int work(char *remote,char *from,int argc,char *argv[])

{
/*waiting 10 secondes for remote server		*/
#define	SECTIME	  10

int status;
int socker;
CTTYP *contact;
char *hostname;
char *port;
int phase;
int proceed;

status=0;
socker=0;
contact=(CTTYP *)0;
hostname=strdup(remote);
port=strdup(SMTPPORT);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting hostname	*/
      if (hostname!=(char *)0) { 	/*always*/ 
	char *ptr;

	if ((ptr=strchr(hostname,':'))!=(char *)0) {
	  *ptr='\000';
	  ptr++;
	  (void) free(port);
	  port=strdup(ptr);
	  }
	}
      break;
    case 1	:	/*openning remote	*/
      contact=eml_opencontact(pro_smtp,hostname,port,SECTIME,&socker);
      if (contact==(CTTYP *)0) {
	char *strloc;
	
	(void) asprintf(&strloc,"Unable to make SMTP contact with <%s> port '%s' (error=<%s>)",
                                 hostname,port,strerror(socker));
	(void) (void) jrl_journalling(false,EVTRT,"%s",strloc);
	(void) free(strloc);
	status=1;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*checking contact	*/
      (void) jrl_journalling(false,EVTRT,"%-*s%s",6,SESSION,contact->session);
      if (contact->established==false) {
	char *strloc;
	
	(void) asprintf(&strloc,"Unable to establish contact with <%s> port '%s' (got=<%s>)",
                                 hostname,port,contact->info);
	(void) (void) jrl_journalling(false,EVTRT,"%s",strloc);
	(void) free(strloc);
	status=2;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*sending from		*/
      (void) jrl_journalling(false,EVTRT,"Connected to <%s/[%s]> port '%s'",
                                          contact->peername,contact->peerip,
                                          contact->servport);
      (void) jrl_journalling(false,EVTRT,"MAIL FROM: <%s>",from);
      status=eml_sendmfrom(contact,from,0);
      (void) jrl_multijrl(EVTRT,"",contact->info);
      if (status!=CMDOK) {
	status=3;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*sending all RCPT	*/
      if (argc>0) {	/*always		*/
	int i;

        for (i=0;i<argc;i++) {
	  (void) jrl_journalling(false,EVTRT,"RCTP TO: <%s>",argv[i]);
	  status=eml_sendcmd(contact,CMDOK,"%s: <%s>",RCPT,argv[i]);
	  (void) jrl_multijrl(EVTRT,"",contact->info);
	  switch (status) {
	    case CMDOK	:	/*RCPT accepted	*/
	      status=0;
	      break;
	    default	:
	      status=4;
	      phase=999;	/*trouble	*/
	      i=argc;
	      break;
	    }
	  }
	}
      break;
    case 5	:	/*preparing data sending*/
      (void) jrl_journalling(false,EVTRT,"DATA");
      status=eml_sendcmd(contact,DATAOK,"%s",DATA);
      (void) jrl_multijrl(EVTRT,"",contact->info);
      if (status!=DATAOK) {
	status=5;
	phase=999;	/*trouble trouble	*/
        }
      break;
    case 6	:	/*sending Email		*/
      status=eml_contactstream(contact,stdin);
      (void) jrl_multijrl(EVTRT,"",contact->info);
      if (status!=CMDOK) {
	status=6;
	phase=999;	/*trouble trouble	*/
        }
      break;
    case 7	:	/*sending End of Email	*/
      status=eml_sendcmd(contact,CMDOK,"%s",DATAEND);
      (void) jrl_multijrl(EVTRT,"",contact->info);
      if (status!=CMDOK) {
	status=7;
	phase=999;	/*trouble trouble	*/
        }
      break;
    case 8	:	/*everything fine	*/
      status=0;
      break;
    default	:
      contact=eml_closecontact(contact);
      (void) jrl_journalling(true,EVTRT,"Session completed");
      proceed=false;	/*SAFE Guard		*/
      break;
    }
  phase++;
  }
(void) free(port);
(void) free(hostname);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Main Program				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
char c;
int status;
GACTYP *gat;
char *from;
char *remote;

clm_foreground=true;
(void) clm_trapsegv(true);
(void) openlog(MAILC,LOG_PID,0);
base=strdup(base);
status=0;
gat=(GACTYP *)0;
from=strdup("");
remote=strdup(LOCALHOST);
(void) openlog(appname,LOG_PID,0);
while ((c=getopt(argc,argv,"F:H:b:d:hv"))!=EOF) {
  switch(c) {
    case   'F'  :
      if ((optarg!=(char *)0)&&(strlen(optarg)>0)) {
        (void) free(from);
        from=strdup(optarg);
	}
      else
	(void) usage();
      break;
    case   'H'  :
      if ((optarg!=(char *)0)&&(strlen(optarg)>0)) {
        (void) free(remote);
        remote=strdup(optarg);
	}
      else
	(void) usage();
      break;
    case   'b'  :
      (void) free(base);
      base=clm_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default     :
      (void) usage();
      break;
    }
  }
(void) clm_getmillisec();
/*application garbage collector		*/
(void) gac_setcurbase(&debug,(int *)0,base);
(void) clm_chgid(MAILC,d_bin);
gat=gac_open(2,CLEMENT,appname);
if (argc>optind) {
  int i;
  char **emails;

  emails=(char **)0;
  for (i=optind;i<argc;i++) {
    emails=(char **)vec_addlstlst((void **)emails,strdup(argv[i]));
    }
  (void) prc_preptitle(argc,argv,environ);
  (void) clm_setjournal(EVTRT,d_log,MAILC);
  status=work(remote,from,argc-optind,emails);
  (void) jrl_closejournals();
  (void) prc_cleantitle();
  emails=(char **)vec_freelstlst((void **)emails,VECFREE);
  }
(void) clm_trashmem(2,gat);
(void) free(remote);
(void) free(from);
(void) free(base);
(void) closelog();
(void) clm_trapsegv(false);
status*=-1;
if ((debug>2)||(status!=0))
  (void) clm_alert(0,"%s exiting with status '%d'",MAILC,status);
return status;
}
