/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	CMLSNDR					*/
/*	Its purpose is to post an Email via	*/
/*	the clement localport			*/
/*						*/
/************************************************/
#include	<sys/types.h>
#include	<sys/utsname.h>
#include	<pwd.h>
#include	<math.h>
#include	<stdio.h>
#include	<errno.h>
#include	<string.h>
#include	<stdlib.h>
#include	<time.h>
#include	<syslog.h>
#include	<utlgac.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unidns.h"
#include	"unieml.h"
#include	"unisql.h"
#include	"unijrl.h"
#include	"comeml.h"
#include	"gescnf.h"
#include	"gesclm.h"
#include	"lvlcnf.h"
#include	"modckr.h"
#include	"modpro.h"

#define	CLMSNDR	"clmsndr"
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",CLMSNDR,clm_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-h ] "
		      "[-q ] "
		      "[-m myemail] "
		      "[-v ] "
		      "[recipient [recipient]....]\n",CLMSNDR);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-q           : queue mode\n");
(void) fprintf(stderr,"\t\t-m myemail   : originator email override\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) exit(-1);
}
/*
^L
*/
/************************************************/
/*						*/
/*	getting local email information		*/
/*						*/
/************************************************/
static char *getmyemail()

{
char *myemail;
char *domain;
struct passwd *passwd;
struct utsname buname;
int phase;
int proceed;

myemail=(char *)0;
domain=(char *)0;
passwd=(struct passwd *)0;
(void) memset(&buname,'\000',sizeof(buname));
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting login name	*/
      if ((passwd=getpwuid(getuid()))==(struct passwd *)0) {
	(void) clm_alert(0,"Unable to get login name");
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*getting domain name	*/
      if (uname(&buname)<0) {
	(void) clm_alert(0,"Unable to get system name");
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*extracting nodename	*/
      if ((domain=strchr(buname.nodename,'.'))==(char *)0) {
	(void) clm_alert(0,"Node <%s> format incorrect",buname.nodename);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*making email		*/
      (void) asprintf(&myemail,"%s@%s",passwd->pw_name,domain+1);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return myemail;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to check if recipient is	*/
/*	acceptable (local,remote,...)		*/
/*						*/
/************************************************/
static void checkrecipients(SQLFILE *sql,DSPTYP *dsp,char **recipients)

{
#define	OPCH	"clmsndr.c:checkrecipients,"

if (recipients!=(char **)0) {
  register int i;

  for (i=0;recipients[i]!=(char *)0;i++) {
    USRTYP *usr;
    RCPTYP *rcp;

    if ((usr=ckr_finduserstatus(sql,dsp->efrom,recipients[i],dsp->peerip,dsp->locip,true,(char **)0))==(USRTYP *)0) {
      (void) clm_alert(0,"%s we should be able to get a status sessions <%s> reipient <%s>",
                          OPCH,dsp->session,recipients[i]);
      continue;
      }
    rcp=(RCPTYP *)calloc(1,sizeof(RCPTYP));
    rcp->rcpt=strdup(recipients[i]);
    switch (usr->rcptflg) {
      case rcpt_resend		:
	rcp->status=m_toresend;
	break;
      case rcpt_forward		:
	rcp->status=m_toforward;
	break;
      case rcpt_forcefwd	:
	rcp->status=m_mustfwd;
	break;
      case rcpt_foreign		:
	rcp->status=m_relayed;
	break;
      case rcpt_islocal		:
      case rcpt_islocalias	:
	rcp->status=m_direct; 
	break;
      case rcpt_banned	:
	rcp->status=m_quarantine;
	break;
      case rcpt_alidirect	:
      case rcpt_direct		:
	rcp->status=m_direct;
	break;
      case rcpt_inactive	:
        (void) asprintf(&(rcp->why),"%s","account not active");
	(void) fprintf(stderr,"session <%s> recipient <%s> %s\n",
			       dsp->session,rcp->rcpt,rcp->why);
	rcp->status=m_rejected;
	break;
      case rcpt_unknown 	:
        (void) asprintf(&(rcp->why),"%s","account not existing");
	(void) fprintf(stderr,"session <%s> recipient <%s> %s\n",
			       dsp->session,rcp->rcpt,rcp->why);
	rcp->status=m_rejected;
	break;
      case rcpt_boxtb 		:
        (void) asprintf(&(rcp->why),"%s","account email to big");
	(void) fprintf(stderr,"session <%s> recipient <%s> %s\n",
			       dsp->session,rcp->rcpt,rcp->why);
	rcp->status=m_rejected;
	break;
      default			:
	(void) fprintf(stderr,"session <%s> recipient <%s> unexpected status='%d'\n",
			       dsp->session,rcp->rcpt,usr->rcptflg);
	rcp->status=m_rejected;
	break;
      }
    if (rcp!=(RCPTYP *)0)
      dsp->rcpts=eml_addnewrcp(dsp->rcpts,&rcp);
    usr=dbc_freeuserinfo(usr);
    }
  }
(void) pro_reportrcpt(dsp->rcpts);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Core transmit program			*/
/*						*/
/************************************************/
static int transmit(char *myemail,char **recipients,SQLFILE *sql,CLAMTYP **clamav)

{
#define	OPTR	"clmsndr.c:transmit,"

int status;
int todo;
int numdone;
DSPTYP *dsp;
const char *virname;
char **headers;
char *outfile;
FILE *fout;
int phase;
int proceed;

status=0;
todo=0;
numdone=0;
dsp=(DSPTYP *)0;
virname=(char *)0;
headers=(char **)0;
outfile=(char *)0;
fout=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*at least ONE recpient */
      if (vec_sizelstlst((void **)recipients)<1) {
	(void) fprintf(stderr,"missing recipient list");
        status=-(10*(phase+1));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*preparing DSP		*/
      dsp=(DSPTYP *)calloc(1,sizeof(DSPTYP));
      dsp->spmlvl=NAN;
      dsp->spmlim=NAN;
      dsp->auth=1;	/*local relay		*/
      dsp->session=clm_uniquename((unsigned int)0);
      (void) asprintf(&(dsp->locip),"%s",LOCIP4);
      (void) asprintf(&(dsp->peerip),"%s",LOCIP4);
      (void) asprintf(&(dsp->peername),"%s_sender",appname);
      dsp->efrom=strdup(myemail);
      (void) jrl_journalling(false,EVJRL,"SID:  %s received via local pipe to '%s'",dsp->session,CLMSNDR);
      (void) jrl_journalling(false,EVJRL,"%s: <%s>",MAILF,myemail);
      if (recipients!=(char **)0) { /*always	*/
	int i;
	
	for (i=0;recipients[i]!=(char *)0;i++) {
	  (void) clm_cleanstring(recipients[i]);
          (void) jrl_journalling(false,EVJRL,"%s: <%s>",RCPT,recipients[i]);
	  }
	}
      break;
    case 2	:	/*openning outputfile	*/
      (void) asprintf(&outfile,"%s/%s",clm_appdir(d_queue),dsp->session);
      if ((fout=fopen(outfile,"w"))==(FILE *)0) {
	(void) fprintf(stderr,"Unable to open outpufile file <%s> (error=<%s>)",outfile,strerror(errno));
        status=-(10*(phase+1));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*scanning input file	*/
      if (stdin!=(FILE *)0) {	/*always	*/
	int inhead;
        char buffer[2048];


	inhead=true;
	(void) fprintf(fout,"Received: by %s id %s; %s\n",CLMSNDR,dsp->session,clm_ascsysstamp(time((time_t *)0)));
	while (fgets(buffer,sizeof(buffer),stdin)!=(char )0) {
	  (void) clm_skiprtlf(buffer);
	  if (strcmp(buffer,".")==0)
	    break;	/*end of Email		*/
	  if (inhead==true) {
	    if (buffer[0]!='\000')
	      headers=(char **)vec_addlstlst((void **)headers,strdup(buffer));
	    else {
	      inhead=false;
	      (void) fprintf(fout,"%sVersion: %s\n",XCLEM,clm_getvers());
              (void) fprintf(fout,"%sID: <%s>\n",XCLEM,dsp->session);
	      }
	    }
	  (void) fprintf(fout,"%s\n",buffer);
	  dsp->mtaille+=strlen(buffer)+1;
	  }
        (void) fclose(fout);
	}
      if (headers==(char **)0) {
	(void) fprintf(stderr,"Email is Empty?\n");
        status=-(10*(phase+1));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*parsing header	*/
      (void) jrl_journalling(false,EVJRL,"Email size: %d",dsp->mtaille);
      if (headers!=(char **)0){	/*always	*/
	EVLTYP *evl;

        if ((evl=hdr_parseheader(dsp->session,headers))==(EVLTYP *)0) {
	  (void) clm_alert(0,"%s Incorrect E-mail Header",OPTR);
          status=-(10*(phase+1));
	  phase=999;	/*trouble trouble	*/
	  }
	else {
	  if (evl->subject!=(char *)0) {
	    dsp->subject=strdup(evl->subject);
            (void) jrl_journalling(false,EVJRL,"Subject: %s",evl->subject);
	    }
	  if (evl->realfrom!=(char *)0) {
	    dsp->hfrom=strdup(evl->realfrom);
            (void) jrl_journalling(false,EVJRL,"E-From: %s",evl->realfrom);
	    }
	  if (evl->msgid!=(char *)0) 
            (void) jrl_journalling(false,EVJRL,"Message-Id: %s",evl->msgid);
	  evl=hdr_freeevl(evl);
	  }
        headers=(char **)vec_freelstlst((void **)headers,VECFREE);
	}
      break;
    case 5	:	/*checking for virus	*/
      switch (cla_checkvirus(clamav,outfile,&virname)) {
	case cla_clean	:
          (void) jrl_journalling(false,EVJRL,"550 Message accepted for delivery (Session ID=<%s>)",dsp->session);
	  break;
	case cla_virus	:
	  if (virname!=(char *)0) {	/*always*/
	    char *virlog;
	    char *forlogs;
	
	    virlog=eml_ascstat(BADVIR,virname);
	    forlogs=eml_forlogs(virlog,(char *)0);
	    (void) jrl_journalling(false,EVJRL,"%s\n\t  %s",CLAVIR,forlogs);
	    (void) fprintf(stderr,"session <%s> found virus <%s>\n",dsp->session,virname);
	    (void) free(forlogs);
	    (void) free(virlog);
            status=-(10*(phase+1));
	    phase=999;	/*trouble trouble	*/
	    }
	  break;
	case cla_nocheck:
	  (void) fprintf(stderr,"%s session <%s> not checked for virus!\n",OPTR,dsp->session);
	  break;
	}
      break;
    case 6	:	/*check	recipient list	*/
      (void) checkrecipients(sql,dsp,recipients);
      break;
    case 7	:	/*dispatching email	*/
      numdone=ckr_dispatch(sql,dsp,&todo); 
      if (todo!=numdone) {
        status=-(10*(phase+1));
        (void) fprintf(stderr,"session <%s> %d Emails not delivered (expected=%d)\n",
                              dsp->session,todo-numdone,todo);
	}
      break;
    default	:	/*SAFE Guard		*/
      (void) jrl_journalling(true,EVJRL,"Session End");
      if (unlink(outfile)<0)
        (void) fprintf(stderr,"Unable to remove ouput file <%s> (error=<%s>)\n",outfile,strerror(errno));
      outfile=clm_freestr(outfile);
      dsp=eml_cleandispatch(dsp);
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	scan stdin to builld recipients list	*/
/*	detect email				*/
/*						*/
/************************************************/
int scanqueue(char *myemail,SQLFILE *sql,CLAMTYP **clamav)

{
int status;
char *fromemail;
char **recipients;
char buffer[2048];

status=0;
fromemail=strdup(myemail);
recipients=(char **)0;
while (fgets(buffer,sizeof(buffer),stdin)!=(char )0) {
  static VOCTYP vocckr[]={
		{1,"FROM:",(void *)0},
		{2,"RCPT:",(void *)0},
		{3,"MAIL:",(void *)0},
		{0,(char *)0,(void *)0}
		};
  VOCTYP *voc;

  (void) clm_cleanstring(buffer);
  if (strlen(buffer)==0)
    continue;
  if ((voc=clm_getvoca(vocckr,buffer))!=(VOCTYP *)0) {
    switch (voc->code) {
      case 1	:	/*get originator data	*/
	(void) clm_cleanstring(buffer);
	if (strlen(buffer)>0) {
	  (void) free(fromemail);
	  fromemail=strdup(buffer);
	  }
	break;
      case 2	:	/*get recipients list	*/
	(void) clm_cleanstring(buffer);
	if (strlen(buffer)>0) {
	  char **split;

	  if ((split=clm_splitstring(buffer,','))!=(char **)0) {
	    int i;

	    for (i=0;split[i]!=(char *)0;i++) 
	      recipients=(char **)vec_addlstlst((void **)recipients,(void *)split[i]);
	    (void) free(split);
	    }
	  }
	break;
      case 3	:	/*scanning Email	*/
	status=transmit(fromemail,recipients,sql,clamav);
        recipients=(char **)vec_freelstlst((void **)recipients,VECFREE); 
	(void) free(fromemail);
	fromemail=strdup(myemail);
	break;		
      default	:
	(void) fprintf(stderr,"Unknown format <%s>",buffer);
	status=-1;
	break;		
      }
    }
  if (status!=0)
    break;
  }
        recipients=(char **)vec_freelstlst((void **)recipients,VECFREE); 
(void) free(fromemail);
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Prepare all needed resource and start	*/
/*	Email transmission.			*/
/*						*/
/************************************************/
int cycle(int queuemode,char *myemail,char **recipients)

{
#define	OPCY	"clmsndr.c:cycle,"

int status;
const char *empty;
SQLFILE *sql;
CLAMTYP *clamav;
int phase;
int proceed;

status=0;
empty=(const char *)0;
sql=(SQLFILE *)0;
clamav=(CLAMTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening SQL data-base	*/
      (void) clm_setjournal(EVJRL,d_log,"events");
      (void) clm_setjournal(EVCKR,d_log,"checker");
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
	(void) fprintf(stderr,"Unable to open data-base (config default?)\n");
        status=-(phase+1);
	phase=999;
	}
      break;
    case 1	:	/*opening clamav	*/
      clamav=ckr_getclamav(sql,(CLAMTYP *)0);
      break;
    case 2	:	/*scanning queuefile	*/
      if (queuemode==true) 
	status=scanqueue(myemail,sql,&clamav);
      else
	status=transmit(myemail,recipients,sql,&clamav);
      break;
    default	:	/*SAFE Guard		*/
      clamav=cla_closescanner(clamav);
      sql=sql_close(sql);
      (void) jrl_closejournals();
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to create a queue file if	*/
/*	the email is sent via stdin		*/
/*						*/
/************************************************/
char *getqueuefile(int argc,char *argv[])

{
#define	FILFORM	"%s/%s-%05d-%08lx"

char *queuefile;
char *emailfile;
FILE *fout;
int phase;
int proceed;

queuefile=(char *)0;
emailfile=(char *)0;
fout=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true)	{
  switch (phase) {
    case 0	:	/*preparing transit file*/
      (void) asprintf(&emailfile,FILFORM,clm_appdir(d_transit),"emailfile",getpid(),clm_getmillisec());
      if ((fout=fopen(emailfile,"w"))==(FILE *)0) {
        (void) fprintf(stderr,"Unable to open file <%s> (error=<%s>)\n",emailfile,strerror(errno));
	proceed=false;
	}
      break;
    case 1	:	/*writing email file	*/
      if (fout!=(FILE *)0) {	/*always	*/
	char buffer[2048];

	while (fgets(buffer,sizeof(buffer),stdin)!=(char )0) 
	  (void) fprintf(fout,"%s",buffer);
        (void) fclose(fout);
	}
      break;
    case 2	:	/*preparing queue file	*/
      (void) asprintf(&queuefile,FILFORM,clm_appdir(d_transit),"queuefile",getpid(),clm_getmillisec());
      if ((fout=fopen(queuefile,"w"))==(FILE *)0) {
        (void) fprintf(stderr,"Unable to open file <%s> (error=<%s>)\n",emailfile,strerror(errno));
	(void) unlink(emailfile);
	(void) free(queuefile);
	queuefile=(char *)0;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*writing queufile	*/
      if (fout!=(FILE *)0) {	/*always	*/
	int i;

        (void) fprintf(fout,"%s",emailfile);
	for (i=0;i<argc;i++)
          (void) fprintf(fout," %s",argv[i]);
        (void) fprintf(fout,"\n");
        (void) fclose(fout);
        }
      break;
    default	:	/*SAFE Guard		*/
      (void) free(emailfile);
      proceed=false;
      break;
    }
  phase++;
  }
return queuefile;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
int status;
int queuemode;
char *myemail;
char **recipients;
GACTYP *gat;
char c;


clm_foreground=true;
(void) clm_trapsegv(true);
(void) openlog(CLMSNDR,LOG_PID,0);
status=0;
queuemode=false;
myemail=(char *)0;
recipients=(char **)0;
gat=(GACTYP *)0;
base=strdup(base);
while ((c=getopt(argc,argv,"b:d:h:qm:v"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=clm_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'q'  :
      queuemode=true;
      break;
    case   'm'  :
      if (myemail!=(char *)0) {
	(void) fprintf(stderr,"Originator email already set to <%s>, exiting\n",myemail);
        (void) exit(-1);
	}
      myemail=strdup(optarg);
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    }
  }
/*application garbage collector		*/
gat=gac_open(2,appname,CLMSNDR);
(void) gac_setcurbase(&debug,&verbose,base);
(void) clm_chgid(CLMSNDR,d_bin);
if (queuemode==false) {
  int i;

  for (i=optind;i<argc;i++)
     recipients=(char **)vec_addlstlst((void **)recipients,(void *)strdup(argv[i]));
  }
if (myemail==(char *)0) {
  if ((myemail=getmyemail())==(char *)0) {
    (void) fprintf(stderr,"Unable to get local email! (config problem?)\n");
    (void) exit(-1);
    }
  }
(void) prc_preptitle(argc,argv,environ);
(void) clm_loadsysconf(CLEMENT,true);
status=cycle(queuemode,myemail,recipients);
(void) prc_cleantitle();
myemail=clm_freestr(myemail); 
recipients=(char **)vec_freelstlst((void **)recipients,VECFREE); 
(void) clm_trashmem(2,gat);
(void) free(base);
(void) closelog();
(void) clm_trapsegv(false);
if ((debug>2)||(status!=0))
  (void) clm_alert(0,"%s exiting with status '%d'",CLMSNDR,status);
return status;
}
