/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	BASANF:					*/
/*	Implement all routine to manage IP	*/
/*	number conversion.			*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<ctype.h>
#include	<string.h>
#include	<stdio.h>
#include	<netdb.h>
#include	<malloc.h>
#include	<arpa/inet.h>

#include	"lowtyp.h"
#include	"subclm.h"
#include	"basafn.h"
/*

*/
/************************************************/
/*						*/
/*	Routine to compare 2 IP num, can be 	*/
/*	IPV4 or IPV6 IP number.			*/
/*						*/
/************************************************/
static int cmpip(AFNTYP *ipnum1,AFNTYP *ipnum2,int mask)

{
int areequal;
int maskbyt;

areequal=false;
if (mask>128)
  mask=128;
maskbyt=mask/8;
if (memcmp(ipnum1->ip,ipnum2->ip,maskbyt)==0) {
  areequal=true;
  mask%=8;
  if (mask!=0) {
    int maskbit;
    int b1;
    int b2;

    maskbit=(~0U<<(8-mask)); 
    b1=((int)ipnum1->ip[maskbyt])&maskbit;
    b2=((int)ipnum2->ip[maskbyt])&maskbit;
    if (b1!=b2)
      areequal=false;
    }
  }
return areequal;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to check if the string is a	*/
/*	pure IPV4 or IPV6 IP.			*/
/*						*/
/************************************************/
int afn_ispureip(char *strip)

{
register int ispure;

ispure=false;
if (strip!=(char *)0) {
  register char *ptr;

  ispure=true;
  for (ptr=strip;(ispure==true)&&(*ptr!='\000');ptr++) {
    switch (*ptr) {
      case '.'	:
      case ':'	:
	break;
      default	:
        if (isdigit(*ptr)==0) 
	  ispure=false;
	break;
      }
    }
  }
return ispure;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert a addr struct to an	*/
/*	IP number in AFNTYP structure.		*/
/*						*/
/************************************************/
AFNTYP *afn_getaddrinfo(struct addrinfo *rp)

{
AFNTYP *afnnum;

afnnum=(AFNTYP *)0;
if (rp!=(struct addrinfo *)0) {
  u_char *ptr;

  ptr=(u_char *)0;
  afnnum=calloc(1,sizeof(AFNTYP));
  afnnum->afntype=rp->ai_family;
  switch (rp->ai_family) {
    case AF_INET	:
      afnnum->afnmask=32;
      ptr=(u_char *)&(((struct sockaddr_in *)rp->ai_addr)->sin_addr.s_addr);
      break;
    case AF_INET6	:
      afnnum->afnmask=128;
      ptr=(u_char *)&(((struct sockaddr_in6 *)rp->ai_addr)->sin6_addr.s6_addr);
      break;
    default		:	/*trouble trouble	*/
      afnnum=afn_freeipnum(afnnum);
      break;
    }
  if (ptr!=(u_char *)0)
    (void) memmove(afnnum->ip,ptr,afnnum->afnmask/8);
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert clean string to be	*/
/*	IP number in AFNTYP structure.		*/
/*						*/
/************************************************/
AFNTYP *afn_getipnum(char *cleanipstr)

{
#define	ZIPV4	"0.0.0.0"

AFNTYP *afnnum;
int phase;
int proceed;

afnnum=(AFNTYP *)0;
phase=0;
proceed=(cleanipstr!=(char *)0);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets say ip format ok	*/
      afnnum=(AFNTYP *)calloc(1,sizeof(AFNTYP));
      afnnum->afntype=AF_INET;
      afnnum->afnmask=32;
      if (strchr(cleanipstr,':')==(char *)0) 
	phase++;	/*must be IPV4		*/
      break;
    case 1	:	/*is it declared IPV6	*/
      if (strcasestr(cleanipstr,STRIPV6)==cleanipstr) 
	cleanipstr+=strlen(STRIPV6);
      afnnum->afntype=AF_INET6;
      afnnum->afnmask=128;
      break;
    case 2	:	/*checking '0' situation*/
      if (strcmp(cleanipstr,"0")==0) 
	cleanipstr=ZIPV4;
      break;
    case 3	:	/*lets convert it	*/
      if (inet_pton(afnnum->afntype,cleanipstr,afnnum->ip)<=0) 
	afnnum=afn_freeipnum(afnnum);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to mask IPNUM according the 	*/
/*	number of by needed			*/
/*						*/
/************************************************/
AFNTYP *afn_maskip(AFNTYP *afnnum,int cidr)

{
if (afnnum!=(AFNTYP *)0) {
  unsigned int num;
  int mxcidr;


  mxcidr=32;
  if (afnnum->afntype==AF_INET6)
    mxcidr=128;
  if (cidr>mxcidr)
    cidr=mxcidr;
  afnnum->afnmask=cidr;
  num=cidr/8;
  cidr%=8;
  if (cidr!=0) {
    afnnum->ip[num]&=(~0U<<(8-cidr));
    num++;
    }
  if (num<AFT)
    (void) memset(afnnum->ip+num,'\000',AFT-num);
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to fee memory used by an AFNTYP	*/
/*						*/
/************************************************/
AFNTYP *afn_freeipnum(AFNTYP *afnnum)

{
if (afnnum!=(AFNTYP *)0) {
  (void) free(afnnum);
  afnnum=(AFNTYP *)0;
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to normalize IP number format	*/
/*	Memory allocated need to be freed later	*/
/*	on.					*/
/*						*/
/************************************************/
char *afn_ipnormal(AFNTYP *afnnum)

{
char *ipnorm;

ipnorm=(char *)0;
if (afnnum!=(AFNTYP *)0) {
  switch (afnnum->afntype) {
    case AF_INET	:
      (void) asprintf(&ipnorm,"%03d.%03d.%03d.%03d",
			      afnnum->ip[0],afnnum->ip[1],
			      afnnum->ip[2],afnnum->ip[3]);
      break;
    case AF_INET6	:
      (void) asprintf(&ipnorm,"%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
			      afnnum->ip[0],afnnum->ip[1],
			      afnnum->ip[2],afnnum->ip[3],
			      afnnum->ip[4],afnnum->ip[5],
			      afnnum->ip[6],afnnum->ip[7],
			      afnnum->ip[8],afnnum->ip[9],
			      afnnum->ip[10],afnnum->ip[11],
			      afnnum->ip[12],afnnum->ip[13],
			      afnnum->ip[14],afnnum->ip[15]);
      break;
    default		:
      (void) clm_alert(0,"subafn.c:afn_ipnormal, unable to fin IP type (bug?)");
    } 
  }
return ipnorm;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to compare two ipnum by taking	*/
/*	mask ip_ipbits, return true if ipbits	*/
/*	are identical, false otherwise.		*/
/*	return -1 if unexpected type		*/
/*						*/
/************************************************/
int afn_cmpipnum(AFNTYP *afnnum1,AFNTYP *afnnum2,int mask)

{
int areequal;
int phase;
int proceed;

areequal=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*two ipnum really	*/
      if ((afnnum1==(AFNTYP *)0)||(afnnum2==(AFNTYP *)0))
	proceed=false;	/*no need to go further	*/
      break;
    case 1	:	/*is mask 0, all IP ok	*/
      if (mask==0) {
	areequal=true;
	proceed=false;
	}
      break;
    case 2	:	/*two ipnum same type	*/
      if (afnnum1->afntype!=afnnum2->afntype)
	proceed=false;	/*no need to go further	*/
      break;
    case 3	:	/*lets compare ip	*/
      if (mask>128)
	mask=128;
      switch(afnnum1->afntype) {
	case AF_INET	:
	  if (mask>32)
	    mask=32;	/*NO break		*/
	case AF_INET6	:
	  areequal=cmpip(afnnum1,afnnum2,mask);
	  break;
	default		:
	  errno=EAFNOSUPPORT;
	  areequal=-1;
	  break;
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return areequal;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a string with	*/
/*	ipnum address in clear format		*/
/*	Address need to	freed.			*/
/*						*/
/************************************************/
char *afn_stripnum(AFNTYP *afnnum)

{
char *strip;

strip=(char *)0;
if (afnnum!=(AFNTYP *)0) {
  char *markr;
  char host[NI_MAXHOST];

  markr="";
  switch (afnnum->afntype) {
    case AF_INET6	:
      markr=STRIPV6;		/*NO Break	*/
    case AF_INET	:
      if (inet_ntop(afnnum->afntype,&(afnnum->ip),host,sizeof(host))!=(char *)0) 
        (void) asprintf(&strip,"%s%s",markr,host);
      else
	(void) asprintf(&strip,"[afn_stripnum error=<%s>]",strerror(errno));
      break;
    default		:
      (void) asprintf(&strip,"[ipvx:0.%d]",afnnum->afntype);
      break;
    }
  }
return strip;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a string with	*/
/*	ipnum address and MASK in clear format	*/
/*	Address need to	freed.			*/
/*						*/
/************************************************/
char *afn_stripnummask(AFNTYP *afnnum,int mask)

{
char *strip;

if ((strip=afn_stripnum(afnnum))!=(char *)0) {
  char *stripmask;

  (void) asprintf(&stripmask,"%s/%d'",strip,mask);
  (void) free(strip);
  strip=stripmask;
  }
return strip;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a revers address as	*/
/*	dynamically allocated string		*/
/*						*/
/************************************************/
char *afn_reversipnum(AFNTYP *afnnum)

{
char * reversip;

reversip=(char *)0;
if (afnnum!=(AFNTYP *)0) {
  int i;

  reversip=strdup("");
  switch (afnnum->afntype) {
    case AF_INET6	:
      for (i=16;i>0;i--) {
	char *ptr;

	ptr=reversip;
	(void) asprintf(&reversip,"%s%x.%x.",ptr,afnnum->ip[i-1]&0xf,afnnum->ip[i-1]>>4);
	(void) free(ptr);
	}
      break;
    case AF_INET	:
      for (i=4;i>0;i--) {
	char *ptr;

	ptr=reversip;
	(void) asprintf(&reversip,"%s%d.",ptr,afnnum->ip[i-1]);
	(void) free(ptr);
	}
      break;
    default		:
      (void) asprintf(&reversip,"%d.0.0.0.0.ukn.",afnnum->afntype);
      break;
    }
  reversip[strlen(reversip)-1]='\000';
  }
return reversip;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a revers generic	*/
/*	domain related to ipnum			*/
/*	dynamically allocated string		*/
/*						*/
/************************************************/
const char *afn_reversdom(AFNTYP *afnnum)

{
const char *reversdom;

reversdom=(const char *)0;
if (afnnum!=(AFNTYP *)0) {
  switch (afnnum->afntype) {
    case AF_INET6	:
      reversdom="ip6.arpa";
      break;
    case AF_INET	:
      reversdom="in-addr.arpa";
      break;
    default		:
      reversdom="ukn.arpa";
      break;
    }
  }
return reversdom;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a formated IP number*/
/*	adding "ipv6" if it is an IPV6 IP number*/
/*	Memory return is a dynamic memory, need	*/
/*	to bee freed later on.			*/
/*						*/
/************************************************/
char *afn_formatip(char *ipnum)

{
#define	OPFO	"basafn.c:afn_formatip,"

char *standard;
struct addrinfo hints;
struct addrinfo *res;
AFNTYP *addrnum;
int status;
int phase;
int proceed;

standard=(char *)0;
(void) memset(&hints,'\000',sizeof(hints));
hints.ai_family=PF_UNSPEC;
hints.ai_socktype=SOCK_STREAM;
hints.ai_flags=HINTFLG;
res=(struct addrinfo *)0;
addrnum=(AFNTYP *)0;
status=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is it Unknown already	*/
      if (strcmp(ipnum,UKNIP)==0) {
	standard=strdup(ipnum);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*extracting addrinfo	*/
      if ((status=getaddrinfo(ipnum,"",&hints,&res))!=0) {
        (void) clm_alert(0,"%s Unable to find addrinfo for <%s> (error=<%s>)",
			   OPFO,ipnum,gai_strerror(status));
	standard=strdup("IP??");
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*getting AFNUM		*/
      if ((addrnum=afn_getaddrinfo(res))==(AFNTYP *)0) {
        (void) clm_alert(0,"%s Found unknown inet family for <%s> IP number",
			   OPFO,ipnum);
	standard=strdup("IP_family??");
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*normalising IP	*/
      if (addrnum!=(AFNTYP *)0) {	/*always*/
	standard=afn_stripnum(addrnum);
        addrnum=afn_freeipnum(addrnum);
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return standard;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return an unformat an IPV4	*/
/*	number to be compatible with inet_pton	*/
/*	requested format.			*/
/*						*/
/************************************************/
char *afn_unformatip(char *ipnum)

{
if ((ipnum!=(char *)0)&&strchr(ipnum,'.')!=(char *)0) {
  char *str;
  int num[4];

  if ((str=strrchr(ipnum,':'))!=(char *)0)
    str++;	/*mixing IPV6+IPV4		*/
  else
    str=ipnum;
  if (sscanf(str,"%d.%d.%d.%d",num,num+1,num+2,num+3)==4) {
    char locip[20];

    num[0]%=256;
    num[1]%=256;
    num[2]%=256;
    num[3]%=256;
    (void) sprintf(locip,"%d.%d.%d.%d",num[0],num[1],num[2],num[3]);
    if (strlen(locip)<=strlen(str))
      (void) strcpy(str,locip);
    }
  }
return ipnum;
}
