/************************************************/
/*						*/
/*	Procedure implementation to manage	*/
/*	PosgreSQL database.			*/
/*						*/
/************************************************/
#include	<stdlib.h>
#include	<fcntl.h>
#include	<syslog.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<errno.h>
#include	<signal.h>
#include	<string.h>
#include	<sys/stat.h>
#include        <pg_config.h>

#include        <libpq-fe.h>

#include	"lowtyp.h"
#include	"utlprc.h"
#include	"subclm.h"
#include	"baspos.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to execut a command		*/
/*	data-base daemon.			*/
/*						*/
/************************************************/
static PGresult *action(PGconn *pf,char *directive)

{
#define	RELAX	1000000

PGresult *statut;
int try;

statut=(PGresult *)0;
try=0;
while (statut==(PGresult *)0) {
  try++;
  switch(PQstatus(pf)) {
    case CONNECTION_OK	:
      if ((statut=PQexec(pf,directive))!=(PGresult *)0)
	break;
      else 
        (void) clm_alert(0,"Command <%s> failed, (error=<%s>), retrying",
			   directive,PQerrorMessage(pf));
	/*NO Break, want DB reset	*/	
    default		:
      (void) clm_alert(0,"baspos.c,action: reseting postgres "
			 "connection (try='%d')",try);
      (void) usleep(RELAX/2);
      (void) PQreset(pf);
      (void) usleep(RELAX/2);
      break;
    }
  if (try>10) {
    (void) clm_alert(0,"baspos.c,action: Unable to carry command <%s>, "
		       "Postgres error!",directive);
    break;    
    }
  }
return statut;
}
#ifdef NEEDED
Removed 2011 Mars 7
/*

*/
/************************************************/
/*						*/
/*	Procedure to set an advisory lock	*/
/*	table MUST be a number definition	*/
/*						*/
/************************************************/
static int mnglockadvisory(POSFILE *pf,char *table,int lock)

{
int status;
char *strlock;
POSRESULT *prs;
char cmd[100];

status=-1;
strlock="lock";
if (lock==false)
  strlock="unlock";
(void) snprintf(cmd,sizeof(cmd),"SELECT pg_advisory_%s(%s)",strlock,table);
if ((prs=pos_tupple(pf,cmd))!=(POSRESULT *)0) {
  (void) PQclear(prs);
  status=0;
  }
return status;
}
#endif
/*

*/
/************************************************/
/*						*/
/*	Procedure to make sure the PostgreSQL	*/
/*	daemon is up and running.		*/
/*						*/
/************************************************/
int pos_isdaemonup(const char *host,const char *sqlport,const char *dbname,int tentative)

{
int itisok;
int i;

itisok=false;
for (i=0;i<tentative;i++) {
  PGconn *pf;

  pf=(PGconn *)0;
  (void) sleep(1);
  if ((pf=PQsetdb(host,sqlport,(char *)0,(char *)0,dbname))!=(PGconn *)0) {
    if (PQstatus(pf)==CONNECTION_OK) {
      (void) clm_alert(1,"PostgreSQL daemon confirmed up and running");
      itisok=true;
      }
    (void) PQfinish(pf);
    }
  if (itisok==true)
    break;
  if (sigterm==true)
    break;
  }
if (itisok!=true) {
  (void) clm_alert(0,"PostgreSQL daemon NOT confirmed up and running");
  }
return itisok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make a connection to	*/
/*	PostgreSQL daemon.			*/
/*						*/
/************************************************/
POSFILE *pos_open(const char *host,const char *sqlport,const char *dbname)

{
#define	OPOP	"baspos.c:pos_open,"

PGconn *pf;
char *z;
int phase;
int proceed;

pf=(PGconn *)0;
z=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the data-base	*/
      if ((pf=PQsetdbLogin(host,sqlport,z,z,dbname,z,z))==(PGconn *)0) {
        (void) clm_alert(0,"%s Unable to open database '%s' on host <%s:%s>",
			   OPOP,dbname,host,sqlport);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*is database ready?	*/
      if (PQstatus(pf)==CONNECTION_BAD) {
        (void) clm_alert(0,"%s Connection to database '%s' failed, cause '%s'",
			   OPOP,dbname,PQerrorMessage(pf));
        (void) PQfinish(pf);
        (void) sleep(2);
        pf=(PGconn *)0;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*do we need to trace	*/
      if (debug>9) {
	char *dbgname;
	FILE *fichier;

	(void) asprintf(&dbgname,"%s/%s",clm_appdir(d_dbgs),dbname);
	if ((fichier=fopen(dbgname,"a+"))!=(FILE *)0)
	  (void) PQtrace(pf,fichier);
	else
          (void) clm_alert(0,"%s Unable to open PQtrace file <%s>",OPOP,dbgname);
	(void) free(dbgname);
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    } 
  phase++;
  }
return (POSFILE *)pf;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to disconnect  from a 	*/
/*	PostgreSQL daemon.			*/
/*						*/
/************************************************/
POSFILE *pos_close(POSFILE *pf)

{
if (pf!=(POSFILE *)0) {
  (void) PQuntrace((PGconn *)pf);
  (void) PQfinish((PGconn *)pf);
  pf=(POSFILE *)0;
  }
else {
  (void) clm_alert(0,"baspos.c,pos_close: database "
		     "access pointer empty (bug?)");
  }
return pf;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the default Postgres*/
/*	access port number.			*/
/*						*/
/************************************************/
const char *pos_getdfltport()

{
return DEF_PGPORT_STR;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do a data-base command	*/
/*	"insert","update", return the number	*/
/*	of row touched by the SQL command,	*/
/*	-1 otherwise.				*/
/*						*/
/************************************************/
int pos_action(POSFILE *pf,char *command)

{
int status;
PGresult *pgstatut;

status=-1;
if ((pgstatut=action((PGconn *)pf,command))!=(PGresult *)0) {
  if (PQresultStatus(pgstatut)!=PGRES_COMMAND_OK) {
    (void) clm_alert(0,"Command <%s> failed, (error=<%s> pgstatut='%d')",
		       command,PQerrorMessage(pf),PQresultStatus(pgstatut));
    }
  else {
    status=atoi(PQcmdTuples(pgstatut));
    }
  (void) PQclear(pgstatut);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract data from the	*/
/*	data-base				*/
/*	Return POSRESULT status pointer (can	*/
/*	be null if not successful).		*/
/*						*/
/************************************************/
POSRESULT *pos_tupple(POSFILE *pf,char *command)

{
PGresult *pgstatut;

if ((pgstatut=action((PGconn *)pf,command))!=(PGresult *)0) {
  if (PQresultStatus(pgstatut)!=PGRES_TUPLES_OK) {
    (void) clm_alert(0,"SQL command <%s> failed",command);
    (void) PQclear(pgstatut);
    pgstatut=(PGresult *)0;
    }
  }
return (POSRESULT *)pgstatut;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the number of tupple*/
/*	related to a previous search.		*/
/*						*/
/************************************************/
int pos_nbrtupple(POSRESULT *posresult)

{
return PQntuples((PGresult *)posresult);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure clean the result area		*/
/*	previously allocated via pos_tupple	*/
/*	related to a previous search.		*/
/*						*/
/************************************************/
POSRESULT *pos_dropresult(POSRESULT *posresult)

{
(void) PQclear(posresult);
return (POSRESULT *)0;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to retrieve specific field in	*/
/*	specific row, previously located via	*/
/*	pos_tupple.				*/
/*						*/
/************************************************/
char *pos_getvalue(POSRESULT *posresult,int row,char *fieldname)

{
char *valeur;
int position;

valeur=(char *)0;
if ((position=PQfnumber((PGresult *)posresult,fieldname))<0) {
  (void) clm_alert(0,"baspos.c,pos_getvalue:, field '%s' unknown",fieldname);
  return valeur;
  }
if (PQgetisnull((PGresult *)posresult,row,position)==false)
  valeur=PQgetvalue((PGresult *)posresult,row,position);
return valeur;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to retrieve specific field in	*/
/*	specific row, accessed with the column	*/
/*	number.					*/
/*						*/
/************************************************/
char *pos_getfield(POSRESULT *posresult,int row,int column)

{
return PQgetvalue((PGresult *)posresult,row,column);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a unix time in acsii*/
/*	format (data-base compatible).		*/
/*						*/
/************************************************/
const char *pos_fromunixtime(time_t timestamp)

{
#define TSTAMP   "%Y-%m-%d %H:%M:%S"

static char unixdate[50];

struct tm *tminfo;

tminfo=localtime(&timestamp);
(void) strftime(unixdate,sizeof(unixdate),TSTAMP,tminfo);
return unixdate;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a unix time_t	*/
/*	from a ascii string (data-base 		*/
/*	compatible).				*/
/*						*/
/************************************************/
time_t pos_tounixtime(char *date)

{
#define	DBDTOUNIX	"%Y-%m-%d %H:%M:%S"
time_t datetime;

datetime=(time_t)0;
if (date!=(char *)0) {
  char *ptr;
  struct tm tm;

  if ((ptr=strchr(date,'.'))!=(char *)0) {
    *ptr='\000';
    }
  (void) memset(&tm,'\000',sizeof(struct tm));
  tm.tm_isdst=-1;
  if (strptime(date,DBDTOUNIX,&tm)!=(char *)0) {
    datetime=mktime(&tm);
    }
  else {
    (void) clm_alert(0,"baspos.c,pos_tounixtime: Unable to convert <%s> "
		       "to time_t",date);
    }
  }
return datetime;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request a lock to access 	*/
/*	set of table in the PostgreSQL data-base*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*	Lock can be 'soft' (lock only the row)	*/
/*	or hard (lock the all table in update)	*/
/*						*/
/************************************************/
int pos_getlock(POSFILE *posfile,LCKTYP **lcklst)

{
#define LOCKMOD "LOCK TABLE %s %s"

int status;
int phase;
int proceed;

status=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Lets Begin		*/
      if ((status=pos_action(posfile,"BEGIN"))<0) 
	phase=999;	/*trouble trouble	*/
      break;
    case 1	:	/*locks to establish	*/
      if (lcklst!=(LCKTYP **)0) {
	int i;
	char fullcmd[300];

	for (i=0;lcklst[i]!=(LCKTYP *)0;i++) {
	  const char *excl;

	  excl="ACCESS SHARE";	
	  if (lcklst[i]->lock==dbd_lockwrite)
	    excl="SHARE ROW EXCLUSIVE";
          (void) snprintf(fullcmd,sizeof(fullcmd),"LOCK TABLE %s IN %s MODE",lcklst[i]->tblname,excl);
	  if ((status=pos_action(posfile,fullcmd))<0) {
	    phase=999;	/*trouble trouble	*/
	    break;
	    }
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request a lock action 	*/
/*	in case we update table component.	*/
/*						*/
/************************************************/
int pos_lock(POSFILE *posfile)

{
return pos_action(posfile,"BEGIN");
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to release pending lock 	*/
/*	set to restrict access on PostgreSQL	*/
/*	data-base tables.			*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int pos_unlock(POSFILE *posfile,int commit)

{
char *cmd;

cmd="COMMIT";
if (commit==false)
  cmd="ROLLBACK";
return pos_action(posfile,cmd);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the current sequence	*/
/*	number					*/
/*						*/
/************************************************/
u_vlong pos_getseq(POSFILE *posfile,char *sequence)

{
u_vlong seqnum;
PGresult *pgstatut;
char cmd[200];

seqnum=(u_vlong)0;
(void) snprintf(cmd,sizeof(cmd),"SELECT last_value FROM %s",sequence); 
if ((pgstatut=pos_tupple(posfile,cmd))!=(PGresult *)0) {
  seqnum=atoll(pos_getvalue(pgstatut,0,"last_value"));
  (void) PQclear((PGresult *)pgstatut);
  }
return seqnum;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to increment the sequence	*/
/*	number					*/
/*						*/
/************************************************/
u_vlong pos_incseq(POSFILE *posfile,char *sequence)

{
u_vlong seqnum;
PGresult *pgstatut;
char cmd[200];

seqnum=(u_vlong)0;
(void) snprintf(cmd,sizeof(cmd),"SELECT NEXTVAL('%s')",sequence); 
if ((pgstatut=action(posfile,cmd))!=(PGresult *)0){
  if (PQresultStatus(pgstatut)==PGRES_TUPLES_OK) 
    seqnum=pos_getseq(posfile,sequence);
  else
    (void) clm_alert(0,"SQL command <%s> failed",cmd);
  (void) PQclear((PGresult *)pgstatut);
  }
return seqnum;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to detect and 'clean' any	*/
/*	single quote within a string, this is	*/
/*	mandatory for string variable in the	*/
/*	data-base.				*/
/*	Single quote "'" are doubled becoming	*/
/*	"''"					*/
/*						*/
/************************************************/
char *pos_cleanquote(register char *dstr)

{
char *cleanstr;
int taille;

cleanstr=strdup("E'");
taille=strlen(cleanstr)+2;
if (dstr!=(char *)0) {
  int max;
  int i,p;
  char *newclean;

  /*to double char ' in the string			*/
  max=strlen((const char *)dstr);
  taille+=max;
  if ((newclean=(char *)realloc(cleanstr,taille))!=(char *)0) {
    cleanstr=newclean;
    p=strlen(cleanstr);
    for (i=0;i<max;i++,p++,dstr++) {
      switch (*dstr) {
        case '\\'		:
        case '\''		:
          taille++;
          if ((newclean=(char *)realloc(cleanstr,taille))!=(char *)0) {
	    cleanstr=newclean;
            cleanstr=(char *)realloc(cleanstr,taille);
            cleanstr[p]=*dstr;
	    }
          p++;
          break;
        default		:
	  break;
        }
      cleanstr[p]=*dstr;
      }
    cleanstr[p]='\000';
    }
  }
(void) strcat(cleanstr,"'");
return cleanstr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format an ipnum to be	*/
/*	stored within an 'inet' postgresql	*/
/*	record.					*/
/*						*/
/************************************************/
char *pos_toipbin(char *ipnum)

{
char *ipbin;

(void) asprintf(&ipbin,"'%s'::inet",ipnum);
return ipbin;
}
