/************************************************/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESTLS:					*/
/*	Take care of all TSL channel management	*/
/*						*/
/************************************************/
#include	<string.h>
#include	<syslog.h>
#include	<unistd.h>
#include	<stdarg.h>
#include	<fcntl.h>
#include	<signal.h>
#include	<netdb.h>
#include	<setjmp.h>
#include	<poll.h>

#define		OPENSSL_NO_KRB5
#include 	<openssl/x509.h>
#include 	<openssl/x509v3.h>
#include 	<openssl/safestack.h>
#include	<openssl/ssl.h>
#include	<openssl/err.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"utlprc.h"
#include	"subclm.h"
#include	"bascnv.h"
#include	"basafn.h"
#include	"unieml.h"
#include	"gestls.h"

#define	MXBUF	16384	/*char in buffer size	*/
#define	MXTRY	30	/*30 second max to 	*/
			/*establish an SSL	*/
			/*connexion		*/

typedef	struct	{
	int tls;	/*connexion en mode TLS	*/
	int handle;	/*direct link handle	*/
	int traceon;	/*trace exchg to stdout	*/
	char **traces;	/*current trace		*/
	time_t lastused;/*last time handle used	*/
	char *rmtname;	/*remote side name	*/
	char *port;	/*remote port number	*/
	int goteof;	/*EOF detected		*/
	int incpt;	/*# char in the buffer	*/
	char *carin;	/*in char buffer	*/
	char *cipherid;	/*SSL version connection*/
	SSL_CTX *ctx;	/*context encryption	*/
	SSL *ssl;	/*ref connection TLS	*/
	BIO *bio;	/*SSL basic I/O		*/
	}TFILE;
/*

*/
/************************************************/
/*						*/
/*	procedure to clear and return socket	*/
/*	status.					*/
/*	Return 0 if ok, not null if trouble	*/
/*						*/
/************************************************/
static int checksockstat(int handle) 

{
int status;
socklen_t len;

status=0;
len=sizeof(status);
if (getsockopt(handle,SOL_SOCKET,SO_ERROR,&status,&len)<0) {
  status=errno;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Setting up the CTX information for	*/
/*	the connection.				*/
/*						*/
/************************************************/
static SSL_CTX *setup_tls_ctx()

{
#define	CA_TRUSTED	"CA_TRUSTED"
#define	CA_CLEMENT	"CA_CLEMENT"
#define TRUSTED		"trusted"
//#define SSL_CIPHER_LIST	"ALL:!LOW"
#define SSL_CIPHER_LIST	"DEFAULT"

SSL_CTX *ctx;
int mode;
int trouble;
const char *certdir;
char *cafile;
char *pemfile;
char *cipherlist;
char *ptr;
int phase;
int proceed;

ctx=(SSL_CTX *)0;
mode=SSL_VERIFY_NONE;
trouble=true;
phase=0;
proceed=true;
certdir=clm_appdir(d_tls);
while (proceed==true) {
  char *errmsg;

  errmsg=(char *)0;
  switch (phase) {
    case  0	:	/*getting CA file	*/
      if ((ptr=getenv(CA_TRUSTED))!=(char *)0) 
	cafile=strdup(ptr);
      else
	(void) asprintf(&cafile,"%s/%s.pem",certdir,TRUSTED);
      if ((ptr=getenv(CA_CLEMENT))!=(char *)0) 
	pemfile=strdup(ptr);
      else
	(void) asprintf(&pemfile,"%s/%s.pem",certdir,appname);
      break;
    case  1	:	/*getting a CTX	data	*/
      if ((ctx=SSL_CTX_new(SSLv23_method()))==(SSL_CTX *)0) {
	errmsg="CTX_new";
        phase=999;	/*trouble trouble	*/
	}
      break;
    case  2	:	/*loading CA verify dir	*/
      if (SSL_CTX_load_verify_locations(ctx,cafile,(char *)0)==0) {
	errmsg="verify_location";
        phase=999;	/*trouble trouble	*/
        }
      break;
    case  3	:	/*loading CA verify dir	*/
      if (SSL_CTX_set_default_verify_paths(ctx)==0) {
	errmsg="verify_paths";
        phase=999;	/*trouble trouble	*/
        }
      break;
    case  4	:	/*loading appl certificat*/
      if (SSL_CTX_use_certificate_chain_file(ctx,pemfile)!=1) {
	errmsg="use_certificat";
        phase=999;	/*trouble trouble	*/
        }
      break;
    case  5	:	/*loading appl certificat*/
      if (SSL_CTX_use_PrivateKey_file(ctx,pemfile,SSL_FILETYPE_PEM)!=1) {
	errmsg="use_privatekey";
        phase=999;	/*trouble trouble	*/
        }
      break;
    case  6	:	/*verify management	*/
      (void) SSL_CTX_set_purpose(ctx,X509_PURPOSE_ANY);
      (void) SSL_CTX_set_verify(ctx,mode,(int(*)())0);
      (void) SSL_CTX_set_verify_depth(ctx,5);
      (void) SSL_CTX_set_options(ctx,SSL_OP_ALL);
      if ((cipherlist=getenv("SSL_CIPHER_LIST"))==(char *)0)
	cipherlist=SSL_CIPHER_LIST;
      if (SSL_CTX_set_cipher_list(ctx,cipherlist)==0) {
	errmsg="set_cipher";
        phase=999;	/*trouble trouble	*/
        }
      break;
    case  7	:	/*allowing partial write*/
      (void) SSL_CTX_set_mode(ctx,SSL_MODE_ENABLE_PARTIAL_WRITE);
      trouble=false;
      break;
    default	:	/*SAFE guard		*/
      if (trouble==true) {
    	(void) clm_alert(0,"gestls.c:setup_tls_ctx, error '%d', trouble with <%s>",
		            phase,errmsg);
        if (ctx!=(SSL_CTX *)0) {
	  (void) SSL_CTX_free(ctx);
	  ctx=(SSL_CTX *)0;
          }
	}
      pemfile=clm_freestr(pemfile);
      cafile=clm_freestr(cafile);
      proceed=false;
      break;
    }
  phase++;
  }
return ctx;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display an TLS error	*/
/*						*/
/************************************************/
static void tlsshowerror(int dbg,SSL *ssl,int code,char *fmt,...)

{
int sslerr;
char *buff;
va_list args;
char info[300];

va_start(args,fmt);
buff="Premature EOF";
(void) vsnprintf(info,sizeof(info),fmt,args);
(void) clm_alert(dbg,info);
switch((sslerr=SSL_get_error(ssl,code))) {
  case SSL_ERROR_ZERO_RETURN		:
    buff="zero return";
    break;
  case SSL_ERROR_WANT_READ		:
    buff="read error";
    break;
  case SSL_ERROR_WANT_WRITE		:
    buff="write error";
    break;
  case SSL_ERROR_WANT_CONNECT		:
    buff="connect error";
    break;
  case SSL_ERROR_WANT_X509_LOOKUP	:
    buff="X509 lookup error";
    break;
  case SSL_ERROR_SYSCALL		:
    if (ERR_get_error()==0)
      buff="Premature EOF";
    else
      buff="syscall error";
    break;
  case SSL_ERROR_SSL			:
    buff=ERR_error_string(ERR_get_error(),(char *)0);
    break;
  default				:
    (void) clm_alert(0,"%s Unexpected sslerr='%d'","gestls.c:tlsshowerror",sslerr);
    break;
  }
(void) clm_alert(dbg,"SSL error=<%s>",buff);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to build a ID string about	*/
/*	the ciphered connection under TLS	*/
/*						*/
/************************************************/
static char *tlsgetcipherid(SSL *ssl)

{
#define MXID 72

char *cipherid;
const SSL_CIPHER *cipher;

cipherid=(char *)0;
if ((cipher=SSL_get_current_cipher(ssl))!=(const SSL_CIPHER *)0) {
  char *version; 
  const char *name; 
  const char *verif;
  int bits;

  version=SSL_CIPHER_get_version(cipher);
  name=SSL_CIPHER_get_name(cipher);
  bits=SSL_CIPHER_get_bits(cipher,0);
  if (strcmp(version,"SSLv3")==0)
    version="TLSv1/SSLv3";
  cipherid=(char *)calloc(1,MXID);
  switch(SSL_get_verify_mode(ssl)) {
    case (SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE)	:
    case (SSL_VERIFY_NONE)				:
      switch(SSL_get_verify_result(ssl)) {
	case (X509_V_OK)	:
          verif="OK";
	  break;
	default			:
          verif="FAIL";
	  break;
	}
      break;
    default			:
      verif="NO";
      break;
    }
  (void) snprintf(cipherid,MXID,"(version=%s cipher=%s bits=%d verify=%s)",
				  version,name,bits,verif);
  }
return cipherid;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to accept a proposal to	*/
/*	encrypt a communication channel.	*/
/*	Return -1 if trouble, 0 otherwise.	*/
/*						*/
/************************************************/
static int tlsaccept(TFILE *tls)

{
#define	OPAC	"gestls.c:tlsaccept,"

int statut;
X509 *peer;
int tic;
int done;

statut=-1;
peer=(X509 *)0;
tic=0;
done=false;
if ((tls->bio=BIO_new_fd(tls->handle,BIO_NOCLOSE))==(BIO *)0) {
  (void) clm_core_dump("%s Unable to get the BIO (error=<%s>)",
		        OPAC,strerror(errno));
  }
(void) SSL_set_bio(tls->ssl,tls->bio,tls->bio);
while (done==false) {
  int sslerr;

  statut=SSL_accept(tls->ssl);
  switch (sslerr=SSL_get_error(tls->ssl,statut)) {
    case SSL_ERROR_NONE	     :
      statut=0;
      if ((peer=SSL_get_peer_certificate(tls->ssl))!=(X509 *)0) {
        if (SSL_get_verify_result(tls->ssl)!=X509_V_OK)
          statut=-1;
        (void) X509_free(peer);
        }
      done=true;
      break;
    case SSL_ERROR_WANT_READ :
      switch(tls_waitforchar(tls,(u_long)1000000,(const sigset_t *)0)) {
        case -1		     :
          switch(errno) {	/*received a signal, lets see...*/
	    case EINTR	     :	/*could be a TERM signal	*/
	      break;	
	    default	     :	/*hummm real code fault report	*/
	      (void) clm_core_dump("%s poll error '%s'",OPAC,strerror(errno));
              done=true;
	      statut=-1;
	      break;
	    }
	  break;
        case  0		     :	/*standard time out		*/
          tic++; 
          if (tic>MXTRY) {
            done=true;
            (void) clm_alert(0,"%s, SSL_accept too long to establish",OPAC);
	    statut=-1;
	    }
          break;
	default		     :
          break;
        }
      break;
#ifdef	SSL_ERROR_WANT_ACCEPT
    case SSL_ERROR_WANT_ACCEPT	:
#endif
    case SSL_ERROR_WANT_WRITE	:
      break;
    case SSL_ERROR_SYSCALL:	/*EOF received		*/
      if (ERR_get_error()==0)
	tls->goteof=true;		
      else
        (void) tlsshowerror(0,tls->ssl,statut,"%s SSL_SYSCALL pbs",OPAC);
      statut=-1;
      done=true;
      break;
    default				:
      (void) tlsshowerror(0,tls->ssl,statut,"%s SSL_accept fatal error "
					    "(sslerr='%d')",OPAC,sslerr);
      statut=-1;
      done=true;
      break; 
    }
  }
return statut;
}
/*

*/
/************************************************/
/*						*/
/*	Procedur to initiate a TLS connection	*/
/*	from the client side.			*/
/*	Return -1 if trouble, 0 otherwise	*/
/*						*/
/************************************************/
static int tlsconnect(TFILE *tls)

{
#define OPCO	"gestls.c:tlsconnect,"

int done;
int statut;
int tic;

done=false;
statut=-1;
tic=0;
if ((tls->bio=BIO_new_fd(tls->handle,BIO_NOCLOSE))==(BIO *)0) {
  (void) clm_core_dump("%s Unable to get the BIO (error=<%s>)",
		        OPCO,strerror(errno));
  }
(void) SSL_set_bio(tls->ssl,tls->bio,tls->bio);
while (done==false) {
  statut=SSL_connect(tls->ssl);
  switch (SSL_get_error(tls->ssl,statut)) {
    case SSL_ERROR_NONE	     :
      done=true;
      statut=0;
      break;
    case SSL_ERROR_WANT_READ :
      switch(tls_waitforchar(tls,(u_long)1000000,(const sigset_t *)0)) {
        case -1		     :
          switch(errno) {	/*received a signal, lets see...*/
	    case EINTR	     :	/*could be a TERM signal	*/
	      break;	
	    default	     :	/*hummm real code fault report	*/
	      (void) clm_core_dump("%s poll error '%s'",OPCO,strerror(errno));
              done=true;
	      statut=-1;
	      break;
	    }
	  break;
        case  0		     :	/*standard time out		*/
          tic++; 
          if (tic>MXTRY) {
            done=true;
            (void) clm_alert(0,"%s, SSL_connect too long to establish",OPCO);
	    statut=-1;
	    }
          break;
	default		     :
          break;
        }
      break;
    case SSL_ERROR_WANT_WRITE	:
    case SSL_ERROR_WANT_CONNECT	:
      break;
    default				:
      (void) tlsshowerror(0,tls->ssl,statut,"%s SSL_connect fatal error",OPCO);
      statut=-1;
      done=true;
      break;
    }
  }
return statut;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a TLS	*/
/*	structure.				*/
/*						*/
/************************************************/
static TFILE *tlsclean(TFILE *tls)

{
if (tls!=(TFILE *)0) {
  tls->traces=(char **)vec_freelstlst((void **)tls->traces,VECFREE);
  tls->cipherid=clm_freestr(tls->cipherid);
  tls->carin=clm_freestr(tls->carin);
  tls->port=clm_freestr(tls->port);
  tls->rmtname=clm_freestr(tls->rmtname);
  (void) free(tls);
  tls=(TFILE *)0;
  }
return tls;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to input character from the	*/
/*	socket.					*/
/*						*/
/************************************************/
static int tlsgetfeed(TLSFILE *tlsfile)

{
#define	OPFEED	"gestls.c:tlsgetfeed,"

TFILE *tls;
int limite;
char *ptr;
int cpt;

tls=(TFILE *)tlsfile;
limite=(MXBUF-1)-tls->incpt;
ptr=tls->carin+tls->incpt;
cpt=0;
errno=0;
switch (tls->tls) {
  case true	:
    cpt=SSL_read(tls->ssl,ptr,limite);
    break;
  default	:
    cpt=read(tls->handle,ptr,limite);
    break;
  }
switch (cpt) {
  case  0	:	/*no character available	*/
    switch (tls->tls) {
      case true			:
	{
	int big;

	switch ((big=SSL_get_error(tls->ssl,0))) {
	  case SSL_ERROR_SYSCALL:	/*EOF received		*/
    	    if (ERR_get_error()!=0)
	      (void) clm_alert(0,"%s wrong EOF",OPFEED);
	    else
              tls->goteof=true;		
	    break;
	  default		:
	    (void) clm_alert(0,"%s Unexpected SSL_read error='%d' Aborting channel",
				OPFEED,SSL_get_error(tls->ssl,0));
	    (void) usleep(100000);	/*100 millisec relax	*/
            tls->goteof=true;		
	    break;
	  }
	}
	break;
      default			:
        tls->goteof=true;		
        switch (errno) {
          case 0		:	/*EOF received		*/
	    break;
	  case EINTR		:	/*expected! error	*/
	  case EAGAIN		:
	  case ECONNRESET	:
	    break;
	  default		:
	    (void) clm_alert(0,"%s Unexpected read error=%d-><%s>",
				OPFEED,errno,strerror(errno));
	    (void) clm_alert(0,"%s limite='%d', incpt='%d'",
				OPFEED,limite,tls->incpt);
	    break;
	  }
        break;
      }
    break;
  case -1	:
    switch (tls->tls) {
      case true	:
	switch (SSL_get_error(tls->ssl,-1)) {
  	  case SSL_ERROR_WANT_READ	:
            (void) usleep(1000);	/*to ease up incoming charactere*/
	    break;
  	  case SSL_ERROR_SYSCALL	:
    	    if (ERR_get_error()==0)
	      tls->goteof=true;		
	    else
              (void) tlsshowerror(0,tls->ssl,cpt,"%s SSL_SYSCALL pbs",OPFEED);
	    break;
	  default	:
            (void) tlsshowerror(0,tls->ssl,cpt,"%s SSL_read problem",OPFEED);
	    break;
	  }
        break;
      default	:
        switch (errno) {
          case 0		:	/*success ???		*/
          case ESRCH		:	/*disconnect received	*/
	  case ECONNRESET	:	/*reset by peer		*/
	  case ETIMEDOUT	:	/*connection timeout	*/
	    break;
	  default		:
	    (void) clm_alert(0,"%s read error=%d-><%s>",
			 	OPFEED,errno,strerror(errno));
	    (void) clm_core_dump("read error");
	    tls->goteof=true;
	    break;
	  }
        break;
      }
    break;
  default	:
    tls->lastused=time((time_t *)0);
    tls->incpt+=cpt;
    /*removing wrong char '\000' received	*/
    while ((ptr=memchr(tls->carin,'\000',tls->incpt))!=(void *)0) {
      *ptr='?';	/*replacing '\000' as '?'	*/
      }
    break;
  }
return (tls->goteof==false);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format remote whois	*/
/*	infomartion				*/
/*						*/
/************************************************/
static char *getrmtwhois(char *demon,char *port,struct addrinfo *rp)

{
char *rmtwhois;
AFNTYP *afnnum;
char *stripnum;

rmtwhois=(char *)0;
afnnum=afn_getaddrinfo(rp);	
stripnum=afn_stripnum(afnnum);
(void) asprintf(&rmtwhois,"'%s' [%s|%s]",demon,stripnum,port);
(void) free(stripnum);
afnnum=afn_freeipnum(afnnum);
return rmtwhois;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to establish the remote	*/
/*	connection, return the handle		*/
/*						*/
/************************************************/
static int doconnect(char *demon,char *port,int swait,struct addrinfo *rp,int *socker)

{
#define	OPDC	"gestls.c:doconnect,"

int handle;
fd_set rset;
fd_set wset;
char *rmtwhois;
struct timeval timeout;
int flags;
int phase;
int proceed;

handle=-1;
FD_ZERO(&rset);
FD_ZERO(&wset);
rmtwhois=getrmtwhois(demon,port,rp);
timeout.tv_usec=0;;
timeout.tv_sec=swait;
flags=0;
*socker=0;
phase=0;
proceed=(rp!=(struct addrinfo *)0);
while (proceed==true) {
  switch (phase) {
    case    0	:	/*lets get a socket	*/
      if ((handle=socket(rp->ai_family,rp->ai_socktype,rp->ai_protocol))<0) 
	(void) clm_core_dump("%s Unable to open socket (error='%s')",
			      OPDC,strerror(errno));
      if ((flags=fcntl(handle,F_GETFL,0))<0) 
	(void) clm_core_dump("%s Unable to get socket descriptor (error='%s')",
			     OPDC,strerror(errno));
      if (fcntl(handle,F_SETFL,flags|O_NONBLOCK|O_ASYNC)<0)
	(void) clm_core_dump("%s Unable to set socket descriptor (error='%s')",
			      OPDC,strerror(errno));
      FD_SET(handle,&rset);	/*FD ready for select	*/
      wset=rset; 
      break;
    case 1	:	/*establishing link	*/
      if (connect(handle,rp->ai_addr,rp->ai_addrlen)<0) {
	if (errno != EINPROGRESS) {
	  (void) clm_alert(1,"%s %s unable to make connection (error=<%s>)",
	      		      OPDC,rmtwhois,strerror(errno));
	  phase=999;	/*trouble trouble	*/
	  }
	}
      break;
    case 2	:	/*wait link established	*/
      switch (select(handle+1,&rset,&wset,(fd_set *)0,&timeout)) {
	case -1	:		
	  (void) clm_alert(1,"%s %s trouble to establish connection (error=<%s>)",
			      OPDC,rmtwhois,strerror(errno));
	  phase=999;	/*trouble trouble	*/
	  break;
	case  0	:		
	  *socker=ETIMEDOUT;
	  (void) clm_alert(6,"%s %s Unable establish connection within %d sec, (error=<%s>)",
			      OPDC,rmtwhois,swait,strerror(*socker));
	  phase=999;	/*trouble trouble	*/
	  break;
	default	:
	  break;	/*everything fine	*/
	}
      break;
    case 3	:	/*link ready?		*/
      if ((*socker=checksockstat(handle))!=0) {
	(void) clm_alert(6,"%s %s Unable to establish socket link (error=<%s>)",
			    OPDC,rmtwhois,strerror(*socker));
	 phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*link ready, It is OK	*/
      (void) clm_alert(6,"%s %s Link established",OPDC,rmtwhois);
      proceed=false;
      break;
    default	:	/*SAFE Guard		*/
      (void) close(handle);
      handle=-1;
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(rmtwhois);
return handle;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the traceon flag within*/
/*	the tls structure.			*/
/*						*/
/************************************************/
void tls_trace(TLSFILE *tlsfile,int traceon)

{
TFILE *tls;

tls=(TFILE *)tlsfile;
tls->traceon=traceon;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extracte and RESET current	*/
/*	trace.					*/
/*						*/
/************************************************/
char **tls_gettraces(TLSFILE *tlsfile)

{
TFILE *tls;
char **traces;

tls=(TFILE *)tlsfile;
traces=tls->traces;
tls->traces=(char **)0;
return traces;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to trace on a debug file the	*/
/*	string reported.			*/
/*						*/
/************************************************/
void tls_traceout(TLSFILE *tlsfile,char *comment,char *str,int parnum)

{
if ((debug>3)&&((str!=(char *)0)&&(parnum>0))) {
  TFILE *tls;
  char *tracetime;
  char *strloc;

  tls=(TFILE *)tlsfile;
  tracetime=clm_ascsystime(time((time_t *)0));
  (void) asprintf(&strloc,"%s %s%.*s",tracetime,comment,parnum,str);
  (void) clm_cleanstring(strloc);
  if (tls->traceon==true) {
    (void) fprintf(stdout,"%s\n",strloc);
    (void) fflush(stdout);
    tls->traces=(char **)vec_addlstlst((void **)tls->traces,(void *)(strdup(strloc)));
    }
  (void) clm_alert(10,"tls_traceout <%s>",strloc);
  (void) free(strloc);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for characters	*/
/*	available on a TLS channel.		*/
/*	millisec is expressed in millisecond	*/
/*						*/
/************************************************/
int tls_waitforchar(TLSFILE *tlsfile,u_long microsec,const sigset_t *st)

{
int status;
TFILE *tls;
struct pollfd polling[1];
sigset_t origmask;
int timeout;

status=-1;
tls=(TFILE *)tlsfile;
polling[0].events=POLLIN|POLLPRI;
polling[0].revents=(short)0;
timeout=microsec/1000;	/*From microsec to millisec	*/
if ((tls==(TFILE *)0)||(tls->handle<0)) {
  errno=EBADF;
  return status;
  }
switch (tls->tls) {
  case true	:
    polling[0].fd=SSL_get_fd(tls->ssl);
    break;
  default	:
    polling[0].fd=tls->handle;
    break;
  }
(void) sigprocmask(SIG_SETMASK,st,&origmask);
status=poll(polling,1,timeout);
(void) sigprocmask(SIG_SETMASK,&origmask,(sigset_t *)0);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a line is		*/
/*	available within the TLS queue.		*/
/*						*/
/************************************************/
int tls_isready(TLSFILE *tlsfile)

{
int isready;
TFILE *tls;

isready=false;
tls=(TFILE *)tlsfile;
if (tls->incpt>0) {
  if (memchr(tls->carin,'\n',tls->incpt)!=(void *)0)
    isready=true;
  }
return isready;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for char on the line	*/
/*						*/
/************************************************/
int tls_checkfeed(TLSFILE *tls,u_long usec,const sigset_t *set)

{
#define	OPCH "gestls.c:tls_checkfeed,"

register int feedok;
register int phase;
register int proceed;

feedok=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*line already in buffer*/
      feedok=tls_isready(tls);
      break;
    case 1	:	/*wait for some action	*/
      switch(tls_waitforchar(tls,usec,set)) {
    	case -1	:
      	  switch(errno) {
	    case EINTR 	:	/*received a signal, lets see...*/
	      break;		/*could be a TERM signal	*/
	    default	:	/*hummm real code fault report	*/
	      (void) clm_core_dump("%s poll error '%s'",OPCH,strerror(errno));
	      break;
	    }
          break;
        case  0	:		/*standard time out		*/
          break;
        default	:		/*got some data on TLS		*/
      	  feedok=tlsgetfeed(tls);
          break;
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  if (feedok==true)
    break;
  phase++;
  }
return feedok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a full line, if the	*/
/*	can't be read in one shot memory is	*/
/*	extended and a new probe is done	*/
/*						*/
/************************************************/
char *tls_getbigline(TLSFILE *tls)

{
#define	 B64K	65536

char *current;
int gotit;

current=(char *)0;
gotit=false;
while (gotit==false) {
  char buffer[512];

  if (tls_fgets(buffer,sizeof(buffer),tls)==(char *)0) 
    break;
  current=clm_astrcat(current,buffer);
  if (strlen(buffer)<sizeof(buffer)-1)
    break;
  if (strlen(current)>B64K) {
    (void) clm_alert(0,"%s SMTP input line bigger than '%d'!",
		       "gestls.c:tls_getbigline,",B64K);
    break;
    }
  }
return current;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to write buffer on the proper	*/
/*	channel (either encrypted via TLS or	*/
/*	plain).					*/
/*	Return the number of char transmitted.	*/
/*						*/
/************************************************/
int tls_write(TLSFILE *tls,char *buffer,int parnum)

{
#define OPWTLS	"gestlst.c:tls_write,"

int gone;
int batched;
TFILE *tfile;
int proceed;

gone=0;
batched=0;
tfile=(TFILE *)tls;
proceed=true;
if (tfile->goteof==true)
  proceed=false;
if (debug>4) {
  char *str;

  str="-->";
  if (tfile->goteof==true)
    str="#eof ->";
  (void) tls_traceout(tls,str,buffer,parnum);
  }
while (proceed==true) {
  switch (tfile->tls) {
    case  true	:
      batched=SSL_write(tfile->ssl,buffer,parnum);
      break;
    default	:
      batched=write(tfile->handle,buffer,parnum);
      break;
    }
  proceed=false;
  switch (batched) {
    case  -1  :
      switch (tfile->tls) {
        case  true	:
          switch (SSL_get_error(tfile->ssl,-1)) {
	    case SSL_ERROR_WANT_READ		:
	    case SSL_ERROR_WANT_WRITE		:
	      (void) usleep(10000);
	      proceed=true;	/*lets proceed again			*/
	      break;
	    default			:
              (void) tlsshowerror(0,tfile->ssl,batched,"%s SSL_write problem",OPWTLS);
	      break;
	    }
          break;
        default		:
	  switch (errno) {
	    case EAGAIN 	:	/*NONBLOCK device not ready yet	*/
	      proceed=true;
					/*NO BREAK			*/
	    case ECONNRESET	:	/*remote just vanish		*/
	      (void) usleep(10000);
	      if (checksockstat(tfile->handle)!=0) {
		(void) clm_alert(1,"%s '%s' (port %d) connection trouble (error=<%s>)",
			           OPWTLS,tfile->rmtname,tfile->port,strerror(gone));
      		tfile->goteof=true;		
		proceed=false;
		gone=0;
		break;
		}
	      break;			/*lets get out with the error	*/
	    default		:
              (void) clm_alert(0,"%s Write problem, (error=<%s>)",
		                  OPWTLS,strerror(errno));
	      break;
	    }
          break;
        }
      break;
    case   0  :		/*no character written	*/
      tfile->goteof=true;		
      switch (tfile->tls) {
        case true			:
	  {
	  switch (SSL_get_error(tfile->ssl,0)) {
	    case SSL_ERROR_SYSCALL:	/*EOF received		*/
    	      if (ERR_get_error()!=0)
	        (void) clm_alert(0,"%s wrong EOF",OPFEED);
	      break;
	    default		:
	      (void) clm_alert(0,"%s Unexpected SSL_write error='%d'",
				  OPWTLS,SSL_get_error(tfile->ssl,0));
	      break;
	    }
	  (void) sleep(1);
	  }
	  break;
        default			:
          switch (errno) {
            case 0		:	/*EOF received		*/
	      break;
	    case EINTR		:	/*expected! error	*/
	    case EAGAIN		:
	    case ECONNRESET	:
	      break;
	    default		:
	      (void) clm_alert(0,"%s Unexpected write error=<%s>",
				  OPWTLS,strerror(errno));
	      break;
	    }
          break;
        }
      break;
    default	:
      if ((parnum-batched)>0) {
	parnum-=batched;
	(void) memmove(buffer,buffer+batched,parnum);
	proceed=true;
	}
      tfile->lastused=time((time_t *)0);
      gone+=batched;
      break;
    }
  }
return gone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait a connection from	*/
/*	remote and assign the structure.	*/
/*	This structure can be used to prepare	*/
/*	a crypted connexion (with TLS)		*/
/*						*/
/************************************************/
TLSFILE *tls_accept(LSTTYP *lst)

{
#define	OPCC	"gestls.c:tls_accept,"

TLSFILE *tls;
int handle;

tls=(TLSFILE *)0;
if ((handle=dev_acceptfile(lst))>=0) {
  TFILE *chantls; 
  int flags;

  chantls=(TFILE *)calloc(1,sizeof(TFILE));
  chantls->carin=calloc(1,MXBUF);
  chantls->handle=handle;
  if ((flags=fcntl(handle,F_GETFL,0))<0)
    (void) clm_core_dump("%s Unable to get socket descriptor (error='%s')",
		          OPCC,strerror(errno));
  if (fcntl(handle,F_SETFL,flags|O_NONBLOCK|O_ASYNC)<0)
    (void) clm_core_dump("%s Unable to set socket descriptor (error='%s')",
		          OPCC,strerror(errno));
  tls=(TLSFILE *)chantls;
  }
return tls;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to put the link in encrypted	*/
/*	mode.					*/
/*	accepting=true, waiting for a TLS	*/
/*	link.					*/
/*	accepting=false establish a TLS	link.	*/
/*	Return -1 if trouble, 0 otherwise	*/
/*						*/
/************************************************/
int tls_gotls(TLSFILE *tlsfile,int accepting)

{
#define	OPST	"gesttls.c:tls_gotls,"
#define TLSWAIT	120

int statut;
TFILE *tls;

statut=-1;
tls=(TFILE *)tlsfile;
if (tls->tls==true) {
  (void) clm_alert(2,"%s TLS link already crypted",OPST);
  }
else {
  (void) SSL_library_init();
  (void) SSL_load_error_strings();
  (void) ERR_clear_error();
  if ((tls->ctx=setup_tls_ctx())!=(SSL_CTX *)0) {
    if ((tls->ssl=SSL_new(tls->ctx))!=(SSL *)0) {
      switch (accepting)	{
        case false	:	/*client initiate connection	*/
          statut=tlsconnect(tls);
	  break;
        case true	:	/*server accept connection	*/
	default		:	/*no break			*/
          statut=tlsaccept(tls);
	  break;
        }
      if (statut<0) {
	if (tls->goteof==false)
          (void) clm_alert(0,"%s Unable to assign fd to SSL",OPST);
        (void) SSL_CTX_free(tls->ctx);
        tls->ctx=(SSL_CTX *)0;
        (void) ERR_free_strings();
	(void) SSL_free(tls->ssl);
	tls->ssl=(SSL *)0;
        }
      else {
	tls->cipherid=tlsgetcipherid(tls->ssl);
        tls->tls=true;
	}
      }
    }
  }
return statut;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the TLS link status	*/
/*						*/
/************************************************/
int tls_intls(TLSFILE *tls)

{
return ((TFILE *)tls)->tls;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the cipher infos	*/
/*	extracted from the TLS connection	*/
/*						*/
/************************************************/
char *tls_getcipher(TLSFILE *tlsfile)

{
char *cipherid;
TFILE *tls;

cipherid=(char *)0;
tls=(TFILE *)tlsfile;
if ((tls->tls==true)&&(tls->cipherid!=(char *)0)) {
  cipherid=strdup(tls->cipherid);
  }
return cipherid;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to terminate a TLS link.	*/
/*						*/
/************************************************/
int tls_close(TLSFILE *tlsfile)

{
#define 	APCO	"gestls.c:tls_close,"
TFILE *tls;
int status;

tls=(TFILE *)tlsfile;
if (tls->tls==true) {
  (void) SSL_set_shutdown(tls->ssl,SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
  switch((status=SSL_shutdown(tls->ssl))) {
    case -1	:
      (void) tlsshowerror(0,tls->ssl,status,"%s SSL_shutdown fatal error",APCO);
      break;
    case  0	:
      if ((status=SSL_shutdown(tls->ssl))!=1) 
        (void) tlsshowerror(0,tls->ssl,status,"%s still in HALF shutdown",APCO);
      break;
    case  1	:		/*good default				*/
      break;
    default	:
      (void) clm_core_dump("%s unexpected shutdown (statut='%d')",APCO,status);
      break;
    }
  (void) SSL_free(tls->ssl);
  (void) SSL_CTX_free(tls->ctx);
  tls->ctx=(SSL_CTX *)0;
  }
if ((status=shutdown(tls->handle,SHUT_RDWR))<0) {
  switch (errno) {
    case ENOTCONN	:	/*already disconnect by other side	*/
      break;
    default		:
      (void) clm_alert(0,"%s unable to shutdown (errno=<%s>)",
			  APCO,strerror(errno));
      break;
    }
  }
(void) close(tls->handle);
tls=tlsclean(tls);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to send characater on socket,	*/
/*	either in plain text or on TLS 		*/
/*	encryption.				*/
/*	This entry always add a '\n\r' to the	*/
/*	string sent on the main channel.	*/
/*						*/
/************************************************/
int tls_printf(TLSFILE *tlsfile,const char *fmt,...)

{
va_list args;
char *local;
int parnum;

va_start(args,fmt);
if ((parnum=vasprintf(&local,fmt,args))>0)
  (void) tls_write(tlsfile,local,parnum);
(void) free(local);
va_end(args);
return parnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to clear up any charactere 	*/
/*	remaining in tsl queue			*/
/*	in plain text or in TSL encryption	*/
/*						*/
/************************************************/
int tls_fpurge(TLSFILE *tlsfile)

{
int status;
TFILE *tls;

status=0;
(void) tls_checkfeed(tlsfile,1000,(const sigset_t *)0);
tls=(TFILE *)tlsfile;
if (tls->incpt>0) {
  tls->incpt=0;
  tls->carin[tls->incpt]='\000';
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to read from a socket either	*/
/*	in plain text or in TSL encryption	*/
/*						*/
/************************************************/
char *tls_fgets(char *s,int size,TLSFILE *tlsfile)

{
#define OPRTLS	"unitls.c:tls_fgets,"

TFILE *tls;
char *str;
int limite;

tls=(TFILE *)tlsfile;
str=(char *)0;
s[0]='\000';
if (tls->incpt>0) {
  char *stop;

  (void) memset(s,'\000',size);
  limite=size-1;
  s[limite]='\000';
  if (limite>tls->incpt)
    limite=tls->incpt;
  (void) memmove(s,tls->carin,limite);
  stop=strchr(s,'\n');
  if (stop==(char *)0) 
    stop=strchr(s,'\r');
  if ((stop=strchr(s,'\n'))!=(char *)0) {
    stop++;
    *stop='\000';
    }
  else {
    if (tls->incpt<size-1) {
      /*lets wait for full line	*/
      *s='\000';
      }
    }
  limite=strlen(s);
  if (limite>0) {
    tls->incpt-=limite;
    (void) memmove(tls->carin,tls->carin+limite,tls->incpt);
    str=s;	/*line is NOT empty	*/
    }
  tls->carin[tls->incpt]='\000';
  }
if (str!=(char *)0) 
  (void) tls_traceout(tlsfile,"<--",str,strlen(str));
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a name related to an	*/
/*	TLS connection.				*/
/*	Return a newly allocated memory which	*/
/*	contents is the needed information.	*/
/*						*/
/************************************************/
char *tls_getaddrname(TLSFILE *tlsfile,PEERTYP peer,char *unknown)

{
TFILE *tls;

tls=(TFILE *)tlsfile;
return soc_getaddrname(tls->handle,peer,unknown);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open a TCP connexion to a	*/
/*	daemon on a specific channel		*/
/*						*/
/************************************************/
TLSFILE *tls_opentcp(char *demon,char *port,int swait,int *socker)

{
#define	OPTC "gestsl.c:tls_opentcp,"

TFILE *tls;
int handle;
int status;
struct addrinfo hints;
struct addrinfo *res;
int phase;
int proceed;

tls=(TFILE *)0;
handle=-1;
status=0;
(void) memset(&hints, 0, sizeof(hints));
hints.ai_family=PF_UNSPEC;
hints.ai_socktype=SOCK_STREAM;
#ifdef AI_ADDRCONFIG
hints.ai_flags=AI_ADDRCONFIG;
#endif
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*get remote IP number	*/
      switch (status=getaddrinfo(demon,port,&hints,&res)) {
	case 0	:	/*everything ok		*/
	  break;
	default	:	/*Trouble		*/
          (void) clm_alert(6,"%s Unable to get '%s' host IP (error='%s')",
		     	      OPTC,demon,gai_strerror(status));
          proceed=false;/*trouble trouble	*/
	  break;
	}
      break;
    case 1	:	/*lets get socket	*/
      if (res!=(struct addrinfo *)0) {
	struct addrinfo *rp;

	for (rp=res;rp!=(struct addrinfo *)0;rp=rp->ai_next) {
	  if ((handle=doconnect(demon,port,swait,rp,socker))>=0) 
	    break;	/*channel open		*/
	  }
	(void) freeaddrinfo(res);
	}
      break;
    case 2	:	/*socket open?		*/
      if (handle<0) {
        (void) clm_alert(6,"%s Unable to make contact with '%s' host on port '%s'",
		     	      OPTC,demon,port);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*link established	*/
      tls=(TFILE *)calloc(1,sizeof(TFILE));
      tls->carin=calloc(1,MXBUF);
      tls->handle=handle;
      tls->rmtname=strdup(demon);
      tls->port=strdup(port);
      break;
    default	:	/*SAFE Guard + Error	*/
      proceed=false;
      break;
    }
  phase++;
  }
return (TLSFILE *)tls;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for data coming from	*/
/*	remote TLS client, same time we detecte*/
/*	if Parent process requested a quickout	*/
/*	or just vanished.			*/
/*	such case we return with a null pointer	*/
/*	Assign a dynamic memory to store input	*/
/*	SMTP line.				*/
/*						*/
/************************************************/
char *tls_waitforaction(TLSFILE *tls,u_int secwait)

{
#define	OPWA	"gestls.c:t,tls_waitforaction"

char *info;
u_long milliwait;
int phase;
int proceed;

info=(char *)0;
milliwait=secwait*100;	/*10 msec increment		*/
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have data ready		*/
      (void) tls_checkfeed(tls,1000,(const sigset_t *)0);
      break;
    case 1	:	/*all time used			*/
      if (milliwait==0)
	phase=999;	/*waiting time expired		*/
      milliwait--;
      break;
    case 2	:	/*lets wait 10 ms for input	*/
      (void) tls_checkfeed(tls,10000,(const sigset_t *)0);
      break;
    case 3	:	/*lets check if EOF		*/
      if (tls_iseof(tls)==true)
	phase=999;	/*trouble trouble		*/
      break;
    case 4	:	/*parent Lost?			*/
      if (prc_checkprocess(getppid())==false)
	phase=999;	/*trouble trouble		*/
      break;
    case 5	:	/*lets start again		*/
      phase=0;		/*on phase 1			*/
      break;
    default	:	/*SAFE guard			*/
      proceed=false;
      break;
    }
  phase++;
  if ((info=tls_getbigline(tls))!=(char *)0) {
    proceed=false;
    }
  }
return info;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report the tls EOF status	*/
/*						*/
/************************************************/
int tls_iseof(TLSFILE *tlsfile)

{
TFILE *tls;

tls=(TFILE *)tlsfile;
return (tls->goteof||sigterm);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report the number of second*/
/*	The TLS connection is IDLE.		*/
/*						*/
/************************************************/
u_long tls_idle(TLSFILE *tlsfile)

{
TFILE *tls;

tls=(TFILE *)tlsfile;
return (u_long)(time((time_t *)0)-tls->lastused);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a reply to a command	*/
/*	extract the final status and come back	*/
/*	with the whole reply.			*/
/*						*/
/************************************************/
int tls_getreply(TLSFILE *tls,u_int mxwait,PROTYP proto,char **reply)

{
#define	OPGE	"gestls.c:tls_getreply,"

int status;
int proceed;

status=ERRPROC;
proceed=true;
while (proceed==true) {
  int sofar;
  char *buffer;

  sofar=0;
  if ((buffer=tls_waitforaction(tls,mxwait))==(char *)0) {
    break;	/*time_out!			*/
    }
  proceed=false;
  switch (proto) {
    case pro_ment  :	/*protocol Mentor	*/
    case pro_smtp  :	/*protocol SMTP		*/
      if (sscanf(buffer,"%d%n",&status,&sofar)==1) {
        if (buffer[sofar]=='-')
          proceed=true;
        }
      break;
    case pro_pop3	:	/*protocol POP	*/
    case pro_pop3t 	:
    case pro_pop3s	:
      switch (buffer[0]) {
	case  '+'	:
	  status=CMDOK;		/*result A_OK	*/
	  break;
	case  '-'	:
	  status=REJECT;	/*result N_AOK	*/
	  break;
	default		: /*no meaningful answer*/
	  proceed=true;
	  break;
	}
      break;
    case pro_imap	:	/*protocol IMAP	*/
    case pro_imapt	:
    case pro_imaps	:
      status=CMDOK;	/*check status later on	*/
      break;
    case pro_ukn	:
    default		:
      (void) clm_alert(0,"%s, unknown protocol '%d' (bug?)",OPGE,(int)proto);
      break;
    }
  if (*reply!=(char *)0) 
    *reply=(char *)realloc((void *)*reply,strlen(*reply)+strlen(buffer)+1);
  else
    *reply=(char *)calloc(strlen(buffer)+1,sizeof(char));
  (void) strcat(*reply,buffer);
  buffer=clm_freestr(buffer);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for a protocol signon	*/
/*	Return 'true' if signon found		*/
/*						*/
/************************************************/
int tls_waitsignon(TLSFILE *tls,u_int mxwait,PROTYP proto,char **reply)

{
int status;
int retcode;

status=false;
if ((retcode=tls_getreply(tls,mxwait,proto,reply))!=ERRPROC) {
  switch (proto) {
    case pro_ment	:
    case pro_smtp	:
      if (retcode==SIGNON)
	status=true;
      break;
    case pro_pop3	:
    case pro_pop3t	:
    case pro_pop3s	:
      if (retcode==CMDOK)
	status=true;
      break;
    case pro_imap	:
    case pro_imapt	:
    case pro_imaps	:
      if ((retcode==CMDOK)&&(strstr(*reply,"* OK")==*reply))
	status=true;
      break;
    case pro_ukn	:	/*nothing to do	*/
    default	   	:
      break;
    }
  }
return status;
}
