/************************************************/
/*						*/
/*	GESTRU:					*/
/*	Take Care of all E-Mail neader analyse	*/
/*						*/
/************************************************/
%{
/*force le debugging du YACC                    */
#define lint    1
#ifdef	DEBUG
#define YYDEBUG 1
#endif

#include	<string.h>
#include	<stdio.h>
#include	<setjmp.h>
#include	<syslog.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<time.h>
#include	<ctype.h>

#include	"lowtyp.h"
#include	"utlgac.h"
#include	"utlvec.h"
#include	"bascnv.h"
#include	"subclm.h"
#include	"unieml.h"
#include	"gestru.h"

int yyparse();		/*syntax parser		*/
static int yylex();	/*lexical parser	*/

/*preamble static function declaration		*/
static void yyerror(char *);

static STRMTY **ystrms;	/*yacc stream info	*/
static int itsok;	/*parsing status	*/
static jmp_buf manystop;/*YACC fully lost	*/
static int cptstop=0;	/*stop received counting*/
static char *ykeyword;	/*ref to current kword	*/

static char **lines;	/*data-header pointer	*/
static int numline;	/*next char header line	*/
static int numrow;	/*next char header row	*/
static int curline;	/*cur char header line	*/
static u_int currow;	/*cur char header row	*/
static int nbrerr;	/*number of error so far*/
static int lastcode;	/*last code retreived	*/

#define	MXSTR	60	/*max line size in debug*/
#define STRIPV6	"ipv6"	/*Ipv6 ip marker	*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicate  DLGTYPstructure	*/
/*						*/
/************************************************/
static DLGTYP **dupdlgs(DLGTYP **dlgs)

{
DLGTYP **ndlgs;

ndlgs=(DLGTYP **)0;
if (dlgs!=(DLGTYP **)0) {
  DLGTYP **ptr;

  for (ptr=dlgs;(*ptr)!=(DLGTYP *)0;ptr++) {
    DLGTYP *ndlg;

    if ((ndlg=tru_dupdlg(*ptr))!=(DLGTYP *)0) 
      ndlgs=(DLGTYP **)vec_addlstlst((void **)ndlgs,(void *)ndlg);
    }
  }
return ndlgs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by DLGTYP	*/
/*	structure.				*/
/*						*/
/************************************************/
static DLGTYP **freedlgs(DLGTYP **dlgs)

{
if (dlgs!=(DLGTYP **)0) {
  int i;

  for (i=0;dlgs[i]!=(DLGTYP *)0;i++) 
    dlgs[i]=tru_freedlg(dlgs[i]);
  (void) free(dlgs);
  dlgs=(DLGTYP **)0;
  }
return dlgs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by streams*/
/*	structure.				*/
/*						*/
/************************************************/
static STRMTY **freestreams(STRMTY **strms)

{
if (strms!=(STRMTY **)0) {
  int i;

  for (i=0;strms[i]!=(STRMTY *)0;i++) 
    strms[i]=tru_freestrm(strms[i]);
  (void) free(strms);
  strms=(STRMTY **)0; 
  }
return strms;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the next character from*/
/*	the data-header				*/
/*						*/
/************************************************/
static char getnextchar()

{
int c;

c=EOF;
if (lines[numline]!=(char *)0) {
  curline=numline;
  currow=numrow;
  if ((numrow==0)&&(debug>0)) {
    char *addstr;

    addstr="";
    if (strlen(lines[numline])>MXSTR)
      addstr="...";
    (void) clm_alert(9,"%6d line=<%.*s%s>",numline+1,MXSTR,lines[numline],addstr);
    }
  c=(int)lines[numline][numrow];
  if (c=='\000') {
    c=' ';
    numrow=0;
    numline++;
    }
  else
    numrow++;
  }
return c;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to skip all blank character	*/
/*						*/
/************************************************/
static int skipblank()

{
int c;

c=EOF;
do {
  c=getnextchar();
  }
while (isblank(c)!=false);
return c;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make a ip/maks string	*/
/*						*/
/************************************************/
static STRMTY *mkipmask(char *ip,int mask,int limit)

{
STRMTY *strm;

strm=(STRMTY *)0;
if (mask>limit) {
  char *trbl;

  (void) asprintf(&trbl,"Mask value unexpected (above %d)",limit);
  (void) yyerror(trbl);
  (void) free(trbl);
  mask=limit;
  }
if (ip==(char *)0)
  (void) yyerror("IP number missing");
else {
  strm=(STRMTY *)calloc(1,sizeof(STRMTY)); 
  strm->mask=mask;
  strm->ip=ip;
  }
return strm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a stream component to	*/
/*	the streamlist list.			*/
/*						*/
/************************************************/
static STRMTY **addstream(STRMTY **streams,STRMTY *stream)

{
if (stream!=(STRMTY *)0) {
  streams=(STRMTY **)vec_addlstlst((void **)streams,(void *)stream);
  }
return streams;
}
/************************************************/
/************************************************/
%}

%union  {
	STRMTY **strms;	/*full streams parsed	*/
	STRMTY *strm;	/*one trust data stream	*/
	DLGTYP *dlg;	/*data stream info	*/
	DLGTYP **dlgs;	/*data stream infos	*/
	long time;	/*time value as a long	*/
	char *strtime;	/*time as a str(long)	*/
	char *string;	/*the info is a string	*/
	char punct;	/*one char only		*/
	long integer;	/*the value is integer	*/
	char *word;	/*the info is a word	*/
	};

%token	STOP	/*Header part end		*/
%token	CRLF	/*Carriage return		*/
%token	PLUS	/* '+' character		*/
%token	MINUS	/* '-' character		*/
%token	EQUAL	/* '=' character		*/
%token	COMMA	/* ',' character		*/
%token	SEMICOL	/* ';' character		*/
%token	SLASH	/* '/' character		*/
%token	COLON	/* ':' character		*/
%token	DOT	/* '.' character		*/
%token	PARLEFT	/* '(' character		*/
%token	PARRIGHT/* ')' character		*/
%token	CURLEFT	/* '{' character		*/
%token	CURIGHT	/* '{' character		*/
%token	BRKLEFT	/* '[' character		*/
%token	BRKRIGHT/* ']' character		*/
%token	PUNCT	/* any other non alphanum char	*/

%token	FROM	/* "From"			*/
%token	BAN	/* "Banned"			*/
%token	STD	/* "Standard"			*/
%token	STATUS	/* "status"			*/
%token	BY	/* "BY"				*/

/*Month name					*/
%token	JAN	/* "Jan"			*/
%token	FEB	/* "Feb"			*/
%token	MAR	/* "Mar"			*/
%token	APR	/* "Apr"			*/
%token	MAY	/* "May"			*/
%token	JUN	/* "Jun"			*/
%token	JUL	/* "Jul"			*/
%token	AUG	/* "Aug"			*/
%token	SEP	/* "Sep"			*/
%token	OCT	/* "Oct"			*/
%token	NOV	/* "Nov"			*/
%token	DEC	/* "Dec"			*/

/*Day name					*/
%token	MON	/* "Mon"			*/
%token	TUE	/* "Tue"			*/
%token	WED	/* "Wed"			*/
%token	THU	/* "Thu"			*/
%token	FRI	/* "Fri"			*/
%token	SAT	/* "Sat"			*/
%token	SUN	/* "Sun"			*/

%token	WORD	/*an alphanumeric word		*/
%token	INTEGER	/*a numeric word		*/
%token	HEXA	/*a hexadecimal word		*/
%token	STRING	/*litteral charactere		*/
%token	IPV6	/*ipv6 number marker		*/
/*
^L
*/
/************************************************/
/*	E-Mail header definition		*/
/************************************************/
/*%pure_parser*/
%start          dlgs

%%
dlgs		:	streams DOT STOP
			{
			ystrms=(STRMTY **)0;
                        if (itsok==true) {
			  ystrms=$<strms>1;
                          YYACCEPT;
			  }
                        else {
			  $<strms>1=freestreams($<strms>1);
                          YYABORT;
			  }
			}
		|	error DOT STOP
			{
			(void) yyerror("Fatal error stop");
			ystrms=(STRMTY **)0;
			YYABORT;
			}
		;

streams		:	stream
			{
			$<strms>$=addstream((STRMTY **)0,$<strm>1);
			}
		|	streams stream
			{
			$<strms>$=$<strms>1;
			if ($<strm>2!=(STRMTY *)0)
			  $<strms>$=addstream($<strms>1,$<strm>2);
			}
		|	streams error
			{
			(void) yyerror("Fatal streams error");
			$<strms>$=freestreams($<strms>1);
			itsok=false;
			}
		;

stream		:	network infos SEMICOL
			{
		        $<strm>$=$<strm>1;
			if ($<strm>1!=(STRMTY *)0) {
			  if ($<dlgs>2!=(DLGTYP **)0)
			    $<strm>1->data=$<dlgs>2;	
			  else
			    itsok=false;
			  }
			else {
			  if ($<dlgs>2!=(DLGTYP **)0)
			    $<dlgs>2=freedlgs($<dlgs>2);
			  itsok=false;
			  }
			}
		|	network error SEMICOL;
			{
			(void) yyerror("Fatal Error Parsing stream data");
			$<strm>$=tru_freestrm($<strm>1);
			}
		;

network		:	ipv4 SLASH INTEGER
			{
			$<strm>$=mkipmask($<string>1,$<integer>3,32);
			}
		|	ipv6 SLASH INTEGER
			{
			$<strm>$=mkipmask($<string>1,$<integer>3,128);
			}
		;


infos		:	info 
			{
			if ($<dlg>1!=(DLGTYP *)0) 
			  $<dlgs>$=(DLGTYP **)vec_addlstlst((void **)0,(void *)$<dlg>1);
			}
		|	infos COMMA info
			{
			if ($<dlg>1!=(DLGTYP *)0) 
			  $<dlgs>$=(DLGTYP **)vec_addlstlst((void **)$<dlgs>1,(void *)$<dlg>3);
			}
		;

info		:	data
			{
			if ($<dlg>1!=(DLGTYP *)0) 
			  (void) asprintf(&($<dlg>1->orig),"%s","");
			$<dlg>$=$<dlg>1;
			}
		|	from data
			{
			if ($<dlg>2!=(DLGTYP *)0) 
			  $<dlg>2->orig=$<string>1;
			$<dlg>$=$<dlg>2;
			}
		|	from datetime capinfo
			{
			$<dlg>$=(DLGTYP *)0;
			if ($<dlg>3!=(DLGTYP *)0) {
			  $<dlg>$=(DLGTYP *)calloc(1,sizeof(DLGTYP));
			  $<dlg>$->orig=$<string>1;
			  if ($<strtime>2!=(char *)0) {
			    $<dlg>$->flgt=atol($<strtime>2);
			    (void) free($<strtime>2);
			    }
			  $<dlg>$->flgdata=false;
			  $<dlg>$->frwdata.frw=$<dlgs>3;
			  }
			}
		;

capinfo		:	CURLEFT infos CURIGHT
			{
			$<dlg>$=$<dlg>2;
			}
		|	error CURIGHT
			{
			$<dlg>$=(DLGTYP *)0;
			}
		;

data		:	datetime STATUS status
			{
			$<dlg>$=(DLGTYP *)calloc(1,sizeof(DLGTYP));
			if ($<strtime>1!=(char *)0) {
			  $<dlg>$->flgt=atol($<strtime>1);
			  (void) free($<strtime>1);
			  }
			$<dlg>$->flgdata=true;
			$<dlg>$->frwdata.status=$<integer>3;
			}
		;

from		:	FROM origine
			{
			$<string>$=$<string>2;
			}
		;

ip		:	ipv4
			{
			$<string>$=$<string>1;
			}
		|	ipv6
			{
			$<string>$=$<string>1;
			}
		;

ipv4		:	byte DOT byte DOT byte DOT byte
			{
			(void) asprintf(&($<string>$),"%ld.%ld.%ld.%ld",
				$<integer>1,$<integer>3,$<integer>5,$<integer>7);
			}
		;

ipv6		:	numipv6
			{
			$<string>$=$<string>1;
			if ($<string>1!=(char *)0) {
			  (void) asprintf(&($<string>$),"ipv6:%s",$<string>1);
			  (void) free($<string>1);
			  }
			}
		|	IPV6 COLON numipv6
			{
			$<string>$=$<string>3;
			if ($<string>3!=(char *)0) {
			  (void) asprintf(&($<string>$),"%s:%s",STRIPV6,$<string>3);
			  (void) free($<string>3);
			  }
			}
		;

numipv6		:	pureipv6
			{
			$<string>$=$<string>1;
			}
		|	pureipv6 COLON COLON pureipv6
			{
			(void) asprintf(&($<string>$),"%s::%s",$<string>1,$<string>4);
			(void) free($<string>1);
			(void) free($<string>4);
			}
		|	pureipv6 COLON COLON ipv4
			{
			(void) asprintf(&($<string>$),"%s::%s",$<string>1,$<string>4);
			(void) free($<string>1);
			(void) free($<string>4);
			}
		|	pureipv6 COLON ipv4
			{
			(void) asprintf(&($<string>$),"%s::%s",$<string>1,$<string>3);
			(void) free($<string>1);
			(void) free($<string>3);
			}
		;

pureipv6	:	hexa 
			{
			$<string>$=$<string>1;
			}
		|	pureipv6 COLON hexa
			{
			(void) asprintf(&($<string>$),"%s:%s",$<string>1,$<string>3);
			(void) free($<string>1);
			(void) free($<string>3);
			}
		;

status		:	BAN
			{
			$<integer>$=BANBAN;
			}
		|	STD
			{
			$<integer>$=BANSTD;
			}
		|	error
			{
			(void) yyerror("Status can be one of ("STRBAN","STRSTD")");
			$<integer>$=BANSTD;
			}
		;

origine		:	domains SLASH BRKLEFT ip BRKRIGHT
			{
			(void) asprintf(&($<string>$),"%s/[%s]",$<string>1,$<string>4);
			(void) free($<string>1);
			(void) free($<string>4);
			}
		|	domains error
			{
			$<string>$=(char *)0;
			(void) free($<string>1);
			}
		|	error
			{
			$<string>$=(char *)0;
			}
		;

domains		:	word
			{
			$<string>$=$<string>1;
			}
		|	domains DOT word
			{
			(void) asprintf(&($<string>$),"%s.%s",$<string>1,$<string>3);
			(void) free($<string>1);
			(void) free($<string>3);
			}
		|	domains MINUS word
			{
			(void) asprintf(&($<string>$),"%s-%s",$<string>1,$<string>3);
			(void) free($<string>1);
			(void) free($<string>3);
			}
		;

datetime	:	days COMMA cachet
			{
			$<strtime>$=$<strtime>3;
			}
		|	cachet
			{
			$<strtime>$=$<strtime>1;
			}
		;

cachet		:	date time
			{
		        $<integer>1+=$<integer>2;
			(void) asprintf(&($<strtime>$),"%ld",$<integer>1);
			}
		|	date
			{
			(void) asprintf(&($<strtime>$),"%ld",$<integer>1);
			}
		;

date		:	INTEGER months INTEGER
			{
			struct tm timing;

			(void) memset(&timing,'\000',sizeof(timing));
			timing.tm_mday=$<integer>1;
			timing.tm_mon=$<integer>2;
	 		if ($<integer>3<100)
			  $<integer>3+=2000;
			timing.tm_year=$<integer>3-1900;
			$<time>$=timegm(&timing);
			}
		|	error 
			{
			$<time>$=time((time_t *)0);
			}
		;

months		:	JAN
			{
			$<integer>$=0;
			}
		|	FEB
			{
			$<integer>$=1;
			}
		|	MAR
			{
			$<integer>$=2;
			}
		|	APR
			{
			$<integer>$=3;
			}
		|	MAY	
			{
			$<integer>$=4;
			}
		|	JUN
			{
			$<integer>$=5;
			}
		|	JUL
			{
			$<integer>$=6;
			}
		|	AUG
			{
			$<integer>$=7;
			}
		|	SEP
			{
			$<integer>$=8;
			}
		|	OCT
			{
			$<integer>$=9;
			}
		|	NOV
			{
			$<integer>$=10;
			}
		|	DEC
			{
			$<integer>$=11;
			}
		|	error
			{
			$<integer>$=0;
			}
		;

time		:	hours zone
			{
			$<time>$=$<time>1-$<time>2;
			}
		|	hours
			{
			$<time>$=$<time>1;
			}
		|	zone
			{
			$<time>$=-$<time>1;
			}
		;

hours		:	hour COLON sixty COLON sixty
			{
			$<time>$=((($<time>1*60)+($<time>3))*60)+$<time>5;
			}
		|	hour COLON sixty
			{
			$<time>$=(($<time>1*60)+($<time>3))*60;
			}
		;

hour		:	INTEGER
			{
			$<time>$=$<integer>1;
			if ($<integer>1>24) {
			  (void) yyerror("Unexpected value (above 24 Hour)");
			  $<time>$=0;
			  } 
			}
		;

sixty		:	INTEGER
			{
			$<time>$=$<integer>1;
			if ($<integer>1>60) {
			  (void) yyerror("Unexpected value (above 60)");
			  $<time>$=0;
			  } 
			}
		;

zone		:	sign INTEGER optime
			{
			$<time>$=(($<integer>2/100)*3600)+(($<integer>2%100)*60);
			$<time>$*=$<integer>1;
			}
		|	WORD
			{
			static struct {
				char *zname;
				int diff;
				}tznom[]={
					  {"UTC",-0},
					  {"Z",-0},
					  {"GMT",-0},
					  {"EST",-5},
					  {"EDT",-4},
					  {"CST",-6},
					  {"CDT",-5},
					  {"MST",-7},
					  {"MDT",-6},
					  {"PST",-8},
					  {"PDT",-7},
					  {"AKST",-9},
					  {"AKDT",-8},
					  {"HST",-10},
					  {"HST",-10},
					  {"BST",+1},
					  {"CET",+1},
					  {"CST",+2},
					  {(char *)0,0}
					  };
			int i;

			$<time>$=0;
			for (i=0;tznom[i].zname!=(char *)0;i++) {
			  if (strcasecmp(tznom[i].zname,$<word>1)==0) {
			    $<time>$=tznom[i].diff*3600;
			    break;
			    }
			  }	
			(void) free($<word>1);
			}
		;

byte		:	INTEGER
			{
			$<integer>$=$<integer>1;
			if ($<integer>1>255) {
			  (void) yyerror("Unexpected value (above 255)");
			  $<integer>$=0;
			  } 
			}
		;

hexa		:	INTEGER
			{
			if ($<integer>1>65535) {
			  (void) yyerror("Unexpected value (above 'FFFF')");
			  $<integer>1=0;
			  } 
			$<integer>$=$<integer>1;
			(void) asprintf(&($<string>$),"%04ld",$<integer>1);
			}
		|	HEXA
			{
			if ($<integer>1>65535) {
			  (void) yyerror("Unexpected value (above 'FFFF')");
			  $<integer>1=0;
			  } 
			$<integer>$=$<integer>1;
			(void) asprintf(&($<string>$),"%04lx",$<integer>1);
			}
		;

days		:	SUN
			{
			$<integer>$=0;
			}
		|	MON
			{
			$<integer>$=1;
			}
		|	TUE
			{
			$<integer>$=2;
			}
		|	WED
			{
			$<integer>$=3;
			}
		|	THU
			{
			$<integer>$=4;
			}
		|	FRI
			{
			$<integer>$=5;
			}
		|	SAT
			{
			$<integer>$=6;
			}
		;

sign		:	MINUS
			{
			$<integer>$=-1;
			}
		|	PLUS
			{
			$<integer>$=+1;
			}
		|	PLUS MINUS
			{
			$<integer>$=+1;
			}
		;

optime		:	PARLEFT comments PARRIGHT
		|
		;


comments	:	comment
		|	comments comment
		;

comment		:	word
			{
			(void) free($<word>1);
			$<word>$=(char *)0;
			}
		;

word		:	INTEGER
			{
			(void) asprintf(&($<string>$),"%ld",$<integer>1);
			}
		|	HEXA
			{
			(void) asprintf(&($<string>$),"%lx",$<integer>1);
			}
		|	keyword
			{
			(void) asprintf(&($<string>$),"%s",ykeyword);
			}
		|	WORD
			{
			$<string>$=$<string>1;
			}
		;

keyword		:	STD
		|	STATUS
		|	BAN
		|	IPV6
		|	FROM
		|	MON
		|	TUE
		|	WED
		|	THU
		|	FRI
		|	SAT
		|	SUN
		|	JAN
		|	FEB
		|	MAR
		|	APR
		|	MAY
		|	JUN
		|	JUL
		|	AUG
		|	SEP
		|	OCT
		|	NOV
		|	DEC
		;
%%
#include	<memory.h>
#include	<ctype.h>
#include	<errno.h>
/*

*/
/************************************************/
/*						*/
/*	Procedure to really clean the YACC	*/
/*	satructure.				*/
/*						*/
/************************************************/
static void cleanyyparse()

{
#ifdef	YYINITSTACKSIZE
#if	YYPATCH >= 20100216
/*procedure already included byacc-1.9		*/
#else
/*to free yyparse dynamically allocated memory	*/
if (yyvs!=(YYSTYPE *)0) {
  (void) free(yyvs);
  yyvs=(YYSTYPE *)0;
  }
if (yyss!=(short *)0) {
  (void) free(yyss);
  yyss=(short *)0;
  }
yystacksize=0;
#endif
#endif
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report a parsing error	*/
/*	debugging mode only.			*/
/*						*/
/************************************************/
static void yyerror(char *info)

{
int dbg;

dbg=0;	/*0 Trusted/Banned streams parsing check*/
if ((clm_foreground==true)||(debug>=dbg)) {
  int i;
  char crlf[3];
  char lasttoken[50];

  (void) strcpy(crlf,"");
  (void) strcpy(lasttoken,"");
  if (strcmp(info,"syntax error")==0)
    (void) sprintf(lasttoken," (last token found='%d')",lastcode);
  for (i=0;lines[curline][i]!='\000';i++)
    if (isspace(lines[curline][i])!=false)
      lines[curline][i]=' ';
  if (currow>=strlen(lines[curline]))
    (void) strcpy(crlf," ");
  (void) clm_alert(dbg,"% 6d %.*s",curline+1,currow+2,lines[curline]);
  (void) clm_alert(dbg,"%*s %*s%s",6,"",currow+1,"^",info);
  }
nbrerr++;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the punct char	*/
/*	value.					*/ 	
/*						*/
/************************************************/
static int punctchar(char c)

{
int code;

code=PUNCT;
switch (c) {
  case ':'	:
    code=COLON;
    break;
  case ';'	:
    code=SEMICOL;
    break;
  case ','	:
    code=COMMA;
    break;
  case '.'	:
    code=DOT;
    break;
  case '+'	:
    code=PLUS;
    break;
  case '-'	:
    code=MINUS;
    break;
  case '='	:
    code=EQUAL;
    break;
  case '/'	:
    code=SLASH;
    break;
  case '['	:
    code=BRKLEFT;
    break;
  case ']'	:
    code=BRKRIGHT;
    break;
  case '('	:
    code=PARLEFT;
    break;
  case ')'	:
    code=PARRIGHT;
    break;
  case '{'	:
    code=CURLEFT;
    break;
  case '}'	:
    code=CURIGHT;
    break;
  default	:
    code=PUNCT;
    break;
  }
return code;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the token value 	*/
/*	from a string				*/
/*						*/
/************************************************/
static int tokenfromstr(char *strtoken,int *disp)

{
/*	Token list				*/
static struct	{
	int code;
	char *string;
	}vocable[]={/*MUST in length order	*/
		   {STD,	STRSTD},
		   {STATUS,	STRSTA},
		   {BAN,	STRBAN},
		   {FROM,	STRFRM},
		   {IPV6,	STRIPV6},
		   {MON,"Mon"},{TUE,"Tue"},{WED,"Wed"},{THU,"Thu"},
		   {FRI,"Fri"},{SAT,"Sat"},{SUN,"Sun"},
		   {JAN,"Jan"},{FEB,"Feb"},{MAR,"Mar"},{APR,"Apr"},
		   {MAY,"May"},{JUN,"Jun"},{JUL,"Jul"},{AUG,"Aug"},
		   {SEP,"Sep"},{OCT,"Oct"},{NOV,"Nov"},{DEC,"Dec"},
		   {WORD,(char *)0}
		   };

int token;
int i;

*disp=0;
token=WORD;
for (i=0;vocable[i].string!=(char *)0;i++) {
  *disp=strlen(vocable[i].string);
  if (strncasecmp(vocable[i].string,strtoken,*disp)==0) {
    ykeyword=vocable[i].string;
    token=vocable[i].code;
    break;
    }
  }
return token;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to backspace in the header	*/
/*	the data-header				*/
/*						*/
/************************************************/
static void backspace()

{
numline=curline;
numrow=currow;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract a string on a line	*/
/*						*/
/************************************************/
static char *getstring()

{
#define STRINC  20

char c;
char *strloc;
int locnumline;
int locnumrow;

locnumline=numline;
locnumrow=numrow;
strloc=(char *)calloc(1,STRINC+5);
while ((c=getnextchar())!=EOF) {
  int bspace;
  char ajout[2];

  bspace=false;
  if ((strlen(strloc)%STRINC)==0) {
    strloc=(char *)realloc(strloc,(((strlen(strloc)/STRINC)+1)*STRINC)+5);
    }
  bspace=((strlen(strloc)>1)&&(strloc[strlen(strloc)]=='\\'));
  if ((c=='\n')&&(bspace==false)) {
    (void) free(strloc);
    strloc=(char *)0;
    numline=locnumline;
    numrow=locnumrow;
    break;
    }
  if ((c=='\"')&&(bspace==false))
    break;
  ajout[0]=c;
  (void) strncat(strloc,ajout,1);
  }
return strloc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get/recognise a lexical	*/
/*	input.					*/
/*						*/
/************************************************/
static int yylex()

{
#define MXINFO	200

int found;
int c;
int proceed;
int phase;
int disp;
int numerique;
char info[MXINFO];

found=STOP;
phase=0;
proceed=true;
disp=0;
numerique=false;
(void) memset(info,'\000',sizeof(info));
c=skipblank();
if (c==EOF) {
  if (cptstop>0) {
    (void) siglongjmp(manystop,1);
    }
  cptstop++;
  return STOP;
  }
while (proceed==true) {
  switch (phase) {
    case     0	:	/*looking for strings	*/
      if (c=='\"') {
        if ((yylval.string=getstring())!=(char *)0) {
	  found=STRING;
	  proceed=false;
	  }
	}
      break;
    case     1	:	/* looking for punct	*/
      if (isalnum(c)==false) {
        found=punctchar(c);
	yylval.punct=(char)c;
	proceed=false;
	}
      break;
    case     2	:	/* looking for keyword	*/
      if ((found=tokenfromstr(lines[numline]+(numrow-1),&disp))!=WORD) {
	char *ptr;

	ptr=lines[numline]+(numrow-1)+disp;
	if (isalnum(*ptr)==false) {
	  numrow+=(disp-1);
	  proceed=false;
	  }
	}
      break;
    case     3	:	/*checking if word could be numeric	*/
      if (isdigit(c)!=false) 
	numerique=true;
      break;
    case     4	:	/*extracting word			*/
      {
      int extract;
      unsigned int inx;

      extract=true;
      inx=0;
      while (extract==true) {
        info[inx++]=c;
        if ((c=getnextchar())==EOF) {
          found=STOP;
	  if (cptstop>0)
	    (void) siglongjmp(manystop,1);
	  cptstop++;
	  extract=false;
	  break;
          }
        if (inx>=sizeof(info))
          extract=false;
        if (extract==false)
	  break;
        extract=(isalnum(c)!=false);
	}
      (void) backspace();
      }
      break;
    case     5	:	/* Is it a pure integer	*/
      if (numerique==true) {
	int i;

	for (i=1;info[i]!='\000';i++) {
	  if (isdigit(info[i])!=false)
	    continue;
	  numerique=false;
	  break;
	  }
	if (numerique==true) {
          yylval.integer=atol(info);
          found=INTEGER;
      	  proceed=false;
	  }
	}
      break;
    case     6	:	/*Is it an exhadecimal	*/
      if (strlen(info)>0) {
	int hexa;
	int i;

	hexa=true;
	for (i=0;info[i]!='\000';i++) {
	  if (isxdigit(info[i])!=false)
	    continue;
	  hexa=false;
	  break;
	  }
	if (hexa==true) {
	  yylval.integer=strtol(info,(char **)0,16);
          found=HEXA;
      	  proceed=false;
	  }
	}
      break;
    default	:	/*safe guard			*/
      yylval.word=strdup(info);
      found=WORD;
      proceed=false;
      break;
    }
  phase++;
  }
#ifdef DEBUG
if (yydebug>0) {
  /*synchronizing yydebug output	*/
  (void) fflush(stdout);
  (void) fflush(stderr);
  }
#endif
lastcode=found;
(void) clm_alert(11,"return token (value='%d')",found);
return found;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to parse an return a stream	*/
/*						*/
/************************************************/
static STRMTY **parsestream(char **trustlst)

{
STRMTY **parsed;
int verybad;

parsed=(STRMTY **)0;
verybad=false;
cptstop=0;
lines=trustlst;
numline=0;
numrow=0;
itsok=true;
nbrerr=0;
#ifdef DEBUG
if (debug>11)	/*debugging all yacc analyze	*/
  yydebug=1;
#endif
if (sigsetjmp(manystop,true)==0) 
  verybad=yyparse(); 
else 
  verybad=true;
parsed=ystrms;
ystrms=(STRMTY **)0;
(void) cleanyyparse();
if (verybad==true) 
  parsed=freestreams(parsed); 
return parsed;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to look at trusted data stream*/
/*	information and build sequence.		*/
/*						*/
/************************************************/
TRUTYP *tru_parsetrust(char *peername,char *peerip,char **trustlst)

{
#define	OPPA	"gestru.y:tru_parsetrust,"

TRUTYP *trust;
GACTYP *gat;
STRMTY **parsed;
int phase;
int proceed;

trust=(TRUTYP *)0;
gat=(GACTYP *)0;
parsed=(STRMTY **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*open Garbage Collector*/
      if ((gat=gac_open(0,appname,"parsing_banned_stream"))==(GACTYP *)0) {
	(void) clm_alert(0,"%s Unable to open garbage collector (errno=<%s> bug?)",
			   OPPA,strerror(errno));
	phase=999;	/*big trouble		*/
	}
      break;
    case 1	:	/*parsing under GC	*/
      if ((parsed=parsestream(trustlst))==(STRMTY **)0)
	phase++;	/*no need to duplicat	*/
      break;
    case 2	:	/*dup stream without GC	*/
      if (parsed!=(STRMTY **)0)	{/*always	*/
	long remaining;
	STRMTY **clone;
	int i;

	(void) gac_gear(gear_down,&remaining);
	clone=(STRMTY **)0;
  	for (i=0;parsed[i]!=(STRMTY *)0;i++) {
    	  STRMTY *loc;

    	  loc=tru_dupstrm(parsed[i]);
          clone=(STRMTY **)vec_addlstlst((void **)clone,(void *)loc);
	  }
  	trust=(TRUTYP *)calloc(1,sizeof(TRUTYP));
  	trust->peername=strdup(peername);
  	trust->peerip=strdup(peerip);
  	trust->strms=clone; 
	(void) gac_gear(gear_up,&remaining);
  	parsed=freestreams(parsed); 
    	}
      break;
    case 3	:	/*cleaning trash	*/
       (void) clm_trashmem(1,gat);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return trust;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free momory used by a	*/
/*	DLGTYP structure.			*/
/*	to a fresh identical structure.		*/
/*						*/
/************************************************/
DLGTYP *tru_freedlg(DLGTYP *dlg)

{
if (dlg!=(DLGTYP *)0) {
  if (dlg->flgdata==false)
    dlg->frwdata.frw=freedlgs(dlg->frwdata.frw);
  dlg->orig=clm_freestr(dlg->orig);
  (void) free(dlg);
  dlg=(DLGTYP *)0;
  }
return dlg;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicate an DLGTYP struct	*/
/*	to a fresh identical structure.		*/
/*						*/
/************************************************/
DLGTYP *tru_dupdlg(DLGTYP *dlg)

{
DLGTYP *ndlg;

ndlg=(DLGTYP *)0;
if (dlg!=(DLGTYP *)0) {
  ndlg=(DLGTYP *)calloc(1,sizeof(DLGTYP));
  ndlg->orig=strdup(dlg->orig);
  ndlg->flgt=dlg->flgt;
  ndlg->flgdata=dlg->flgdata;
  if (ndlg->flgdata==true)
    ndlg->frwdata.status=dlg->frwdata.status;
  else
    ndlg->frwdata.frw=dupdlgs(dlg->frwdata.frw);
  }
return ndlg;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locat within STRMTY struct	*/
/*	a BANNED information coming from the	*/
/*	same origin.				*/
/*						*/
/************************************************/
DLGTYP **tru_locdlg(STRMTY *strm,DLGTYP *dlg)

{
DLGTYP **spoted;

spoted=(DLGTYP **)0;
if ((strm!=(STRMTY *)0)&&(strm->data!=(DLGTYP **)0)) {
  DLGTYP **data;

  for (data=strm->data;*data!=(DLGTYP *)0;data++) {
    if (strcmp((*data)->orig,dlg->orig)==0) {
      spoted=data;
      break;
      }
    }
  }
return spoted;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump data within DLGTYP	*/
/*	structure.				*/
/*						*/
/************************************************/
char **tru_dumpdlgs(DLGTYP **dlgs,int tab)

{
static char *strcode[]={"UKN",STRSTD,STRBAN};

char **strdlgs;

strdlgs=(char **)0;
if (dlgs!=(DLGTYP **)0) {
  int i;

  for (i=0;dlgs[i]!=(DLGTYP *)0;i++) {
    char *str;
    char **strs;
    char sdate[100];

    str=(char *)0;
    strs=(char **)0;
    (void) strcpy(sdate,clm_ascsysstamp(dlgs[i]->flgt));
    switch (dlgs[i]->flgdata) {
      case false	:
	if ((strs=tru_dumpdlgs(dlgs[i]->frwdata.frw,tab+2))!=(char **)0) {
	  int j;

	  (void) asprintf(&str,"%*s%s %s %s",tab,"",STRFRM,dlgs[i]->orig,sdate);
          strdlgs=(char **)vec_addlstlst((void **)strdlgs,(void *)str);
	  (void) asprintf(&str,"%*s{",tab+2,"");
          strdlgs=(char **)vec_addlstlst((void **)strdlgs,(void *)str);
          for (j=0;strs[j]!=(char *)0;j++) {
            strdlgs=(char **)vec_addlstlst((void **)strdlgs,(void *)strs[j]);
	    str=strs[j];
	    }
          (void) free(strs);
          str[strlen(str)-1]='\000';
	  (void) asprintf(&str,"%*s},",tab+2,"");
          strdlgs=(char **)vec_addlstlst((void **)strdlgs,(void *)str);
	  }
	break;
      default		:
 	if (dlgs[i]->frwdata.status>BANBAN)
	  dlgs[i]->frwdata.status=BANUKN;	
	if (strlen(dlgs[i]->orig)>0)
	  (void) asprintf(&str,"%*s%s %s %s %s %s,",
			        tab,"",STRFRM,dlgs[i]->orig,sdate,
				STRSTA,strcode[dlgs[i]->frwdata.status]);
	else
	  (void) asprintf(&str,"%*s%s %s %s,",
			        tab,"",sdate,STRSTA,strcode[dlgs[i]->frwdata.status]);
        strdlgs=(char **)vec_addlstlst((void **)strdlgs,str);
	break;
      }
    if ((dlgs[i+1]==(DLGTYP *)0)&&(str!=(char *)0)&&(strlen(str)>0))
      str[strlen(str)-1]=';';
    }
  }
return strdlgs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicate an STRMTY struct	*/
/*	to a fresh identical structure.		*/
/*						*/
/************************************************/
STRMTY *tru_dupstrm(STRMTY *strm)

{
STRMTY *nstrm;

nstrm=(STRMTY *)0;
if (strm!=(STRMTY *)0) {
  nstrm=(STRMTY *)calloc(1,sizeof(STRMTY));
  nstrm->ip=strdup(strm->ip);
  nstrm->mask=strm->mask;
  nstrm->data=dupdlgs(strm->data);
  }
return nstrm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free the memory used by	*/
/*	a STRMTY * structure			*/
/*						*/
/************************************************/
STRMTY *tru_freestrm(STRMTY *strm)

{
if (strm!=(STRMTY *)0) {
  strm->data=freedlgs(strm->data);
  strm->ip=clm_freestr(strm->ip);
  (void) free(strm);
  strm=(STRMTY *)0;
  }
return strm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free the memory used by	*/
/*	a TRUTYP ** structure			*/
/*						*/
/************************************************/
TRUTYP *tru_freetrusts(TRUTYP *dattrusts)

{
if (dattrusts!=(TRUTYP *)0) {
  dattrusts->strms=freestreams(dattrusts->strms);
  dattrusts->peerip=clm_freestr(dattrusts->peerip);
  dattrusts->peername=clm_freestr(dattrusts->peername);
  (void) free(dattrusts);
  dattrusts=(TRUTYP *)0; 
  }
return dattrusts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do an ASCII dump of 	*/
/*	trust data.				*/
/*						*/
/************************************************/
char **tru_dumpstrm(char *myname,char *myip,STRMTY *strm)

{
char **strstrms;

strstrms=(char **)0;
if (strm!=(STRMTY *)0) {
  char *str;
  char **dlgs;

  (void) asprintf(&str,"#-Received by %s/[%s]",myname,myip);
  strstrms=(char **)vec_addlstlst((void **)strstrms,(void *)str);
  (void) asprintf(&str,"%s/%d",strm->ip,strm->mask);
  strstrms=(char **)vec_addlstlst((void **)strstrms,(void *)str);
  if ((dlgs=tru_dumpdlgs(strm->data,4))!=(char **)0) {
    int j;
    
    for (j=0;dlgs[j]!=(char *)0;j++) 
      strstrms=(char **)vec_addlstlst((void **)strstrms,(void *)dlgs[j]);
    (void) free(dlgs);
    }
  }
return strstrms;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to preapre ASCII conversion	*/
/*	of the stream, adding the origin.	*/
/*						*/
/************************************************/
void tru_setorigin(TRUTYP *dattrusts)

{
if (dattrusts!=(TRUTYP *)0) {
  char *origin;

  (void) asprintf(&origin,"%s/[%s]",dattrusts->peername,dattrusts->peerip);
  if (dattrusts->strms!=(STRMTY **)0) {
    STRMTY **ptr;

    for (ptr=dattrusts->strms;*ptr!=(STRMTY *)0;ptr++) {
      if ((*ptr)->data!=(DLGTYP **)0) {
	DLGTYP **dlgs;
	int count;

	dlgs=(*ptr)->data;
	count=vec_sizelstlst((void **)dlgs);
	if ((count==1)&&(strlen((*dlgs)->orig)==0)) {
	  (void) free((*dlgs)->orig);
	  (*dlgs)->orig=strdup(origin);
	  }
	else {
	  DLGTYP *dlg;

	  dlg=(DLGTYP *)calloc(1,sizeof(DLGTYP));
	  dlg->orig=strdup(origin);
	  dlg->flgt=time((time_t *)0);
	  dlg->flgdata=false;
	  dlg->frwdata.frw=dlgs;
          (*ptr)->data=(DLGTYP **)vec_addlstlst((void **)0,(void *)dlg);
	  }
	}
      }
    }
  (void) free(origin);
  }
}
