/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLCKR:					*/
/*	Tool to handle user status detection	*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<string.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"basafn.h"
#include	"subclm.h"
#include	"unidns.h"
#include	"unilda.h"
#include	"unijrl.h"
#include	"geseml.h"
#include	"comeml.h"
#include	"lvlckr.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if found MX could	*/
/*	be loopback to clement (MX looping	*/
/*	detection).				*/
/*						*/
/************************************************/
static int itsaloopback(SQLFILE *sql,char *domain)

{
register int loopback;
char **iplst;

loopback=false;
if ((iplst=dns_getfirstmxiplst(domain))!=(char **)0) {
  BININF **binlst;

  binlst=dbc_getbinds(sql,(BININF **)0,pro_smtp);
  binlst=dbc_getbinds(sql,binlst,pro_smtps);
  if (binlst!=(BININF **)0) {
    int i;

    for (i=0;(loopback==false)&&(iplst[i]!=(char *)0);i++) {
      int j;

      for (j=0;binlst[j]!=(BININF *)0;j++) {
	if (strcmp(binlst[j]->ip,iplst[i])==0) {
	  loopback=true;
	  break;
	  }
	}
      }
    binlst=dev_freebindinf(binlst);
    }
  iplst=(char **)vec_freelstlst((void **)iplst,VECFREE);
  }
return loopback;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to retrieve ldap user info	*/
/*						*/
/************************************************/
static RCPTFLG getuserldapinfo(int ssl,GRPTYP *grp,USRTYP *usr)

{
RCPTFLG rcptflg;
char *official;
int phase;
int proceed;

rcptflg=rcpt_unknown;
official=strdup(usr->username);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*checking is user exist*/
      if (lda_emailknown(ssl,grp->authrmt,grp->authport,grp->ldapbase,&official)==false)
	phase=999;	/*user unknown		*/
      break;
    case 1	:	/*is user to forward	*/
      rcptflg=rcpt_islocal;
      if (strlen(official)>0) {	/*always	*/
	char **lists;

        if ((lists=lda_emailaliases(ssl,grp->authrmt,grp->authport,grp->ldapbase,official))!=(char **)0) {
	  rcptflg=rcpt_islocalias;
	  usr->numalias=vec_sizelstlst((void **)lists);
	  lists=(char **)vec_freelstlst((void **)lists,VECFREE);
	  }
	}
      break;
    case 2	:	/*is site remote too	*/
      if ((grp->smtprmt!=(char *)0)&&(grp->smtpport!=0)) {
	switch (rcptflg) {
	  case rcpt_islocal	:
            rcptflg=rcpt_forward;
	    break;
	  case rcpt_islocalias	:
            rcptflg=rcpt_isalias;
	    break;
	  default		:	/*never	*/
	    break;
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(official);
return rcptflg;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if override record	*/
/*	do match and grab the current status.	*/
/*						*/
/************************************************/
static OVRENUM checkover(char *from,OVRTYP *ovr,AFNTYP *afnnum)

{
#define	TRBCHK	"Warning!, Overrides record (ovrid='%lu') orgip=<%s> table contents is ill-formated"

register OVRENUM decide;
register AFNTYP *afnorgip;
	 int cidr;
register int phase;
register int proceed;

decide=tr_maybe;
afnorgip=(AFNTYP *)0;
cidr=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is domain the same	*/
      if (strcmp(from,ovr->orgemail)!=0)
	phase=999;	/*org do not match	*/
      break;
    case 1	:	/*Is IP acceptable	*/
      if (ovr->orgip!=(char *)0) { /*always true*/
        char orgip[100];

	if (sscanf(ovr->orgip,"%99[^/]/%d",orgip,&cidr)==2) {
          if ((afnorgip=afn_getipnum(orgip))==(AFNTYP *)0) {
            (void) jrl_journalling(false,EVCKR,TRBCHK,ovr->ovrid,ovr->orgip);
	    phase=999;	/*trouble touble	*/
	    }
	  }
        }
      break;
    case 2	:	/*comparing IP number	*/
      if (afn_cmpipnum(afnorgip,afnnum,cidr)==true)
        decide=ovr->ovrride;
      afnorgip=afn_freeipnum(afnorgip);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return decide;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a specific remote	*/
/*	is ready to accept email for a specific	*/
/*	domain.					*/
/*						*/
/************************************************/
static int locatedomain(GRPTYP *grp,char *domain)

{
/*remote must answer within 15 second		*/
#define LOCOUT	15	
#define	NORMT	"We are unable to contact remote forwarder <%s/%d>"
#define	NORMFRM	"From not accepted by remote forwarder <%s/%d>"
#define	RMTRCPT	"%s: <%s@%s>"

int accepted;
CTTYP *contact;
int status;
int socker;
int phase;
int proceed;

accepted=false;
contact=(CTTYP *)0;
status=ERRPROC;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*contact with remote	*/
      contact=eml_opencontact(pro_smtp,grp->smtprmt,grp->smtpport,LOCOUT,&socker);
      if ((contact==(CTTYP *)0)||(contact->established==false)) {
        (void) jrl_journalling(false,EVCKR,NORMT,grp->smtprmt,grp->smtpport);
	proceed=false;
	}
      break;
    case 1	:	/*make from		*/
      if (eml_sendmfrom(contact,"",100)!=CMDOK) {
        (void) jrl_journalling(false,EVCKR,NORMFRM,grp->smtprmt,grp->smtpport);
	phase=999;
	}
      break;
    case 2	:	/*cheking user		*/
      status=eml_sendcmd(contact,CMDOK,RMTRCPT,RCPT,POSTM,domain);
      switch (status) {
	case CMDOK	:
	  accepted=true;
	  break;
	default		: 	
	  if ((status>=TMPERR)&&(status<SYNERR))
	    accepted=true;
	  break;
	}
      break;
    default	:	/*SAFE Guard		*/
      contact=eml_closecontact(contact);
      proceed=false;
      break;
    }
  phase++;
  }
return accepted;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to discover new domain to be	*/
/*	forwarded when clement is MX but the	*/
/*	domain is not yet known localy		*/
/*						*/
/************************************************/
static int findoutdomain(SQLFILE *sql,char *domain)

{
int done;
GRPTYP **grplst;
int phase;
int proceed;

done=false;
grplst=(GRPTYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting group list	*/
      if ((grplst=dbc_getgrouplst(sql))==(GRPTYP **)0) 
	phase=999;	/*no group ?		*/
      break;
    case 1	:	/*scanning group list	*/
      if (grplst!=(GRPTYP **)0) {	/*always*/
	int i;

	for (i=0;grplst[i]!=(GRPTYP *)0;i++) {
	  if ((grplst[i]->smtprmt==(char *)0)||(strlen(grplst[i]->smtprmt)==0))
	    continue;
	  if (locatedomain(grplst[i],domain)==true) {
	    (void) dbc_adddomtogroup(sql,grplst[i],domain);
	    done=true;
	    break;
	    }
	  }
        grplst=dbc_freegrouplst(grplst);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to detect if remote user could*/
/*	handle by a remote SMTP server		*/
/*	originator is acceptable by recipient	*/
/*						*/
/************************************************/
static void istoforward(SQLFILE *sql,GRPTYP *group,USRTYP *usr)

{
#define	OPIS	"lvlckr.c:istoforward,"
#define	NOCTA	"We are unable to contact forwarder <%s/%d>"
#define	NOFRM	"From not accepted by forwarder <%s/%d>"
#define	TOTRBL	"Forwarder <%s:%d> replied to:\n\t  %s: <%s>, with:\n\t  %s"

#define	FRWTOUT	15	/*forwarder should answer within 15 sec	*/

CTTYP *contact;
int status;
int socker;
int phase;
int proceed;

contact=(CTTYP *)0;
status=ERRPROC;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*check within DB first	*/
      usr->rcptflg=dbc_getintusr(sql,usr->username);
      if (usr->rcptflg!=rcpt_uknstatus)
	phase=999;	/*found one user	*/
      break;
    case 1	:	/*contact with forwarder*/
      /*lets assum forwarder won't answer	*/
      usr->rcptflg=rcpt_notchecked;
      contact=eml_opencontact(pro_smtp,group->smtprmt,group->smtpport,FRWTOUT,&socker);
      if ((contact==(CTTYP *)0)||(contact->established==false)) {
        (void) jrl_journalling(false,EVCKR,NOCTA,group->smtprmt,group->smtpport);
	phase=999;
	}
      break;
    case 2	:	/*sending from		*/
      if (contact->established==true) {	/*always*/
	char *dfltfrom;

	dfltfrom=dbc_getdfltfrom(sql,eml_getdomain(usr->username));
      	if (eml_sendmfrom(contact,dfltfrom,100)!=CMDOK) {
          (void) jrl_journalling(false,EVCKR,NOFRM,group->smtprmt,group->smtpport);
          contact=eml_closecontact(contact);
	  phase=999;
	  }
	(void) free(dfltfrom);
	}	
      break;
    case 3	:	/*cheking user		*/
      status=eml_sendcmd(contact,CMDOK,FMSM,RCPT,usr->username);
      if ((status>=CMDOK)&&(status<=MAILPAL))
	status=CMDOK;
      if ((status>=TMPERR)&&(status<SYNERR))
	status=TMPERR;
      if (status>=SYNERR)
	status=REJECT;
      switch (status) {
	case CMDOK	:
          usr->rcptflg=rcpt_forward;
	  break;
	case TMPERR	:
          usr->rcptflg=rcpt_notchecked;
	  break;
	case REJECT	:
	default		: 	
          usr->rcptflg=rcpt_unknown;
	  break;
	}
      if ((status!=CMDOK)&&(contact->info!=(char *)0)) {
	 char *forlogs;

	 forlogs=eml_forlogs(contact->info,(char *)0);
         (void) jrl_journalling(false,EVCKR,TOTRBL,
				      group->smtprmt,group->smtpport,
				      RCPT,usr->username,forlogs);
	 forlogs=clm_freestr(forlogs);
	 }
      break;
    case 4	:	/*inserting user/status	*/
      (void) dbc_setintusr(sql,usr->username,usr->rcptflg);
      contact=eml_closecontact(contact);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to to replace if needed	*/
/*	current domain by alias domain.		*/
/*	if an alias is found and replace current*/
/*	domain return 1.			*/
/*						*/
/************************************************/
int ckr_checkdomover(SQLFILE *sql,char **rcpt,int eventjrl)

{
int done;
char *ptr;

done=false;
if ((ptr=strchr(*rcpt,'@'))!=(char *)0) {
  char *newdom;

  if ((newdom=dbc_isitadomainalias(sql,ptr+1))!=(char *)0) {
    char *old;

    old=strdup(*rcpt);
    *ptr='\000';
    ptr=*rcpt;
    (void) asprintf(rcpt,"%s@%s",ptr,newdom);
    (void) jrl_journalling(false,eventjrl,"Domain override from <%s> to <%s>",old,*rcpt);
    (void) free(old);
    (void) free(newdom);
    (void) free(ptr);	/*removing old username	*/
    done=true;
    }
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if wee need to	*/
/*	override domain part for user data	*/
/*						*/
/************************************************/
void ckr_domoverride(SQLFILE *sql,USRTYP *usr)

{
char *ptr;
char *newdom;
int phase;
int proceed;

ptr=(char *)0;
newdom=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting current dom*/
      if ((ptr=strchr(usr->username,'@'))!=(char *)0) {
	newdom=dbc_isitadomainalias(sql,ptr+1);
	if (newdom!=(char *)0)
	  phase++;	/*no need to check	*/
	}		/*localhost.localdomain	*/
      break;
    case 1	:	/*is it localhost?	*/
      break;
    case 2	:	/*do we have an override*/
      if (newdom==(char *)0)
	phase=999;	/*no need to go further	*/
      break;
    case 3	:	/*replacing with new dom*/
      *ptr='\000';	/*take user name only	*/
      ptr=usr->username;
      usr->domover=true;
      (void) asprintf(&(usr->username),"%s@%s",ptr,newdom);
      (void) free(newdom);
      (void) free(ptr);	/*removing old username	*/
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a user id can be	*/
/*	confirmed existing and extract the	*/
/*	current status.				*/
/*	server).				*/
/*						*/
/************************************************/
void ckr_isuserknown(SQLFILE *sql,USRTYP *usr)

{
GRPTYP *group;
int phase;
int proceed;

group=(GRPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we know domain?	*/
      if ((group=dbc_getgroup(sql,usr->username))==(GRPTYP *)0) 
	proceed=false;	/*No, must be foreign	*/
      break;
    case 1	:	/*is group expired	*/
      if ((group->active==false)||((group->upto>0)&&(group->upto<time((time_t *)0)))) {
	usr->rcptflg=rcpt_inactive;
	phase=999;	/*account is not active	*/
	}
      break;
    case 2	:	/*is user a local one	*/
      if ((group->smtprmt==(char *)0)||(strlen(group->smtprmt)==0)) {
        int ssl;
	RCPTFLG ldapflg;

        (void) dbc_getuserinfo(sql,usr);
	ldapflg=usr->rcptflg;
        ssl=false;
	switch (group->authprot) {
	  case pro_ldaps:	/*asking to ldap*/
	    ssl=true;		/*NO BREAK	*/
	  case pro_ldap	:
	    ldapflg=getuserldapinfo(ssl,group,usr);
	    break;
	  default    	:	/*domain is local*/
	    break;
	  }
        if (ldapflg!=usr->rcptflg) {
          usr->rcptflg=ldapflg; 
	  usr->active=false;
	  switch (ldapflg) {
	    case rcpt_islocal	:
	      usr->active=true;
				/*NO BREAK	*/
	    case rcpt_unknown	:
	      (void) dbc_updateuser(sql,usr);
	      break;
	    default		:
	      break;		/*nothing to do	*/
            }
          }
	}
      else {
	/*domain on remote server		*/
        (void) istoforward(sql,group,usr);
	}
      break;
    default	:	/*SAFE guard		*/
      group=dbc_freegroup(group);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return all meaningful data	*/
/*	related to one user.			*/
/*						*/
/************************************************/
USRTYP *ckr_getuserinfo(SQLFILE *sql,char *rcpt,char *locip)

{
#define OPKN	"lvlckr.c:ckr_ckr_getuserinfo,"

#define	ISTRB	\
	"Receiving Host (IP=%s) is declared MX for domain <%s> but domain is unknown to me"


USRTYP *usr;
char **iplist;
int phase;
int proceed;

usr=dbc_mkuserinfo(rcpt,rcpt_uknstatus);
iplist=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*is domain an alias	*/
      (void) ckr_domoverride(sql,usr);
      if (usr->rcptflg!=rcpt_uknstatus)
	phase=999;	/*found a status	*/
      break;
    case 1	:	/*is domain a known one	*/
      switch (dbc_getactive(sql,eml_getdomain(usr->username))) {
	case da_good	:	/*active domain	*/
          (void) ckr_isuserknown(sql,usr);
	  phase=999;	/*no need to go further	*/
	  break;
	case da_resend	:	/*to resend 	*/
          usr->rcptflg=rcpt_resend;
	  phase=999;	/*no need to go further	*/
	  break;
	case da_lock	:	/*not active 	*/
          usr->rcptflg=rcpt_inactive;
	  phase=999;	/*no need to go further	*/
	  break;
	case da_unknown	:	/*MX?		*/
	  break;	/*new domain??		*/
	}
      break;
    case 2	:	/*is user remote	*/
      switch (dns_isgoodmx(eml_getdomain(usr->username),locip)) {
	case dns_ourmx 	:/*could be ours ????	*/
	case dns_bestmx	:/*must be our domain 	*/
	  if (findoutdomain(sql,eml_getdomain(usr->username))==true)
            (void) ckr_isuserknown(sql,usr);
          else {
	    (void) clm_alert(0,ISTRB,locip,eml_getdomain(usr->username));
	    /*not ready to take care of domain	*/
	    usr->rcptflg=rcpt_foreign;
	    }
	  break;
	case dns_extmx	:/*mail to foreign MX	*/
	  usr->rcptflg=rcpt_foreign;
	  if (itsaloopback(sql,eml_getdomain(usr->username))==true) {
	    usr->rcptflg=rcpt_loopback;
	    break;	/*No MX Check		*/
	    }
	  if (dns_gotmxorip(usr->username)==false) 
	    usr->rcptflg=rcpt_nomx;
	  break;
	case dns_nomx	:/*No official MX FOUND!	*/
	  usr->rcptflg=rcpt_nomx;
	  if ((iplist=dns_getiplist(eml_getdomain(usr->username)))!=(char **)0) {
	    iplist=(char **)vec_freelstlst((void **)iplist,VECFREE);
	    usr->rcptflg=rcpt_foreign;
	    }
	  break;
	default		:
	  (void) clm_alert(0,"%s Unexpect mx status for domain <%s> (bug?)",
			      OPKN,eml_getdomain(usr->username));
	  break;
	}
      break;
    default	:	/*SAFE Guard		*/
      if (usr!=(USRTYP *)0) {	/*always	*/
	GRPTYP *grp;

        if ((grp=dbc_getgroup(sql,usr->username))!=(GRPTYP *)0) {
          usr->grpid=grp->groupid;
	  grp=dbc_freegroup(grp);
	  }
	}
      proceed=false;
      break;
    }
  phase++;
  }
return usr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if an override	*/
/*	directive is found  within the local	*/
/*	data-base.				*/
/*						*/
/************************************************/
OVRENUM ckr_getoverride(SQLFILE *sql,char *from,char *rcpt,char *locip)

{
#define	OPOV	"lvlckr.c:ckr_getoverride,"

OVRENUM decide;
OVRTYP **over;
AFNTYP *afnnum;
int phase;
int proceed;

decide=tr_maybe;
over=(OVRTYP **)0;
afnnum=(AFNTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*loading override	*/
      if ((over=dbc_getoverride(sql,from,rcpt))==(OVRTYP **)0)
        proceed=false;	/*nothing to do		*/
      break;
    case 1	:	/*getting localip	*/
      if ((afnnum=afn_getipnum(locip))==(AFNTYP *)0) {
	(void) clm_alert(0,"%s Unable to scan IP <%s> (error=<%s>)",OPOV,locip,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*checking for match	*/
      if (over!=(OVRTYP **)0) {	/*always true	*/
	int i;

	for (i=0;over[i]!=(OVRTYP *)0;i++) {
	  if ((decide=checkover(from,over[i],afnnum))!=tr_maybe) {
	    (void) dbc_overhit(sql,over[i]);
	    break;	/*we found a status	*/
	    }
	  }
	}
      afnnum=afn_freeipnum(afnnum);
      break;
    default	:	/*SAFE guard		*/
      over=dbc_freeoverride(over);
      proceed=false;
      break;
    }
  phase++;
  }
return decide;
}
