/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLSPM:					*/
/*	Spam detection routine level.		*/
/*						*/
/************************************************/
#include	<time.h>
#include	<errno.h>
#include	<syslog.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<stdio.h>
#include	<pwd.h>
#include	<sys/stat.h>
#include	<string.h>
#
#include	"lowtyp.h"
#include	"subclm.h"
#include	"unijrl.h"
#include	"gestls.h"
#include	"lvlspm.h"

/*defining spam server variable			*/
#define	CLEMENT_SPMDSRV	"SPMDSRV"

/*spam proctocol Level				*/
#define	SPAMPROT	"SPAMC/1.3"

/*maximun waiting for spamd to respond		*/
#define	MXSPMWAIT    30	

typedef	struct	{
	char *port;	/*spamd used port	*/
	char *hostname;	/*spamd host name	*/
	char *username;	/*spamd access login	*/
	}SPAMFILE;
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the file size in byte	*/
/*						*/
/************************************************/
static long fsize(FILE *fichier)

{
long taille;
struct stat bufstat;

taille=0;
if (fstat(fileno(fichier),&bufstat)<0) {
  (void) clm_alert(0,"lvlspm.c:fsize, Unable to get file stat (bug?)");
  }
else
  taille=bufstat.st_size;
return taille;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect the answer from	*/
/*	spam checker daemon			*/
/*						*/
/************************************************/
static char *getspmansw(TLSFILE *fspam)

{
#define	OPGT	"lvlspm.c:getspmansw,"

char *finding;
int phase;
int proceed;

finding=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  char *ligne;

  /*waiting 300 seconds	*/
  if ((ligne=tls_waitforaction(fspam,10*MXSPMWAIT))==(char *)0) {
    (void) clm_alert(0,"%s timeout waiting for spamd",OPGT);
    break;
    }
  switch(phase) {
    case   0	:
      if (strstr(ligne,"EX_OK")==(char *)0) {
	(void) clm_alert(0,"%s unexpected spamd answer <%s>",OPGT,ligne);
	phase=999;	/*get out quick			*/
	}
      break;
    case   1	:
      if (strlen(ligne)>0) {
	float result;
	float limite;
	char code[100];

        if (sscanf(ligne,"Spam: %19s ; %f / %f",code,&result,&limite)!=3) {
	  (void) clm_alert(0,"%s Unexpected spamd answer format <%s>",OPGT,ligne);
	  }
	else {
	  (void) asprintf(&finding,"%4.2f/%4.2f",result,limite);
	  }
	}
			/*no break			*/
    default	:	/*safe guard			*/
      proceed=false;
      break;
    }
  (void) free(ligne);
  phase++;
  }
return finding;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to detecte if the spam daemon	*/
/*	(spamassassin) is up and running.	*/
/*						*/
/************************************************/
static int isdaemonup(SPAMFILE *spamfile)

{
#define	OPUP	"lvlspm.c:isspamdup,"

int status;
TLSFILE *fspam;
char *ligne;
int code;
int socker;
int phase;
int proceed;

status=false;
fspam=(TLSFILE *)0;
ligne=(char *)0;
code=0;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the connection*/
      if ((fspam=tls_opentcp(spamfile->hostname,spamfile->port,MXSPMWAIT,&socker))==(TLSFILE *)0) 
	phase=999;	/*trouble trouble	*/
      break;
    case 1	:	/*sending a spamd wakeup*/
      (void) tls_printf(fspam,"PING %s\r\n",SPAMPROT);
      if ((ligne=tls_waitforaction(fspam,30))==(char *)0) {
        (void) clm_alert(0,"%s timeout waiting for spamd on <%s/%s>",
			    OPUP,spamfile->hostname,spamfile->port);
	phase=999;	/*trouble trouble	*/
	}
      (void) tls_close(fspam);
      break;
    case 2	:	/*scanning answer	*/
      status=true;
      if (sscanf(ligne,"%*s %d",&code)!=1)
        (void) clm_alert(0,"%s Unexpected ping reply <%s> to PING",ligne);
      else
	if (code!=0)
	  status=false;
      (void) free(ligne);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to contact spam detector and	*/
/*	request about the file spamming level.	*/
/*						*/
/************************************************/
static char *asktodaemon(FILE *fichier,TLSFILE *fspam,char *username)

{
#define OPAS	"levspm.c:asktodaemon,"

char *finding;
long taille;
int phase;
int proceed;

finding=(char *)0;
taille=fsize(fichier);
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case   0	:	/*sending header to spammassassin	*/
      (void) tls_printf(fspam,"CHECK %s\r\n",SPAMPROT);
      if (username==(char *)0)
	username=cuserid((char *)0);
      (void) tls_printf(fspam,"User: %s\r\n",username);
      (void) tls_printf(fspam,"Content-length: %ld\r\n",taille);
      (void) tls_printf(fspam,"\r\n");
      break;
    case   1	:	/*sending file core to spamassassin	*/
      while (taille>0) {
	int max;
        int collected;
        char buffer[1025];

	max=sizeof(buffer)-1;
	if (max>taille)
	  max=taille;
        (void) memset(buffer,'\000',sizeof(buffer));
	if ((collected=fread(buffer,1,max,fichier))<0) {
	  (void) clm_alert(0,"%s Unable to fully collect file, remaining=%d/%d",
			     OPAS,taille,fsize(fichier));
	  proceed=false;
	  break;
          }
	(void) tls_write(fspam,buffer,collected);
        taille-=collected;
        }
      break;
    case   2	:	/*extracting returned information	*/
      finding=getspmansw(fspam);
      break;
    default	:	/*safe guard				*/	
      proceed=false;
      break;
    }
  phase++;
  }
return finding;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open a link with the	*/
/*	spamassassin daemon (spamd)		*/
/*						*/
/************************************************/
SPAMTYP *spm_openspamd(char *hostname,char *port,char *username)

{
#define	OPSD		"lvlspm.c:clm_openspamd,"

#define	SPAMDHOST	"localhost.localdomain"
#define	SPAMDPORT	"783"

SPAMFILE *spamfile;
int phase;
int proceed;

spamfile=(SPAMFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting default values*/
      if (hostname==(char *)0)
	hostname=SPAMDHOST;
      if (port==(char *)0)
	port=SPAMDPORT;
      if (username==(char *)0)
	username=appname;	/*set to clement*/
      spamfile=(SPAMFILE *)calloc(1,sizeof(SPAMFILE));
      spamfile->hostname=strdup(hostname);
      spamfile->port=strdup(port);
      spamfile->username=strdup(username);
      break;
    case 1	:	/*is daemon up and ready*/
      if (isdaemonup(spamfile)==false) {
	(void) clm_alert(3,"%s Unable to contact <%s/%s> spamassassin daemon",
			   OPSD,hostname,port);
	spamfile=spm_closespamd((SPAMTYP *)spamfile);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return (SPAMTYP *)spamfile;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free data used by 		*/
/*	spam detector use.			*/
/*						*/
/************************************************/
SPAMTYP *spm_closespamd(SPAMTYP *spamd)

{
register SPAMFILE *spamfile;

spamfile=(SPAMFILE *)spamd;
if (spamfile!=(SPAMTYP *)0) {
  spamfile->username=clm_freestr(spamfile->username);
  spamfile->port=clm_freestr(spamfile->port);
  spamfile->hostname=clm_freestr(spamfile->hostname);
  (void) free(spamfile);
  spamfile=(SPAMFILE *)0;
  }
return (SPAMTYP *)spamfile;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the spam detector	*/
/*	finding about the file.			*/
/*						*/
/************************************************/
char *spm_checklevel(SPAMTYP **spamd,char *mailid)

{
#define	OPCL	"lvlspm.c:spm_checklevel,"

char *finding;
SPAMFILE *spamfile;
char *mailname;
FILE *fichier;
TLSFILE *fspam;
int socker;
int phase;
int proceed;

finding=(char *)0;
spamfile=(SPAMFILE *)(*spamd);
(void) asprintf(&mailname,"%s/%s",clm_appdir(d_queue),mailid);
fichier=(FILE *)0;
fspam=(TLSFILE *)0;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Spamdaemon UP?	*/
      if (spamfile==(SPAMFILE *)0) 
	proceed=false;	/*NO spamassassin	*/
      break;
    case 1	:	/*lets open file	*/
      if ((fichier=fopen(mailname,"r"))==(FILE *)0) {
	(void) clm_alert(3,"%s Unable to open file <%s> (error=<%s> bug!?)",
			    OPCL,mailname,strerror(errno));
	proceed=false;	/*Trouble trouble	*/
	}
      break;
    case 2	:	/*lets contact spamd	*/
      if ((fspam=tls_opentcp(spamfile->hostname,spamfile->port,MXSPMWAIT,&socker))==(TLSFILE *)0) {
	/*lets say SPAMD lost			*/
        (void) clm_alert(0,"%s Unable to access <%s> port <%s> within '%d' sec (config?)",
			      OPCL,spamfile->hostname,spamfile->port,MXSPMWAIT);
	*spamd=spm_closespamd(*spamd);
	phase=999;
	}
      break;
    case 3	:	/*lets contact spamd	*/
      if ((finding=asktodaemon(fichier,fspam,spamfile->username))==(char *)0) {
        (void) clm_alert(0,"%s Unable to get answer from <%s> port <%s>",
			      OPCL,spamfile->hostname,spamfile->port);
	*spamd=spm_closespamd(*spamd);
	}
      (void) tls_close(fspam);
      break;
    default	:	/*SAFE Guard		*/
      (void) fclose(fichier);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(mailname);
return finding;
}
