/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESSTA:					*/
/*	routine to manage email statistic	*/
/*	collection.				*/
/*						*/
/************************************************/
#include	<string.h>
#include	<time.h>
#include	<stdio.h>
#include	<errno.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"lvlsta.h"

#define MINUTE	 60	/*stat collection slice	*/
#define	MINS	 70	/*collect # of 1 minutes*/
#define	STATS	(((int)m_undefined)+3)
#define	VOLIN	((int)(m_undefined)+1)
#define	VOLOUT	((int)(m_undefined)+2)

#define	NAMAX	60
#define COLFRM	"%s %s  %ld:%lu:%lu:%lu:%lu:%lu:%lu:%lu:%lu\n"


typedef	u_long STAR[MINS][STATS];

typedef	struct	{
	char *grpname;	/*groupname name	*/
	char *domname;	/*domain name		*/
	STAR stats;	/*status collectors	*/
	}COLTYP;

typedef struct	{
	time_t last;	/*last collection date	*/	
	time_t dump;	/*last collection date	*/	
	COLTYP **cols;	/*collection info	*/
	}ALLSTATS;
/*

*/
/************************************************/
/*						*/
/*	Procedure to standardize group and	*/
/*	domain lenght.				*/
/*						*/
/************************************************/
static char *setstd(char *name)

{
register int taille;
char *str;

taille=strlen(name);
str=strdup(name);
if (taille>NAMAX) {
  register char *ptr;

  ptr=str+(taille-NAMAX);
  (void) memmove(str,ptr,strlen(ptr)+1);
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format stats to str format	*/
/*						*/
/************************************************/
static char *tostr(time_t temps,COLTYP *cols,u_long st[STATS])

{
char *str;

(void) asprintf(&str,COLFRM,cols->grpname,
			    cols->domname,
			    temps,
			    st[VOLIN],st[VOLOUT],
			    st[m_direct],st[m_relayed],
			    st[m_quarantine],st[m_rejected],
			    st[m_empty],st[m_virusfound]);
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a new statistic	*/
/*	collection within the collection group	*/
/*						*/
/************************************************/
static COLTYP **addcol(COLTYP **cols,char *grpname,char *domname)

{
COLTYP *col;

col=(COLTYP *)calloc(1,sizeof(COLTYP));
col->grpname=strdup(grpname);
col->domname=strdup(domname);
cols=(COLTYP **)vec_addlstlst((void **)cols,(void *)col);
return cols;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to locate a specific domain	*/
/*	statistic collection within collection	*/
/*						*/
/************************************************/
static COLTYP *findcol(COLTYP **cols,char *grpname,char *domname)

{
COLTYP *col;

col=(COLTYP *)0;
if (cols!=(COLTYP **)0) {
  int i;

  for (i=0;cols[i]!=(COLTYP *)0;i++) {
    if ((strcmp(cols[i]->domname,domname)==0) &&
        (strcmp(cols[i]->grpname,grpname)==0)) {
      col=cols[i];
      break;
      }
    }
  }
return col;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add number to collection	*/
/*	Return the status found.		*/
/*						*/
/************************************************/
static void addnum(COLTYP *col,MENTYP mstatus,u_long in,u_long out)

{
switch (mstatus) {
  case m_trouble	:
  case m_empty		:
  case m_indata		:
    mstatus=m_empty;
    break;
  case m_piped		:
  case m_acceptable	:
  case m_direct		:
    mstatus=m_direct;
    col->stats[0][VOLIN]+=in;
    break;
  case m_mustfwd	:
  case m_toforward	:
    col->stats[0][(int)m_direct]++;
		/*NO BREAK		*/
  case m_toresend	:
  case m_relayed	:
    mstatus=m_relayed;
    col->stats[0][VOLIN]+=in;
    col->stats[0][VOLOUT]+=out;
    break;
  case m_stayput	:
    mstatus=m_quarantine;
		/*NO BREAK		*/
  case m_quarantine	:
    col->stats[0][VOLIN]+=in;
    break;
  case m_discarded	:
  case m_rejected	:
  case m_noaddress	:
    mstatus=m_rejected;
    break;
  case m_ipbanned	:
    break;
  case m_virusfound	:
    break;
  case m_toresolv	:
  case m_progress	:
  case m_delivered	:
  case m_forwarded	:
  case m_delayed	:
  case m_undefined	:
    break;
  }
col->stats[0][(int)mstatus]++;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump specific memory info	*/
/*						*/
/************************************************/
static int dumpstats(time_t temps,COLTYP **cols,int hours)

{
#define	OPDU	"lvlsta.c:dumpstats,"
#define	STNA	"stats"

int done;
char *name;
char *statname;
char *filename;
FILE *fichier;
int phase;
int proceed;

done=false;
name=(char *)0;
statname=(char *)0;
filename=(char *)0;
fichier=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*building file name	*/
      name="minutly";
      if (hours==true)
	name="hourly";
      (void) asprintf(&statname,"%s%s/%s.%s",clm_appdir(d_varlib),STNA,STNA,name);
      (void) asprintf(&filename,"%s.new",statname);
      if ((fichier=fopen(filename,"w"))==(FILE *)0) {
	(void) clm_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			    OPDU,filename,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case  1	:	/*dumping information	*/
      if (cols!=(COLTYP **)0) {
	int i;

	for (i=0;cols[i]!=(COLTYP *)0;i++) {
	  char *str;
	  u_long *infos;
	  u_long stats[STATS];

	  str=(char *)0;
	  infos=&(cols[i]->stats[1][0]);
          (void) (void) memcpy(stats,infos,sizeof(stats));
	  if (hours==true) {
	    int j;

	    for (j=2;j<60;j++) {
	      MENTYP k;

	      infos=&(cols[i]->stats[j][0]);
	      for (k=(MENTYP)0;k<m_undefined;k++) {
	        stats[k]+=*(infos+k);
		}
	      }
	    }
	  if ((str=tostr(temps,cols[i],stats))!=(char *)0) {
	    (void) fprintf(fichier,"%s",str);
	    (void) free(str);
	    }
	  }
	}
      break;
    case  2	:	/*closing the file	*/
      (void) fclose(fichier);
      break;
    case  3	:	/*renameing filename	*/
      (void) unlink(statname);
      if (link(filename,statname)<0)
	(void) clm_alert(0,"%s Unable to link <%s> to <%s> (bug?)",
			   OPDU,filename,statname);
      else
	(void) unlink(filename);
      break;
    default	:	/*SAFE guard		*/
      (void) free(filename);
      (void) free(statname);
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a statistic	*/
/*	collection area.			*/
/*						*/
/************************************************/
STATYP *sta_mkstats()

{
ALLSTATS *allstats;

allstats=(ALLSTATS *)calloc(1,sizeof(ALLSTATS));
allstats->last=time((time_t *)0);
allstats->last+=10;	/*adding starting delay	*/
allstats->last-=(allstats->last%MINUTE);
allstats->dump=allstats->last;
allstats->cols=addcol(allstats->cols,ALLDOM,ALLDOM);
return (STATYP *)allstats;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	statistic collection area.		*/
/*	collection area.			*/
/*						*/
/************************************************/
STATYP *sta_freestats(STATYP *stats)

{
ALLSTATS *allstats;

allstats=(ALLSTATS *)stats;
if (allstats!=(ALLSTATS *)0) {
  if (allstats->cols!=(COLTYP **)0) {
    int i;

    for (i=0;allstats->cols[i]!=(COLTYP *)0;i++) {
      register COLTYP *col;

      col=allstats->cols[i];
      (void) free(col->domname);
      (void) free(col->grpname);
      (void) free(col);
      }
    (void) free(allstats->cols);
    }
  (void) free(allstats);
  allstats=(ALLSTATS *)0;
  }
return (STATYP *)allstats;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to move stats by MINUTE	*/
/*	increment if needed.			*/
/*						*/
/************************************************/
STATYP *sta_movestats(STATYP *stats)

{
ALLSTATS *allstats;

allstats=(ALLSTATS *)stats;
if (allstats!=(ALLSTATS *)0) {
  while ((allstats->last+MINUTE)<time((time_t *)0)) {
    int i;

    for (i=0;allstats->cols[i]!=(COLTYP *)0;i++) {
      u_long *dest;
      u_long *src;

      dest=&(allstats->cols[i]->stats[1][0]);
      src=&(allstats->cols[i]->stats[0][0]);
      (void) memmove(dest,src,(STATS*(MINS-1))*(sizeof(u_long)));
      (void) memset(src,'\000',STATS*(sizeof(u_long)));
      }
    allstats->last+=MINUTE;
    }
  }
return (STATYP *)allstats;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump statistic to a file,	*/
/*	This file will allow a shell script to	*/
/*	feed rrdtools and draw graphic.		*/
/*						*/
/************************************************/
int sta_dumpstats(STATYP *stats)

{
#define	OPEX	"lvlsta.c:sta_dumpstats,"
#define	SHDIR	"shell"
#define	STSH	"statistic"

int done;
ALLSTATS *allstats;

done=false;
allstats=(ALLSTATS *)stats;
if (allstats!=(ALLSTATS *)0) {
  while ((allstats->dump+MINUTE)<time((time_t *)0)) {
    done&=dumpstats(allstats->dump+MINUTE,allstats->cols,false);
    done|=dumpstats(allstats->dump+MINUTE,allstats->cols,true);
    allstats->dump+=MINUTE;
    done=true;
    }
  if (done==true) {
    char *scriptpath;
    char *argv[10];

    (void) asprintf(&scriptpath,"%s%s/%s.sh",clm_appdir(d_usrlib),SHDIR,STSH);
    argv[0]="/bin/sh";
    argv[1]=scriptpath;
    argv[2]=base;
    argv[3]=(char *)0;
    switch(fork()) {
      case -1	:
	(void) clm_alert(0,"%s Unable to start subprocess (error=<%s>)",
			 OPEX,strerror(errno));
	break;
      case  0	:	/*process itself	*/
	if (getuid()==0) {	/*is changing ID possible	*/
	  int egid;
	  int euid;
	
	  egid=getegid();
	  euid=geteuid();
	  if (setregid(egid,egid)<0)
	    (void) clm_alert(0,"%s Unable to set GID to '%d' (error=<%s>)",
				OPEX,getegid(),strerror(errno));
	  if (setreuid(euid,euid)<0)
	    (void) clm_alert(0,"%s Unable to set UID to '%d' (error=<%s>)",
				OPEX,geteuid(),strerror(errno));
	  }
	(void) clm_argvtrace(20,"Starting execv <%s>",argv);
	(void) execv(argv[0],argv);
	(void) clm_argvtrace(0,"Unable to execv, aborting! <%s>",argv);
	(void) exit(-1);/*trouble trouble	*/
	break;
      default	:
	(void) usleep(1000);
	break;
      }
    (void) free(scriptpath);
    }
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump all current statistic	*/
/*	a single line message			*/
/*						*/
/************************************************/
char **sta_statstomsg(time_t temps,STATYP *stats,int ndxtime)

{
#define	MXTAILLE	500

char **statlst;

statlst=(char **)0;
if (stats!=(STATYP *)0) {
  char *statmsg;
  int curtaille;
  ALLSTATS *allstats;
  int i;

  statmsg=(char *)calloc(3,sizeof(char));
  curtaille=0;
  allstats=(ALLSTATS *)stats;
  for (i=0;allstats->cols[i]!=(COLTYP *)0;i++) {
    u_long *infos;
    u_long st[STATS];
    char *str;
    char *newstatmsg;

    infos=&(allstats->cols[i]->stats[ndxtime][0]);
    (void) memcpy(st,infos,sizeof(st));
    str=tostr(temps,allstats->cols[i],st);
    curtaille+=strlen(str);
    if ((newstatmsg=(char *)realloc((void *)statmsg,(curtaille+5)*sizeof(char)))!=(char *)0) {
      statmsg=newstatmsg;
      (void) strcat(statmsg,str);
      }
    (void) free(str);
    if (curtaille>MXTAILLE) {
      statlst=(char **)vec_addlstlst((void **)statlst,(void *)statmsg); 
      curtaille=0;
      statmsg=(char *)calloc(3,sizeof(char));
      }
    }
  if (curtaille>0)
    statlst=(char **)vec_addlstlst((void **)statlst,(void *)statmsg); 
  else
    (void) free(statmsg);
  }
return statlst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transfer all stats from	*/
/*	a message list to current statistic	*/
/*	data.					*/
/*						*/
/************************************************/
STATYP *sta_msgtostats(STATYP *stats,char **msglst)

{
#define	OPMS	"lvlsta.c:sta_msgtostats,"

ALLSTATS *allstats;

allstats=(ALLSTATS *)stats;
if (msglst!=(char **)0) {
  int i;

  for (i=0;msglst[i]!=(char *)0;i++) {
    int found;
    COLTYP *col;
    char *grpname;
    char *domname;
    char *locgrp;
    char *locdom;
    time_t curtime;
    int phase;
    int proceed;
    u_long st[STATS];

    found=0;
    col=(COLTYP *)0;
    grpname=(char *)calloc(strlen(msglst[i]),sizeof(char));
    domname=(char *)calloc(strlen(msglst[i]),sizeof(char));
    locgrp=(char *)0;
    locdom=(char *)0;
    curtime=0;
    phase=0;
    proceed=true;
    while (proceed==true) {
      switch (phase) {
	case 0	:	/*scanning data		*/
          found=sscanf(msglst[i],COLFRM,grpname,domname,&curtime,
					st+VOLIN,st+VOLOUT,
					st+m_direct,st+m_relayed,
					st+m_quarantine,st+m_rejected,
					st+m_empty,st+m_virusfound);
	  if (found<11) {
	    (void) clm_alert(0,"%s Unable to scan string<%s>",OPMS,msglst[i]);
	    phase=999;	/*trouble trouble	*/
	    }
	  break;
	case 1	:	/*standard grp, dom size*/
          locgrp=setstd(grpname);
          locdom=setstd(domname);
          break;
	case 2	:	/*lets find the col data*/
          if ((col=findcol(allstats->cols,locgrp,locdom))==(COLTYP *)0) {
	    allstats->cols=addcol(allstats->cols,locgrp,locdom);
	    phase--;	/*lets try again	*/
	    }
          break;
	case 3	:	/*adding all data	*/
          if (col!=(COLTYP *)0) {	/*always*/
	    int k;

	    for (k=0;k<STATS;k++)
	      col->stats[0][k]+=st[k];
	    }
          break;
	case 4	:	/*cleaning		*/
	  (void) free(locdom);
	  (void) free(locgrp);
	  break;
        default	:	/*SAFE guard		*/
          proceed=false;
	  break;
        }
      phase++;
      }
    (void) free(domname);
    (void) free(grpname);
    }
  }
return stats;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect all stats in the	*/
/*	proper statistique array.		*/
/*						*/
/************************************************/
STATYP *sta_collationne(STATYP *stats,char *grpname,char *domname,
		        MENTYP mstatus,u_long in,u_long out)

{
ALLSTATS *allstats;
COLTYP *col;
char *locgrp;
char *locdom;
int phase;
int proceed;

allstats=(ALLSTATS *)stats;
col=(COLTYP *)0;
locgrp=(char *)0;
locdom=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sparing trouble	*/
      if ((allstats==(ALLSTATS *)0)||(allstats->cols==(COLTYP **)0)) {
        (void) clm_alert(0,"lvlsta.c:sta_collationne, cols missing (bug?)");
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*standard grp, dom size*/
      locgrp=setstd(grpname);
      locdom=setstd(domname);
      break;
    case 2	:	/*lets find the col data*/
      if ((col=findcol(allstats->cols,locgrp,locdom))==(COLTYP *)0) {
	allstats->cols=addcol(allstats->cols,locgrp,locdom);
	phase--;	/*lets try again	*/
	}
      break;
    case 3	:	/*adding data		*/
      (void) addnum(col,mstatus,in,out);
      break;
    case 4	:	/*cleaning		*/
      (void) free(locdom);
      (void) free(locgrp);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return (STATYP *)allstats;
}
