/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLTRT:					*/
/*	Email transit management.		*/
/*						*/
/************************************************/
#include	<time.h>
#include	<errno.h>
#include	<syslog.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<stdio.h>
#include	<pwd.h>
#include	<sys/stat.h>
#include	<string.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"unijrl.h"
#include	"unidns.h"
#include	"comeml.h"
#include	"gesdbc.h"
#include	"gesque.h"
#include	"gestru.h"
#include	"lvltrt.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to clean DLG structure if the	*/
/*	data was received from destination.	*/
/*						*/
/************************************************/
static DLGTYP **purgedata(char *outip,DLGTYP **data)

{
if (data!=(DLGTYP **)0) {
  register int i;
  char *dest;

  dest=outip;
  if (strstr(dest,STRIPV6)==dest)
     dest+=strlen(STRIPV6);
  for (i=0;data[i]!=(DLGTYP *)0;i++) {
    register int dopurge;

    dopurge=false;
    if (data[i]->flgdata==false) {
      data[i]->frwdata.frw=purgedata(outip,data[i]->frwdata.frw);
      if (data[i]->frwdata.frw==(DLGTYP **)0)
	dopurge=true;
      }
    if (strlen(data[i]->orig)>0) {
      char *ptr;

      if ((ptr=strchr(data[i]->orig,'['))!=(char *)0) {
        char *ip;

        ip=strdup(ptr+1);
	if ((ptr=strrchr(ip,']'))!=(char *)0)
	  *ptr='\000';
	ptr=ip;
        if (strstr(ptr,STRIPV6)==ptr)
	  ptr+=strlen(STRIPV6);
	if (strcmp(dest,ptr)==0) 
	  dopurge=true;
	(void) free(ip);
	}
      }
    if (dopurge==true) {
      register int taille;

      taille=vec_sizelstlst((void **)(data+i));
      data[i]=tru_freedlg(data[i]);
      (void) memmove(data+i,data+i+1,taille*sizeof(DLGTYP *));
      i--;
      }
    }
  if (data[0]==(DLGTYP *)0) {
    (void) free(data);
    data=(DLGTYP **)0;
    }
  }
return data;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to clean stream type to remove*/
/*	data already received from origin	*/
/*						*/
/************************************************/
static STRMTY **purgestrms(char *outip,STRMTY **strms)

{
if (strms!=(STRMTY **)0) {
  register int i;

  for (i=0;strms[i]!=(STRMTY *)0;i++) {
    if ((strms[i]->data=purgedata(outip,strms[i]->data))==(DLGTYP **)0) {
      int taille;

      taille=vec_sizelstlst((void **)(strms+i));
      strms[i]->ip=clm_freestr(strms[i]->ip);
      (void) free(strms[i]);
      (void) memmove(strms+i,strms+i+1,taille*sizeof(STRMTY *));
      i--;
      }
    }
  if (strms[0]==(STRMTY *)0) {
    (void) free(strms);
    strms=(STRMTY **)0;
    }
  }
return strms;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert a stream feed and	*/
/*	send it to trusted remote		*/
/*						*/
/************************************************/
static void sendtrustdata(CTTYP *contact,STRMTY **strms)

{
#define	OPND	"lvltrt.c:sendtrustdata,"

if (strms!=(STRMTY **)0) { /*always	*/
  register int i;

  for (i=0;strms[i]!=(STRMTY *)0;i++) {
    register int j;
    register char **data;

    if ((data=tru_dumpstrm(contact->locname,contact->locip,strms[i]))==(char **)0) {
      (void) clm_alert(0,"%s No trusted data to feed to <%s> (bug?)",OPND,contact->peername);
      break;
      }
    for (j=0;data[j]!=(char *)0;j++) {
      if (strstr(data[j],"#-")==data[j])
	continue;
      (void) eml_printfcrlf(contact,"%s",data[j]);
      }
    data=(char **)vec_freelstlst((void **)data,VECFREE);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to feed remote trusted server	*/
/*	with trusted feed.			*/
/*	status:					*/
/*		0	sending not successful	*/
/*		1	sending successful	*/
/*	       -1	remote not trusting us	*/
/*						*/
/*						*/
/************************************************/
static int sendingfeed(char *outip,TRUTYP *tru)

{
#define	OPSE	"lvltrt.c:sendingfeed,"

int done;
CTTYP *contact;
int socker;
int phase;
int proceed;

done=0;
contact=(CTTYP *)0;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case 0	:	/*connecting remote	*/
      if (strstr(outip,STRIPV6)==outip)
  	outip+=strlen(STRIPV6);
      contact=eml_opencontact(pro_smtp,outip,SMTPPORT,60,&socker);
      if (contact==(CTTYP *)0) {
	(void) clm_alert(0,"%s Unable to establish contact with trusting <%s>",OPSE,outip);
	proceed=false;
	}
      break;
    case 1	:	/*remote trusting us?	*/
      if ((contact->ctxflag&VALDTED)==0) {
	(void) clm_alert(0,"%s Trusted status lost with <%s>",OPSE,contact->peername);
	done=-1;	/*we lost trust status!	*/
	phase=999;	/*no need to go further	*/
	}
      break;
    case 2	:	/*sending TRUST request	*/
      if (eml_sendcmd(contact,TRUSTOK,"%s",HLOTRS)!=TRUSTOK) {
	(void) clm_alert(0,"%s remote <%s> not accepting <%s> command (bug?)",OPSE,contact->peername,HLOTRS);
	done=-1;	/*we lost trust status!	*/
	phase=999;	/*get out fast		*/
	}
      break;
    case 3	:	/*sending data out	*/
      (void) sendtrustdata(contact,tru->strms);
      break;
    case 4	:	/*sending end marker	*/
      if (eml_sendcmd(contact,CMDOK,"%s",DATAEND)!=CMDOK) {
	(void) clm_alert(0,"%s remote <%s> not accepting our trust feed (bug?)",OPSE,contact->peername);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 5	:	/*status is OK		*/
      done=1;
      break;
    default	:	/*SAFE Guard		*/
      (void) jrl_journalling(true,EVTRU,"Sent trusted data to <%s> status='%d'",contact->peername,done);
      contact=eml_closecontact(contact);
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to move a data-file within	*/
/*	the trouble queue			*/
/*						*/
/************************************************/
static int totrblq(SQLFILE *sql,QUETYP *q,time_t curtime)

{
#define	OPTR	"lvltrt.c:totrblq,"


int done;
char *datafile;
char *newfile;

done=true;
datafile=que_getdatafilename(d_transit,q);
newfile=que_getdatafilename(d_trblq,q);
if (link(datafile,newfile)<0) {
  done=false;		/*oups trouble		*/
  switch (errno) {
    case EXDEV	:
      done=clm_dupfile(datafile,newfile);
      break;
    default	:
      (void) clm_alert(0,"%s Unable to move file <%s> to <%s> (error=<%s>)",
			  OPTR,datafile,newfile,strerror(errno));
      break;
    }
  }
(void) unlink(datafile);
(void) free(newfile);
(void) free(datafile);
q->nexttime=curtime+(360*24*3600);
q->lasttime=curtime;
q->progress=QTRBL;
q->status=FAILED;
q->retry++;
if (q->why!=(char *)0) {
  char *ptr;

  while ((ptr=strchr(q->why,'\n'))!=(char *)0)
    *ptr=' ';
  while ((ptr=strchr(q->why,'\r'))!=(char *)0)
    *ptr=' ';
  }
if (done==true) {
  (void) jrl_journalling(false,EVTRT,"Moving transaction %s,<%s@%s> to trouble queue",
	      			      q->session,q->recname,q->recdom);
  (void) dbc_updqueue(sql,q);
  }
else
  (void) dbc_delqueue(sql,q);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a transaction from	*/
/*	Queuing reference.			*/
/*	Removing both the record and the file	*/
/*	linked to it.				*/
/*						*/
/************************************************/
int trt_delqueue(SQLFILE *sql,QUETYP *queue)

{
#define	OPDL	"lvltrt.c:trt_delqueue,"

int done;

done=false;
if (queue!=(QUETYP *)0) {
  char *datafile;

  done=true;
  datafile=que_getdatafilename(d_transit,queue);
  if (unlink(datafile)<0) {
    (void) clm_alert(0,"%s Unable to remove file <%s> (error=<%s>) (bug?)",
		        OPDL,datafile,strerror(errno));
    }
  (void) dbc_delqueue(sql,queue);
  (void) free(datafile);
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage queue status.	*/
/*	Removing data file if email transit	*/
/*	reached a final status.			*/
/*						*/
/************************************************/
QUETYP **trt_cleanqueues(SQLFILE *sql,QUETYP **queues)

{
#define	OPCL	"lvltrt.c:trt_cleanqueues,"

if (queues!=(QUETYP **)0) {
  time_t curtime;
  int i;

  curtime=time((time_t *)0);
  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    switch (queues[i]->status) {
      case RETRY	:	/*rmt not ready	*/
      case INRELAY	:	/*no contact	*/
	queues[i]->progress=QWAIT;
        (void) dbc_updqueue(sql,queues[i]);
	break;
      case FAILED	:	/*big failure	*/
      case BOUNCE	:	/*bounced	*/
	queues[i]->progress=QTRBL;
        (void) dbc_updqueue(sql,queues[i]);
	break;
      case CMDOK	:	/*transmitted	*/
	(void) trt_delqueue(sql,queues[i]);
	break;
      default		:	/*move it to trbl queue*/
	(void) clm_alert(0,"%s Unexpected status='%d' for transaction <%s-%d> (bug?)",
			    OPCL,queues[i]->status,queues[i]->session,queues[i]->transit);
	(void) totrblq(sql,queues[i],curtime);
	break;
      }
    queues[i]=que_freequeue(queues[i]);
    }
  (void) free(queues);
  queues=(QUETYP **)0;
  }
return queues;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove 'old' record	*/
/*	from the data-base.			*/
/*						*/
/************************************************/
void trt_dbcleanup(SQLFILE *sql,time_t curtime)

{
QUETYP **queues;
int phase;
int proceed;

curtime-=(3*H24);
queues=(QUETYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*easy cleanup		*/
      (void) dbc_dbcleanup(sql);
      break;
    case 1	:	/*inqueue to clean	*/
      if ((queues=dbc_getoldqueues(sql,curtime,QTRBL))!=(QUETYP **)0) {
	int i;

	for (i=0;queues[i]!=(QUETYP *)0;i++)
	  (void) trt_delqueue(sql,queues[i]);
	queues=que_freequeues(queues);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to write down the resend file	*/
/*	within the transit directory.		*/
/*						*/
/************************************************/
u_int trt_transit(SQLFILE *sql,char *outip,char *smtpmx,char *smtpport,
		  char *from,DSPTYP *dsp,char *infile,RCPTYP **rcpts)

{
#define	OPSI	"lvltrt.c:trt_transit,"

int numdone;
QUETYP **queues;
int phase;
int proceed;

numdone=0;
queues=(QUETYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*preparing data		*/
      queues=que_mkqueues(outip,smtpmx,smtpport,dsp,from,mm_std,rcpts);
      break;
    case   1	:	/*checking for trouble		*/
      if (queues==(QUETYP **)0) {
	(void) clm_alert(0,"%s Unable to make QUETYP (bug!)",OPSI);
	phase=999;	/*we are really in big trouble	*/
  	}
      break;
    case   2	:	/*dumping transit record	*/
      if (dbc_createqueues(sql,queues)==0) {
	(void) clm_alert(0,"%s Unable to update database with queue  (bug!)",OPSI);
	phase=999;	/*we are really in big trouble	*/
	}
      break;
    case   3	:	/*dumping transit record	*/
      if (queues!=(QUETYP **)0) {	/*always	*/
	char *newshead;
	int i;

        (void) asprintf(&newshead,"Set New transit %s",dsp->session);
	(void) clm_fasprintf(&newshead,"\n\t\t%08s from <%s>","",from);
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  int ready;
	  char *cmt;
	  char *lfile;

	  ready=true;
	  cmt=""; 
	  lfile=que_getdatafilename(d_transit,queues[i]);
	  if (link(infile,lfile)<0) {
	    (void) clm_alert(0,"%s Unable to link file <%s> to file <%s> (errno=<%s>))",
				OPSI,infile,lfile,strerror(errno));
	    cmt=" not queued (trouble?)";
	    ready=false;
	    }
	  (void) clm_fasprintf(&newshead,"\n\t\t%08d to   <%s@%s>%s",
					  queues[i]->transit,queues[i]->recname,queues[i]->recdom,cmt);
	  (void) free(lfile);
	  if (ready==true)
            (void) dbc_rdyqueue(sql,queues[i]);
	  numdone++;
          }
	(void) jrl_journalling(false,EVCKR,"%s",newshead);
	(void) free(newshead);
	}
      break;
    default	:	/*SAFE guard			*/
      queues=que_freequeues(queues);
      proceed=false;
      break;
    }
  phase++;
  }
return numdone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a banned stream to	*/
/*	trusting remote site			*/
/*						*/
/*	status:					*/
/*		0	sending not successful	*/
/*		1	sending successful	*/
/*	       -1	remote not trusting us	*/
/*						*/
/************************************************/
int trt_feedbanned(char *outip,TRUTYP *tru)

{
#define	OPFE	"lvltrt.c:trt_feedbanned,"

int done;
int phase;
int proceed;

done=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*security		*/
      if (tru==(TRUTYP *)0) {
	(void) clm_alert(0,"%s Unexpected empty data (bug!)",OPFE);
	phase=9999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*purging stream	*/
      if ((tru->strms=purgestrms(outip,tru->strms))==(STRMTY **)0) 
	phase=999;	/*nothing to be send	*/
      break;
    case 2	:	/*feeding remote	*/
      done=sendingfeed(outip,tru);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
