/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Implement very sub level procedure to 	*/
/*	debug potential process problem		*/
/*						*/
/************************************************/
#include	<unistd.h>
#include	<fcntl.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<signal.h>
#include	<errno.h>
#include	<sys/resource.h>
#include	<sys/prctl.h>
#include	<sys/time.h>
#include	<sys/ioctl.h>

#include	"lowtyp.h"
#include	"subprc.h"

/*Process command override structure (proc)	*/
typedef	struct	{
	int max;	/*title max size	*/
	char *title;	/*title storage area	*/
	}TITLTYP;

/*storage area for /proc title display		*/
static  TITLTYP *title=(TITLTYP *)0;
/*

*/
/************************************************/
/*						*/
/*	Procedure to demonize the current	*/
/*	process once in dive mode.		*/
/*						*/
/************************************************/
static void demonize()

{
#define	MXSTD	3
#define	DEVNULL	"/dev/null"

int i;

for (i=0;i<MXSTD;i++) {
  int fd;

  if ((fd=open(DEVNULL,O_RDONLY))>0) {
    (void) dup2(fd,i);
    (void) close(fd);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to catch a SIGALRM signal	*/
/*						*/
/************************************************/
static void trppace(int sig)

{
#define	OPTR	"subprc.c:trppace,"

static u_long tics=(u_long)0;

switch (sig) {
  case SIGALRM   :
    tics++;
    break;
  default        :
    (void) fprintf(stderr,"%s Unexpected signal <%s> received\n",
			  OPTR,sys_siglist[sig]);
    (void) fflush(stderr);
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free used title memory	*/
/*						*/
/************************************************/
void prc_cleantitle()

{
if (title!=(TITLTYP *)0) {
  if (environ!=(char **)0) {
    int i;

    for (i=0;environ[i]!=(char *)0;i++) {
      (void) free(environ[i]);
      environ[i]=(char *)0;
      }
    (void) free(environ);
    environ=(char **)0;
    }
  (void) free(title);
  title=(TITLTYP *)0;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find and limit space to	*/
/*	be used as status information available	*/
/*	from proc (ps ax)			*/
/*						*/
/************************************************/
void prc_preptitle(int argc,char *argv[],char *env[])

{
extern char **environ;

char *lastend;

lastend=(char *)0;
(void) prc_cleantitle();
title=(TITLTYP *)calloc(1,sizeof(TITLTYP));
if (argv!=(char **)0) {
  int i;

  title->title=argv[0];
  for (i=1;i<argc;i++) {
    lastend=argv[i]+strlen(argv[i]);
    argv[i]=(char *)0;
    }
  }
if (env!=(char **)0) {
  int i;

  environ=(char **)0;
  for (i=0;env[i]!=(char *)0;i++) {
    char *valeur;

    lastend=env[i]+strlen(env[i]);
    valeur=strdup(env[i]);
    (void) putenv(valeur);
    }
  }
title->max=lastend-title->title;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update title information.	*/
/*	title information is available via the	*/
/*	"ps" command.				*/
/*						*/
/************************************************/
void prc_settitle(const char *fmt,...)

{
va_list args;

va_start(args,fmt);
if ((title!=(TITLTYP *)0)&&(title->title!=(char *)0)) {
  (void) bzero(title->title,title->max);
  (void) vsnprintf(title->title,title->max,fmt,args);
  }
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to put a process in background*/
/*	mode.					*/
/*						*/
/************************************************/
void prc_divedivedive(int foreground)

{
#define	OPDI	"dbgprc.c:prc_divedivedive,"

if (foreground==0) {
  switch (fork()) {
    case -1	:
      (void) fprintf(stderr,"%s, Unable to dive! (error=<%s>)",
			     OPDI,strerror(errno));
      /*lets continue if foregorund mode		*/
      break;
    case  0	:
      if (setsid()<0) 
        (void) fprintf(stderr,"%s, Unable to setsid! (error=<%s>)",
			       OPDI,strerror(errno));
      (void) demonize();
      /*we are now in background mode		*/
      break;
    default	:
      /*waiting for ballat to fill up :-}}	*/
      (void) sleep(1);	
      (void) exit(0);	/*just exit		*/
      break;		/*never reached		*/
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to start or stop a 'Pacemaker'*/
/*	within the application.			*/
/*						*/
/************************************************/
void prc_pace(u_long millisec,int onoff)

{
static int done=false;
struct sigaction oldsa;

if (onoff==true) {
  if (done==false) {
    struct sigaction newsa;
    struct itimerval period;

    newsa.sa_flags=0;
    newsa.sa_handler=trppace;
    (void) sigemptyset(&newsa.sa_mask);
    (void) sigaction(SIGALRM,&newsa,&oldsa);
    period.it_value.tv_sec=millisec/1000;
    period.it_value.tv_usec=millisec*1000;
    period.it_interval.tv_sec=period.it_value.tv_sec;
    period.it_interval.tv_usec=period.it_value.tv_usec;
    (void) setitimer(ITIMER_REAL,&period,(struct itimerval *)0);
    done=true;
    }
  }
else {
  if (done==true) {
    struct itimerval period;

    period.it_value.tv_sec=0;
    period.it_value.tv_usec=0;
    period.it_interval.tv_sec=period.it_value.tv_sec;
    period.it_interval.tv_usec=period.it_value.tv_usec;
    (void) setitimer(ITIMER_REAL,&period,(struct itimerval *)0);
    (void) sigaction(SIGALRM,&oldsa,(struct sigaction *)0);
    done=false;
    }
  }
}
