/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIDEV:					*/
/*	Take Care of all access to devices	*/
/*	could be system dependent.		*/
/*						*/
/************************************************/
#include	<stdlib.h>
#include	<unistd.h>
#include	<syslog.h>
#include	<string.h>
#include	<fcntl.h>
#include	<errno.h>
#include	<time.h>
#include	<netdb.h>
#include	<ctype.h>
#include	<signal.h>
#include	<sys/ioctl.h>
#include	<net/if.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"utlprc.h"
#include	"subclm.h"
#include	"basafn.h"
#include	"unidev.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	binding info.				*/
/*						*/
/************************************************/
static BININF *freebinding(BININF *binding)

{
if (binding!=(BININF *)0) {
  binding->port=clm_freestr(binding->port);
  binding->ip=clm_freestr(binding->ip);
  (void) free(binding);
  binding=(BININF *)0; 
  }
return binding;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to accept a new handle from	*/
/*	the list of connect request.		*/
/*						*/
/************************************************/
static int getnewhandle(LSTTYP *lst)

{
#define	GTPF	"unidev.c:getnewhandle,"

int newhandle;
socklen_t taille;
struct sockaddr addr;
int flags;

newhandle=-1;
taille=sizeof(addr);
if ((newhandle=accept(lst->handle,&addr,&taille))<0) {
  switch (errno)  {
    case EAGAIN		:
      /*well process not quick enough the handle was already	*/
      /*accepted by another clement???.				*/
      break;
    case EINVAL		:	/*no socket avail anymore	*/
      break;
    default		:
      (void) clm_alert(0,"%s Unexpected error on IP/PORT <%s/%d> errno=<%s>",
		         GTPF,lst->binding->ip,lst->binding->port,
			 strerror(errno));
      break;
    }
  return newhandle;
  }
if ((flags=fcntl(newhandle,F_GETFL,0))<0)
  (void) clm_core_dump("%s Unable to get socket descripteur (error='%s')",
		        GTPF,strerror(errno));
if ((flags=fcntl(newhandle,F_SETFL,flags|O_NONBLOCK|O_ASYNC))<0)
  (void) clm_core_dump("%s Unable to set socket descripteur (error='%s')",
		        GTPF,strerror(errno));
return newhandle;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a 	*/
/*	LSTTYP information.			*/
/*						*/
/************************************************/
LSTTYP *dev_onefree(LSTTYP *lst)

{
if (lst!=(LSTTYP *)0) {
  lst->binding=freebinding(lst->binding);
   if (lst->process!=(pid_t *)0)
     (void) free(lst->process);
  (void) free(lst);
  lst=(LSTTYP *)0;
  }
return lst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close one waiting channel	*/
/*	an terminating all related process.	*/
/*						*/
/************************************************/
PRCTYP *dev_oneclose(PRCTYP *prc,LSTTYP *lst)

{
if (shutdown(lst->handle,SHUT_RDWR)<0) {
  (void) clm_alert(0,"unidev.c:devoneclose, shutdown error on IP/PORT <%s/%d> "
		     "(error='%s')",lst->binding->ip,lst->binding->port,
		     strerror(errno));
  }
if (lst->binding->iteration>0) {
  register pid_t *pidlst;
  register unsigned int i;

  pidlst=(pid_t *)calloc(lst->binding->iteration,sizeof(pid_t));
  for (i=0;i<lst->binding->iteration;i++) {
    if (lst->process[i]!=(pid_t)0) {
      prc=prc_addtoprc(prc,lst->process[i]);
      }
    }
  (void) free(pidlst);
  }
(void) close(lst->handle);
lst->handle=-1;
return prc;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to bind a socket to a specific*/
/*	address and return the linked handle	*/
/*						*/
/************************************************/
int bindhandle(struct addrinfo *ai,char *port,int mxstrick)

{
#define	OPBI	"unidev.c:bindhandle,"

int handle;
int phase;
int proceed;

handle=-1;
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*trouble phase		*/
      (void) close(handle);
      handle=-1;
      phase=999;	/*check for trouble	*/
      break;
    case 1	:	/*lets get a socket	*/
      if ((handle=socket(ai->ai_family,ai->ai_socktype,ai->ai_protocol))<0) {
        (void) clm_alert(0,"%s Unable to open socket for <%s> (error='%s')",
			   OPBI,ai->ai_canonname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*getting socket option	*/
      if (handle>=0) {	/*always		*/
	struct linger slg;
	char buf[30];
	
	slg.l_onoff=true;
	slg.l_linger=5;	/*5 sec timeout on close*/
	if (setsockopt(handle,SOL_SOCKET,SO_REUSEADDR,buf,sizeof(buf))<0) {
          (void) clm_alert(0,"%s Unable to get socket option for <%s> (error='%s')",
			     OPBI,ai->ai_canonname,strerror(errno));
	  phase=-1;	/*trouble cleanup phase	*/
	  break;
	  }
	if (setsockopt(handle,SOL_SOCKET,SO_LINGER,&slg,sizeof(slg))<0) {
          (void) clm_alert(0,"%s Unable to set socket option for <%s> (error='%s')",
			     OPBI,ai->ai_canonname,strerror(errno));
	  phase=-1;	/*trouble cleanup phase	*/
	  }
	}
      break;
    case 3	:	/*getting socket option	*/
      if (handle>=0) {	/*always		*/
	int flags;

	if ((flags=fcntl(handle,F_GETFL,0))<0) {
          (void) clm_alert(0,"%s Unable to set socket descriptor for <%s> (error='%s')",
			   OPBI,ai->ai_canonname,strerror(errno));
	  phase=-1;	/*trouble cleanup phase	*/
	  break;
	  }
	if (fcntl(handle,F_SETFL,flags|O_NONBLOCK|O_ASYNC)<0) {
          (void) clm_alert(0,"%s Unable to set socket in non block mode for <%s> (error='%s')",
			     OPBI,ai->ai_canonname,strerror(errno));
	  phase=-1;	/*trouble cleanup phase	*/
	  }
	}
      break;
    case 4	:	/*getting socket option	*/
      if (bind(handle,ai->ai_addr,ai->ai_addrlen)<0) {
	(void) clm_alert(0,"%s Unable to bind on IP/port <%s/%s> (error='%s')",
			   OPBI,ai->ai_canonname,port,strerror(errno));
	phase=-1;	/*trouble cleanup phase	*/
	}
      break;
    case 5	:	/*preparing listen	*/
      if (listen(handle,5)<0) {
	(void) clm_alert(0,"%s Unable to listen on IP/port <%s/%s> (error='%s')",
			   OPBI,ai->ai_canonname,port,strerror(errno));
	phase=-1;	/*trouble cleanup phase	*/
	}
      break;
    case 6	:	/*reporting listen	*/
      (void) clm_alert(1,"Listening on IP/PORT <%s/%s>",ai->ai_canonname,port);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      if ((handle<0)&&(mxstrick>0)) {
	(void) sleep(1);
	mxstrick--;
	phase=-1;	/*lets try again	*/
	proceed=true;
	}
      break;
    }
  phase++;
  }
return handle;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to bind a list of interface	*/
/*	to a specific port.			*/
/*						*/
/************************************************/
int dev_mullisten(LSTTYP **lst)

{
#define MXSTRICK 1
#define	MULOPEN	"unidev.c:dev_mullisten,"

int fullyok;

fullyok=true;
if (lst!=(LSTTYP **)0) {
  uid_t gid;
  uid_t uid;
  struct addrinfo hints;
  int i;

  gid=getegid();	/*let be back to root if needed	*/
  uid=geteuid(); 	/*to open device on < 1024 port	*/
  (void) memset(&hints, 0, sizeof(hints));
  hints.ai_family=PF_UNSPEC;  
  hints.ai_flags=HINTFLG;
  hints.ai_socktype=SOCK_STREAM;
  if (setegid(getgid())<0) {
    (void) clm_core_dump("%s Unable to set the Egid to '%d'",MULOPEN,getgid());
    }
  if (seteuid(getuid())<0) {
    (void) clm_core_dump("%s Unable to set the Euid to '%d'",MULOPEN,getuid());
    }
  for (i=0;(lst[i]!=(LSTTYP *)0);i++) {
    int status;
    struct addrinfo *res;

    if (lst[i]->handle>=0)
      continue;		/*handle alreay open			*/
    if ((lst[i]->lasttry+5)>time((time_t)0))
      continue;		/*binding try done less than 5 sec ago	*/
    switch (status=getaddrinfo(lst[i]->binding->ip,lst[i]->binding->port,&hints,&res)) {
      case 0 	:	/*everything OK				*/
	break;
      default	:
        lst[i]->lasttry=time((time_t *)0);
        (void) clm_alert(0,"%s Unable to find a address about IP:port '%s:%s' (error='%s')",
                            MULOPEN, lst[i]->binding->ip,lst[i]->binding->port,gai_strerror(status));
        continue;		/*no need to go further			*/
	break;
      }
    lst[i]->handle=bindhandle(res,lst[i]->binding->port,MXSTRICK);
    lst[i]->lasttry=time((time_t *)0);
    if (lst[i]->handle<0)
      fullyok=false;
    }
  (void) setegid(gid);
  (void) seteuid(uid);
  }
return fullyok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open a channel/process to	*/
/*	every binded interface.			*/
/*						*/
/************************************************/
void dev_mulopen(LSTTYP **lst,
		 void (*tlsact)(int compool,LSTTYP *channel,int iteration))

{
#define	OPML	"unidev.c:dev_mulopen,"

if (lst!=(LSTTYP **)0) {
  int i;

  for (i=0;lst[i]!=(LSTTYP *)0;i++) {
    unsigned int j;

    if (lst[i]->handle<0)
      continue;
    for (j=0;j<lst[i]->binding->iteration;j++) { 
      if (lst[i]->process[j]==(pid_t)0) {
        lst[i]->process[j]=fork();
	switch (lst[i]->process[j]) {
	  case   -1	:
	    (void) clm_alert(0,"%s Unable to fork process (error=<%s>)",
				OPML,strerror(errno));
				
	    (void) sleep(1);
	    break;
	  case	  0	:
	    (void) tlsact(CLMPOOL,lst[i],j);
	    (void) exit(0);
	    break;
	  default	:
	    break;
          }
	}
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove bindinf from	*/
/*	interface and specific port.		*/
/*						*/
/************************************************/
void dev_mulclose(LSTTYP **lst)

{
if (lst!=(LSTTYP **)0) {
  int i;
  PRCTYP *prc;

  prc=(PRCTYP *)0;
  for (i=0;lst[i]!=(LSTTYP *)0;i++) {
    if (lst[i]->handle<0)
      continue;
    prc=dev_oneclose(prc,lst[i]);
    }
  prc=prc_childkill(prc,5);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait a connection from a	*/
/*	client, return NULL if quickout flag	*/
/*	is set, a FILE pointer otherwise.	*/
/*						*/
/************************************************/
int dev_acceptfile(LSTTYP *lst)

{
#define	ACPTF	"unidev.c:dev_acceptfile,"
int newhandle;

newhandle=-2;
if (lst!=(LSTTYP *)0) {
  int proceed;

  proceed=true;
  while (proceed==true) {
    fd_set reading;
    struct timeval relax;

    proceed=false;
    if (lst->handle<0) 
      break;
    FD_ZERO(&reading);
    FD_SET(lst->handle,&reading);
    relax.tv_sec=1;	/*1 sec relax time	*/
    relax.tv_usec=0;
    switch (select(lst->handle+1,&reading,(fd_set *)0,(fd_set *)0,&relax)) {
      case	-1	:
	switch (errno) {
	  case EINTR	:	/*received a signal, lets see...*/
	    break;		/*getting out fast		*/
	  default	:
	    (void) clm_core_dump("%s select error '%s'",ACPTF,strerror(errno));
	    break;
	  }
        break;
      case      0	:	/*to check if parent lost each	*/
				/*'relax' time, if not lets wait*/
        proceed=(prc_checkprocess(getppid())&&(sigterm==false));
	break;
      default		: 
	newhandle=getnewhandle(lst);
	break;
      }
    }
  }
return newhandle;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to duplicat binding contents	*/
/*						*/
/************************************************/
BININF *dev_dupbind(BININF *binding)

{
BININF *neu;

neu=(BININF *)0;
if (binding!=(BININF *)0) {
  neu=(BININF *)calloc(1,sizeof(BININF));
  neu->proto=binding->proto;
  neu->ip=strdup(binding->ip);
  neu->port=strdup(binding->port);
  neu->iteration=binding->iteration;
  }
else
  (void) clm_core_dump("%s, duplicating null binding info!",
		       "unidev.c:dev_dupbind,");
return neu;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a bind	*/
/*	definition.				*/
/*						*/
/************************************************/
BININF **dev_freebindinf(BININF **bindref)

{
if (bindref!=(BININF **)0) {
  int i;

  for (i=0;bindref[i]!=(BININF *)0;i++) {
    bindref[i]=freebinding(bindref[i]);
    }
  (void) free(bindref);
  bindref=(BININF **)0;
  }
return bindref;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set memory used by a bind	*/
/*	definition.				*/
/*						*/
/************************************************/
BININF *dev_mkbindinf(PROTYP proto,char *ip,char *port,int iteration)

{
BININF *bininf;

bininf=(BININF *)calloc(1,sizeof(BININF));
bininf->proto=proto;
bininf->ip=strdup(ip);
bininf->port=strdup(port);
bininf->iteration=iteration;
return bininf;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set convert a protocol name*/
/*	to a PROTYP.				*/
/*						*/
/************************************************/
PROTYP dev_getproto(char *strprot)

{
#define	OPCK		"unidev.c:dev_getproto,"

static VOCTYP vocprot[]={
	{pro_smtp,SMTP,(void *)0},	/*SMTP plain link	*/
	{pro_smtps,SMTPS,(void *)0},	/*SMTP SSL link		*/
	{pro_pop3t,POP3TLS,(void *)0},	/*pop3 secured TLS link	*/
	{pro_pop3s,POP3S,(void *)0},	/*pop3 secured SSL link	*/
	{pro_pop3,POP3,(void *)0},	/*pop3 plain link	*/
	{pro_imapt,IMAPTLS,(void *)0},	/*IMAP secured TLS link	*/
	{pro_imaps,IMAPS,(void *)0},	/*IMAP secured TLS link	*/
	{pro_imap,IMAP,(void *)0},	/*IMAP plain link	*/
	{pro_ldaps,LDAPS,(void *)0},	/*LDAP secured SSL link	*/
	{pro_ldap,LDAP,(void *)0},	/*LDAP plain link	*/
	{pro_ment,MENTOR,(void *)0},	/*Mentor plain link	*/
	{0,(char *)0,(void *)0}
	};

PROTYP proto;

proto=pro_ukn;
if (strprot!=(char *)0) {	/*always	*/
  VOCTYP *voc;

  if ((voc=clm_getvoca(vocprot,strprot))==(VOCTYP *)0) 
    (void) clm_alert(0,"%s Remote server unknown protocol <%s>! (config?)",OPCK,strprot);
  else 
    proto=voc->code;
  }
return proto;
}
