/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Define low level routine to access	*/
/*	DNS record.				*/
/*						*/
/************************************************/
#ifndef UNIDNS
#define UNIDNS
#include	<netdb.h> 
#include	"basafn.h"

#define	T_SPF	99	/*SPF type in DNS	*/

/*SAFE Inc. own domain definition		*/
#define	SAFEDOM		"safe.ca"

/*defining localhos IP number			*/
#define LOCALHOST	"localhost.localdomain"
#define LOCALDOMAIN	"localdomain"

typedef	struct	{
	int preference;	/*MX preference		*/
	char *mxname;	/*MX hostname		*/
	}MXTYP;

typedef	enum	{
	spf_pass,	/*OK if condition apply	*/
	spf_fail,	/*NOK if condition apply*/
	spf_softfail,	/*Msg origin is dubious	*/
	spf_neutral,	/*same status as no SPF	*/
	spf_relax,	/*access MUST be relaxed*/
	spf_timeout,	/*no answer within time	*/
	spf_unknown	/*Unknown SPF directive	*/
	}SPFENUM;

typedef	enum	{
	dns_bestmx,	/*we are the BEST MX	*/
	dns_ourmx,	/*we are MX		*/
	dns_extmx,	/*external MX		*/
	dns_nomx	/*No MX for domain	*/
	}DNSENUM;

extern AFNTYP *dns_isipaddress(char *ipaddr);
extern char *dns_addrip(struct addrinfo *addinf);
extern char **dns_addriplist(struct addrinfo *addrlst);
extern char **dns_getspf(char *hostname);
extern char **dns_gettxt(char *domain);
extern int dns_gooddomdns(char *domain);
extern int dns_isfqdn(char *domain);
extern MXTYP **dns_getmx(char *domain);
extern MXTYP *dns_cleanonemx(MXTYP *mx);
extern MXTYP **dns_cleanmx(MXTYP **mxlist);
extern char **dns_getfirstmxiplst(char *domain);
extern char **dns_getptr(AFNTYP *afnnum);
extern char **dns_getiplist(char *hostname);
extern char *dns_getfirstip(char *hostname);
extern char *dns_getmyname();
extern int dns_matchIPrec(char *domain,AFNTYP *afnnum,int mask);
extern char *dns_getipclass(char *ipnum,int cidr);
extern DNSENUM dns_isgoodmx(char *domain,char *listenip);
extern int dns_chkversion(char *domain,char **version);
extern int dns_chksafeversion(char **version);
extern int dns_gotmxorip(char *email);
extern MXTYP **dns_getmxorip(char *domain);
#endif
