/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIEML:					*/
/*	Take Care of all access low level SMTP	*/
/*	formating.				*/
/*						*/
/************************************************/
#include	<ctype.h>
#include	<pwd.h>
#include	<stdio.h>
#include	<errno.h>
#include	<string.h>
#include	<time.h>
#include	<math.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unieml.h"

/*R-file avail command				*/
#define	SESSID	"SESSID"/*Session uniq ID	*/
#define	SUBJECT	"SBJECT"/*subject information	*/
#define	MSGID	"MSGID"	/*Email MSGID		*/
#define	PEERIP	"PEERIP"/*connexion remote IP	*/
#define	PEERNM	"PEERNM"/*connexion remote name	*/
#define	LOCIP	"LOCIP"	/*connexion local IP	*/
#define	LOCPORT	"LCPORT"/*connexion local port	*/
#define	SPMLVL	"SPMLVL"/*E-mail spam level	*/
#define	MTAILLE	"MTAILL"/*Email size		*/


int eml_orgn;		/*flag to allow ORGN	*/

/*

*/
/************************************************/
/*						*/
/*	Procedure to check a string and update	*/
/*	dsp structure with it			*/
/*						*/
/************************************************/
static void upddsp(DSPTYP *dsp,char *line)

{
#define	OPUD	"unieml.c:upddsp,"

static VOCTYP vocdis[]={
		{1,MFROM,(void *)0},
		{2,HFROM,(void *)0},
		{3,SUBJECT,(void *)0},
		{4,RCPT,(void *)0},
		{5,PEERIP,(void *)0},
		{6,PEERNM,(void *)0},
		{7,LOCIP,(void *)0},
		{8,LOCPORT,(void *)0},
		{9,SESSID,(void *)0},
		{10,SPMLVL,(void *)0},
		{11,MTAILLE,(void *)0},
		{12,MSGID,(void *)0},
		{0,(char *)0,(void *)0}
		};

VOCTYP *voc;

if ((voc=clm_getvoca(vocdis,line))!=(VOCTYP *)0) {
  RCPTYP *rmt;
 
  rmt=(RCPTYP *)0; 
  switch (voc->code) {
    case  1	:	/*grabbing Mfrom	*/
      if (dsp->efrom!=(char *)0) {
        (void) clm_alert(0,"%s multiple M-from <%s>?",OPUD,dsp->efrom);
        (void) free(dsp->efrom);
        }
      dsp->efrom=strdup(line);
      if (strlen(dsp->efrom)>0) {
	char *ptr;

	if ((ptr=strchr(dsp->efrom,' '))!=(char *)0) {
	  *ptr='\000';
	  dsp->grpin=atoi(ptr+1);
	  }
	}
      break;
    case  2	:	/*grabbing Hfrom	*/
      if (dsp->hfrom!=(char *)0) {
        (void) clm_alert(0,"%s multiple H-from <%s>?",OPUD,dsp->hfrom);
        (void) free(dsp->hfrom);
        }
      dsp->hfrom=strdup(line);
      break;
    case  3	:	/*grabbing subject	*/
      if (dsp->subject!=(char *)0) {
        (void) clm_alert(0,"%s multiple subject <%s>?",OPUD,dsp->subject);
        (void) free(dsp->subject);
        }
      dsp->subject=strdup(line);
      break;
    case  4	:	/*grabbing ONE RCPT	*/
      rmt=(RCPTYP *)calloc(1,sizeof(RCPTYP));
      rmt->rcpt=calloc(strlen(line),sizeof(char));
      rmt->why=calloc(strlen(line),sizeof(char));
      if (sscanf(line,"%s %d %[ -~]",rmt->rcpt,(int *)(&(rmt->status)),rmt->why)<2) {
        (void) clm_alert(0,"%s Unable to scan <%s>?",OPUD,line);
	rmt=eml_freercpt(rmt);
        }
      else {
        switch (rmt->status) {
          case m_stayput   :
          case m_quarantine:
            dsp->quarneeded=true;
			  /*NO BREAK	*/
          case m_acceptable:
          case m_direct    :
          case m_toresend  :
          case m_toresolv  :
          case m_toforward :
          case m_mustfwd   :
          case m_relayed   :
            break;
          case m_indata    :	/*do not count	*/
          case m_trouble   :	/*do not count	*/
          case m_virusfound:	/*do not count	*/
          case m_discarded :	/*do not count	*/
          case m_empty	   :	/*do not count	*/
          case m_noaddress :	/*do not count	*/
          case m_ipbanned  :	/*do not count	*/
          case m_rejected  :	/*do not count	*/
            break;
          default	    :	/*Big trouble!	*/
            (void) clm_alert(0,"%s unexpected status <%s> for user <%s>?",
			        OPUD,eml_strmntstatus(rmt->status),rmt->rcpt);
            break;
          }
        dsp->rcpts=(RCPTYP **)vec_addlstlst((void **)dsp->rcpts,(void *)rmt);
        }
      break;
    case  5	:	/*grabbing peerip	*/
      if (dsp->peerip!=(char *)0) {
        (void) clm_alert(0,"%s multiple peerip <%s>?",OPUD,dsp->peerip);
        (void) free(dsp->peerip);
        }
      dsp->peerip=strdup(line);
      if (strlen(dsp->peerip)>0) {
	char *ptr;

	if ((ptr=strchr(dsp->peerip,' '))!=(char *)0) {
	  *ptr='\000';
	  dsp->auth=atoi(ptr+1);
	  }
	}
      break;
    case  6	:	/*grabbing peername	*/
      if (dsp->peername!=(char *)0) {
        (void) clm_alert(0,"%s multiple peername <%s>?",OPUD,dsp->peername);
        (void) free(dsp->peername);
        }
      dsp->peername=strdup(line);
      break;
    case  7	:	/*grabbing locip	*/
      if (dsp->locip!=(char *)0) {
        (void) clm_alert(0,"%s multiple locip <%s>?",OPUD,dsp->locip);
        (void) free(dsp->locip);
        }
      dsp->locip=strdup(line);
      break;
    case  8	:	/*grabbing locport	*/
      dsp->locport=strdup(line);
      break;
    case  9	:	/*getting session ID	*/
      if (dsp->session!=(char *)0) {
        (void) clm_alert(0,"%s multiple session <%s>?",OPUD,dsp->session);
        (void) free(dsp->session);
        }
      dsp->session=strdup(line);
      break;
    case 10	:	/*getting spam level	*/
      (void) sscanf(line,"%f/%f",&(dsp->spmlvl),&(dsp->spmlim));
      break;
    case 11	:	/*getting spam level	*/
      (void) sscanf(line,"%lu",&(dsp->mtaille));
      break;
    case 12	:	/*grabbing msgid	*/
      if (dsp->msgid!=(char *)0) {
        (void) clm_alert(0,"%s multiple msgid <%s>?",OPUD,dsp->msgid);
        (void) free(dsp->msgid);
        }
      dsp->msgid=strdup(line);
      break;
    default	:
      (void) clm_alert(0,"%s Unable to understand line <%s>",OPUD,line);
      break;
    }
  }
else 
  (void) clm_alert(0,"%s Unable to vocalise line <%s>",OPUD,line);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the ASCII meaning of	*/
/*	Mentor flag value			*/
/*						*/
/************************************************/
const char *eml_strmntstatus(MENTYP menstat)

{
const char *code;

code="ERRPROC";
switch (menstat) {
  case m_indata		:
    code="Indata";
    break;
  case m_relayed	:
    code="Relayed";
    break;
  case m_toforward	:
    code="To Forward";
    break;
  case m_forwarded	:
    code="Forwarded";
    break;
  case m_mustfwd	:
    code="Must be forwarded";
    break;
  case m_toresend	:
    code="Resend";
    break;
  case m_acceptable	:
    code="Acceptable";
    break;
  case m_direct		:
    code="Direct";
    break;
  case m_stayput	:
  case m_quarantine	:
    code="Quarantine";
    break;
  case m_noaddress	:
    code="no_address";
    break;
  case m_ipbanned	:
    code="IP_banned";
    break;
  case m_rejected	:
    code="Rejected";
    break;
  case m_discarded	:
    code="Discarded";
    break;
  case m_empty	:
    code="Empty";
    break;
  case m_virusfound	:
    code="Virus_found";
    break;
  case m_trouble	:
    code="Trouble";
    break;
  case m_progress	:
    code="in progress";
    break;
  case m_delivered	:
    code="delivered";
    break;
  case m_toresolv	:
    code="Aliased from:";
    break;
  case m_delayed	:
    code="Delayed";
    break;
  case m_piped		:
    code="Piped";
    break;
  case m_undefined	:
    break;
  }
return code;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the ASCII meaning of	*/
/*	status according RFC 821 standard.	*/
/*	Return a dynamically allocated memory	*/
/*						*/
/************************************************/
char *eml_ascstat(int status,...)

{
#define	MXCMT	6	/*Max # of status line	*/

char *result;
va_list args;
const char *strcmt[MXCMT];

result=(char *)0;
va_start(args,status);
(void) memset(strcmt,0,sizeof(strcmt));
strcmt[0]="??";
switch (status) {
  case  CMDOK	:
    strcmt[0]="Requested mail action okay, completed";
    break;
  case  RELAYOK	:
    strcmt[0]="E-mail will be relayed";
    break;
  case  FWRDOK	:
    strcmt[0]="E-mail will be forwarded";
    break;
  case  SENDOK	:
    strcmt[0]="E-mail will be sent to a better MX";
    break;
  case  ALIOK	:
    strcmt[0]="Address will be aliased";
    break;
  case  QUITOK	:
    strcmt[0]="%s closing connection";
    break;
  case  IDOK	:
    strcmt[0]="OK Authenticated";
    break;
  case DLYUSR 	:
    strcmt[0]="Grey listing in progress";
    break;
  case ALREAD 	:
    strcmt[0]="You have a connection still in progress";
    strcmt[1]="please give me a break and retry later";
    break;
  case ONHOLD  	:
    strcmt[0]="Server not allowed to received mail for <%s>";
    break;
  case NEWSITE 	:
    strcmt[0]="Checking your Clearance, please retry in few minutes";
    break;
  case NOSNDR 	:
    strcmt[0]="Domain sender process not connected, please retry in few minutes";
    break;
  case NOCHKR 	:
    strcmt[0]="Domain checker not available";
    break;
  case NOANSWR 	:
    strcmt[0]="Remote server not responding anymore";
    break;
  case CHKUSR  	:
    strcmt[0]="I need to double check recipient address, please retry in few minutes";
    break;
  case DATAOK  	:
    strcmt[0]="Enter mail, end with \".\" on a line by itself";
    break;
  case TRUSTOK 	:
    strcmt[0]="Enter trusted data, end with \".\" on a line by itself";
    break;
  case TOOMANY  	:
    strcmt[0]="Too many mail pending for <%s>, please retry in few minutes";
    break;
  case  SYNERR	:
    strcmt[0]="syntax error, command unrecognized...<%s>";
    break;
  case  BADPAR	:
    strcmt[0]="syntax error in parameters or arguments (%s)";
    break;
  case  CMDBAD	:
    strcmt[0]="command not implemented (%s)";
    break;
  case  BADSEQ	:
    strcmt[0]="%s (Bad sequence of commands)";
    break;
  case  MAXDAY	:
    strcmt[0]="<%s> Sent too many email to outside (already %d today)";
    break;
  case  BADAUTH	:
    strcmt[0]="AUTH mechanism %s not available";
    break;
  case  HDRLOOP	:
    strcmt[0]="Email loop detected (%d Receive in header)";
    break;
  case  NOOP	:
    strcmt[0]="Sorry, we do not allow this operation"; 
    break;
  case  DOMNRES	:
    strcmt[0]="<%s>.... Sender domain must resolve";
    break;
  case  REJECT	:
    strcmt[0]="Trouble with mailbox, status=<%s>";
    strcmt[1]="Requested mail action not taken";
    break;
  case  NOLOC	:
    strcmt[0]="<%s> account is not existing";
    strcmt[1]="Requested mail action not taken";
    break;
  case  MALABRT	:
    strcmt[0]="Requested mail action aborted (%s)";
    break;
  case  BADID	:
    strcmt[0]="authentication failed";
    break;
  case  EXPIRED	:
    strcmt[0]="mailbox is currently locked (%s)";
    strcmt[1]="Requested mail action not taken";
    break;
  case  CMDABRT	:
    strcmt[0]="Requested mail action aborted: (%s)";
    break;
  case  FAILED	:
    strcmt[0]="transaction failed (%s)";
    break;
  case  NORELAY	:
    strcmt[0]="Relaying in behalf of user <%s> not allowed";
    strcmt[1]="Requested mail action aborted";
    break;
  case  BANNED	:
    strcmt[0]="Email from <%s> are banned";
    strcmt[1]="Requested mail action not taken";
    break;
  case  WRGRLY	:
    strcmt[0]="Relaying using originator email <%s> is not allowed (domain is unknown to us)";
    break;
  case  BLCKLST	:
    strcmt[0]="Note: %s";
    strcmt[1]="Originator server IP <%s> black listed! see above note.";
    break;
  case  ISBIG	:
    strcmt[0]="Email is found too big for recipient (%s)";
    break;
  case  MBFULL	:
    strcmt[0]="Recipient mail box is full (%s)";
    break;
  case  SPAMMER	:
    strcmt[0]="detected spammer (%s)";
    break;
  case  NOSPAMS	:
    strcmt[0]="Your E-mail has been detected as coming from a Spamming server.";
    strcmt[1]="In order to accept your first emails we need to validate your Email server";
    strcmt[2]="Please click on the bellow URL, to be activated.";
    strcmt[3]="ATTENTION validation needed!!! %s";
    break;
  case  MISSMX	:
    strcmt[0]="No valid MX found for recipient domain name (%s)";
    break;
  case  BADSPF	:
    strcmt[0]="Server with IP <%s> is NOT SPF compliant for domain used";
    strcmt[1]="If this is an error, You may need to change YOUR SPF definition record";
    strcmt[2]="within DNS for domain '%s' (see your IT support Person)";
    strcmt[3]="Email rejected according SPF definition";
    break;
  case  BADRVRS	:
    strcmt[0]="Server reverse-address for IP <%s> is undefined";
    strcmt[1]="If this is an error, You may need to change YOUR DNS definition record";
    strcmt[2]="within DNS reverse-address definition (see your IT support Person)";
    strcmt[3]="Email rejected because your server reverse-address is mandatory";
    break;
  case  NOTGOOD	:
    strcmt[0]="your email server not allowed to send email to me (%s)";
    break;
  case  NODNS	:
    strcmt[0]="No DNS found for <%s> provided as HELO data";
    break;
  case  BADMX	:
    strcmt[0]="No valid MX found for originator domain name (%s)";
    break;
  case  OPRELAY	:
    strcmt[0]="Requested mail action aborted (%s/[%s] detected as an open relay)";
    break;
  case  BADVIR	:
    strcmt[0]="virus <%s> found in the E-mail";
    strcmt[1]="Requested mail delivery aborted";
    break;
  case  MANYRJT	:
    strcmt[0]="Too many reject already, aborting connection";
    break;
  case  BULKOVR	:
    strcmt[0]="Too many user specified (%d RCPT already)";
    break;
  case  IPBAN	:
    strcmt[0]="Your IP <%s> is fully banned, mail to <%s> refused";
    strcmt[1]="Requested mail delivery aborted";
    break;
  case  UKNUSR	:
    strcmt[0]="Your ID <%s> is unknown to us, we can't relay your email";
    strcmt[1]="Requested mail relaying aborted";
    break;
  default	:
    (void) clm_alert(0,"unieml.c,eml_ascstat: status '%d' unexpected",status);
  case  ERRPROC	:
    strcmt[0]="Requested action aborted (%s)";
    break; 
  }
if (strcmt[0]!=(char *)0) {
  char *strloc;
  int i;

  strloc=(char *)0;
  for (i=0;strcmt[i]!=(const char *)0;i++) {
    char space;
    char *addon;

    space='-';
    if (strcmt[i+1]==(char *)0)
      space=' ';
    (void) asprintf(&addon,"%d%c%s",status,space,strcmt[i]);
    if (strloc!=(char *)0)
      strloc=clm_astrcat(strloc,"\r\n");
    strloc=clm_astrcat(strloc,addon);
    (void) free(addon);
    }
  (void) vasprintf(&result,strloc,args);
  (void) free(strloc);
  }
va_end(args);
return result;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove duplicate space and	*/
/*	lowercase the string contents		*/
/*						*/
/************************************************/
char *eml_standardize(char *str)

{
if (str!=(char *)0) {
  register char *ptr;

  ptr=str;
  while (*ptr!='\000') {
    if (isblank(*ptr)!=0) {
      register char *next;

      *ptr=' ';
      ptr++;
      next=ptr; 
      while (isblank(*next))
	next++;
      if (ptr!=next)
	(void) memmove(ptr,next,strlen(next)+1);
      continue;
      }
    if (isupper(*ptr)!=0) {
      *ptr=(char)tolower(*ptr);
      ptr++;
      continue;
      }
    switch (*ptr) {
      case '\\'	:/*dito				*/
        *ptr='_';
	break;
      default	:
	if (isascii(*ptr)==0) 
        *ptr='_';
	break;
      }
    ptr++;
    }
  /*we CAN'T have a space as first char	*/
  if (isblank(*str))
    (void) memmove(str,str+1,strlen(str+1)+1);
  /*we CAN'T have a space as last char	*/
  if (*str!='\000') {
    ptr=str+strlen(str)-1;
    if (isblank(*ptr))
      *ptr='\000';
    }
  }
return str;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by RCPTYP	*/
/*	definition.				*/
/*						*/
/************************************************/
RCPTYP *eml_freercpt(RCPTYP *rcpt)

{
if (rcpt!=(RCPTYP *)0) {
  rcpt->inbehalf=clm_freestr(rcpt->inbehalf);
  rcpt->note=clm_freestr(rcpt->note);
  rcpt->why=clm_freestr(rcpt->why);
  rcpt->rcpt=clm_freestr(rcpt->rcpt);
  (void) free(rcpt);
  rcpt=(RCPTYP *)0;
  }
return rcpt;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by RCPTYP	*/
/*	list					*/
/*						*/
/************************************************/
RCPTYP **eml_freercpts(RCPTYP **rcpts)

{
if (rcpts!=(RCPTYP **)0) {
  register int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    rcpts[i]=eml_freercpt(rcpts[i]);
    }
 (void) free(rcpts);
 rcpts=(RCPTYP **)0;
 }
return rcpts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to merge a recipient list to	*/
/*	another recipient list.			*/
/*						*/
/************************************************/
RCPTYP **eml_mergercpts(RCPTYP **rcpts,RCPTYP **rcp)

{
RCPTYP **neulst;

if ((neulst=eml_duprcpts(rcp))!=(RCPTYP **)0) {
  if (rcpts!=(RCPTYP **)0) {
    int t1;
    int t2;

    t1=vec_sizelstlst((void **)rcpts);
    t2=vec_sizelstlst((void **)neulst);
    rcpts=(RCPTYP **)realloc((void *)rcpts,(t1+t2+1)*sizeof(RCPTYP *));
    (void) memmove(rcpts+t1,neulst,(t2+1)*sizeof(RCPTYP *));
    (void) free(neulst);
    }
  else
    rcpts=neulst;
  }
return rcpts;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to add a recipient to		*/
/*	recipient list if NOT already within	*/
/*	current list				*/
/*	Remove reference to remote if can't be	*/
/*	added to the list.			*/
/*						*/
/************************************************/
RCPTYP **eml_addnewrcp(RCPTYP **rcpts,RCPTYP **rcp)

{
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    if (strcmp(rcpts[i]->rcpt,(*rcp)->rcpt)==0) {
      *rcp=eml_freercpt(*rcp);
      break;
      }
    }
  }
if (*rcp!=(RCPTYP *)0)
  rcpts=(RCPTYP **)vec_addlstlst((void **)rcpts,(void *)(*rcp));
return rcpts;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to duplicate on RCPT struct	*/
/*						*/
/************************************************/
RCPTYP *eml_duprcpt(const RCPTYP *rcpt)

{
RCPTYP *neu;

neu=(RCPTYP *)0;
if (rcpt!=(RCPTYP *)0) {
  neu=(RCPTYP *)calloc(1,sizeof(RCPTYP));
  neu->rcpt=strdup(rcpt->rcpt);
  if (rcpt->inbehalf!=(char *)0)
    neu->inbehalf=strdup(rcpt->inbehalf);
  neu->status=rcpt->status;
  neu->grpid=rcpt->grpid;
  neu->smtpstatus=rcpt->smtpstatus;
  if (rcpt->why!=(char *)0)
    neu->why=strdup(rcpt->why);
  }
return neu;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to duplicate an RCPT list	*/
/*						*/
/************************************************/
RCPTYP **eml_duprcpts(RCPTYP **rcpts)

{
RCPTYP **nlst;

nlst=(RCPTYP **)0;
if (rcpts!=(RCPTYP **)0) {
  register int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    RCPTYP *neu;

    neu=eml_duprcpt((const RCPTYP *)rcpts[i]);
    nlst=(RCPTYP **)vec_addlstlst((void **)nlst,(void *)neu);
    }
  }
return nlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free a CKRTYP structure	*/
/*						*/
/************************************************/
CKPTYP *eml_freeckrrcpt(CKPTYP *ckrrcpt)

{
if (ckrrcpt!=(CKPTYP *)0) {
  ckrrcpt->rcpt=clm_freestr(ckrrcpt->rcpt);
  ckrrcpt->locip=clm_freestr(ckrrcpt->locip);
  ckrrcpt->locport=clm_freestr(ckrrcpt->locport);
  ckrrcpt->efrom=clm_freestr(ckrrcpt->efrom);
  ckrrcpt->heloname=clm_freestr(ckrrcpt->heloname);
  ckrrcpt->peerip=clm_freestr(ckrrcpt->peerip);
  (void) free(ckrrcpt);
  ckrrcpt=(CKPTYP *)0;
  }
return ckrrcpt;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to scan a CHKRCPT request and*/
/*	return the build structure.		*/
/*						*/
/************************************************/
CKPTYP *eml_scanckrrqst(char *ckrrqst)

{
register CKPTYP *scanned;
register char *cpy;
register char *start;
register char *ptr;
int phase;
int proceed;

scanned=(CKPTYP *)calloc(1,sizeof(CKPTYP));
cpy=strdup(ckrrqst);
start=cpy;
ptr=(char *)0;
phase=1;
proceed=true;
while (proceed==true) {
  ptr=(char *)0;
  if ((start!=(char *)0)&&((ptr=strchr(start,' '))!=(char *)0)) {
    *ptr='\000';
    ptr++;
    }
  switch (phase) {
    case 0	:	/*fatal error		*/
      scanned=eml_freeckrrcpt(scanned);
      proceed=false;
      break;
    case 1	:	/*extracting peerip	*/
      scanned->peerip=strdup(start);
      break;
    case 2	:	/*extracting helo data	*/
      scanned->heloname=strdup(start);
      break;
    case 3	:	/*extracting from	*/
      scanned->efrom=strdup(start);
      break;
    case 4	:	/*extracting taille	*/
      (void) sscanf(start,"%lu",&(scanned->taille));
      break;
    case 5	:	/*extracting locip	*/
      scanned->locip=strdup(start);
      break;
    case 6	:	/*extracting locport	*/
      scanned->locport=strdup(start);
      break;
    case 7	:	/*extracting rcpt	*/
      scanned->rcpt=strdup(start);
      break;
    case 8	:	/*extracting auth flag	*/
      (void) sscanf(start,"%u",&(scanned->ctxflag));
      break;
    default	:	/*SAFE guard		*/
      if (start!=(char *)0)
	phase=-1;	/*not all string sanned	*/
      else
        proceed=false;
      break;
    }
  phase++;
  start=ptr;
  }
(void) free(cpy);
return scanned;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a CHKRCPT request and*/
/*	return a message structure		*/
/*						*/
/************************************************/
char *eml_mkckrrqst(register CKPTYP *scn)

{
/*RCPT check message exchange format		*/
#define CKFRM	"%s %s %s %lu %s %s %s %u"

char *msg;

(void) asprintf(&msg,CKFRM,scn->peerip,scn->heloname,scn->efrom,scn->taille,
			   scn->locip,scn->locport,scn->rcpt,scn->ctxflag);
return msg;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the domain part	*/
/*	within an E-mail address		*/
/*	if no domain part found return NULL	*/
/*						*/
/************************************************/
char *eml_getdomain(char *email)

{
char *domain;
char *ptr;

domain=email;
if ((ptr=strchr(email,'@'))!=(char *)0)
  domain=ptr+1;
return domain;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the domain part	*/
/*	within an E-mail address		*/
/*	but set cut the domain part by inserting*/
/*	a '\000' at the '@' position		*/
/*						*/
/************************************************/
char *eml_splitdomain(char *email)

{
char *domain;

if ((domain=strchr(email,'@'))!=(char *)0) {
  *domain='\000';
  domain++;
  }
else
  domain=email;
return domain;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert a dispatch struct	*/
/*	to a dispatch string			*/
/*						*/
/************************************************/
char **eml_setdispatch(DSPTYP *dsp)

{
char **dspstr;
char *dlbstr;
char *trpstr;
char *strflt;
int i;

dspstr=(char **)0;
dlbstr="%s %s";
trpstr="%s %s %d";
strflt="%s %f/%f";
for (i=0;i<11;i++) {
  char *str;

  str=(char *)0;
  switch(i) {
    case   0	:
      (void) asprintf(&str,dlbstr,SESSID,dsp->session);
      break;
    case   1	:
      (void) asprintf(&str,dlbstr,LOCPORT,dsp->locport);
      break;
    case   2	:
      (void) asprintf(&str,trpstr,PEERIP,dsp->peerip,dsp->auth);
      break;
    case   3	:
      (void) asprintf(&str,dlbstr,PEERNM,dsp->peername);
      break;
    case   4	:
      (void) asprintf(&str,dlbstr,LOCIP,dsp->locip);
      break;
    case   5	:
      (void) asprintf(&str,trpstr,MFROM,dsp->efrom,dsp->grpin);
      break;
    case   6	:
      if (dsp->hfrom!=(char *)0)
        (void) asprintf(&str,dlbstr,HFROM,dsp->hfrom);
      break;
    case   7	:
      if (dsp->subject!=(char *)0)
        (void) asprintf(&str,dlbstr,SUBJECT,dsp->subject);
      break;
    case   8	:
      (void) asprintf(&str,strflt,SPMLVL,dsp->spmlvl,dsp->spmlim);
      break;
    case   9	:
      (void) asprintf(&str,"%s %lu",MTAILLE,dsp->mtaille);
      break;
    case  10	:
      if (dsp->msgid!=(char *)0)
        (void) asprintf(&str,dlbstr,MSGID,dsp->msgid);
      break;
    }
  if (str!=(char *)0)
    dspstr=(char **)vec_addlstlst((void **)dspstr,(void *)str);
  }
if (dsp->rcpts!=(RCPTYP **)0) {
  int j;

  for (j=0;dsp->rcpts[j]!=(RCPTYP *)0;j++) {
    register RCPTYP *rcploc;
    char *str;

    rcploc=dsp->rcpts[j];
    str=(char *)0;
    (void) asprintf(&str,"%s %s %d %s",RCPT,rcploc->rcpt,rcploc->status,rcploc->why);
    dspstr=(char **)vec_addlstlst((void **)dspstr,(void *)str);
    }
  }
return dspstr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to create a dispatch structure*/
/*						*/
/************************************************/
DSPTYP *eml_getdispatch(char **dspstr)

{

DSPTYP *dsp;
int phase;
int proceed;

dsp=(DSPTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*preparing DSP memory	*/
      dsp=(DSPTYP *)calloc(1,sizeof(DSPTYP));
      dsp->spmlvl=NAN;
      dsp->spmlim=NAN;
      break;
    case  1	:	/*reading line		*/
      if (dspstr!=(char **)0) {
	int i;

	for (i=0;dspstr[i]!=(char *)0;i++) {
	  (void) upddsp(dsp,dspstr[i]);
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return dsp;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	dispatch struture.			*/
/*						*/
/************************************************/
DSPTYP *eml_cleandispatch(DSPTYP *dsp)

{
if (dsp!=(DSPTYP *)0) {
  dsp->rcpts=eml_freercpts(dsp->rcpts);
  dsp->msgid=clm_freestr(dsp->msgid);
  dsp->subject=clm_freestr(dsp->subject);
  dsp->hfrom=clm_freestr(dsp->hfrom);
  dsp->efrom=clm_freestr(dsp->efrom);
  dsp->inbehalf=clm_freestr(dsp->inbehalf);
  dsp->rejectto=clm_freestr(dsp->rejectto);
  dsp->peername=clm_freestr(dsp->peername);
  dsp->peerip=clm_freestr(dsp->peerip);
  dsp->locport=clm_freestr(dsp->locport);
  dsp->locip=clm_freestr(dsp->locip);
  dsp->session=clm_freestr(dsp->session);
  (void) free(dsp);
  dsp=(DSPTYP *)0;
  }
return dsp;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to generat an advise answer	*/
/*	from DSP information.			*/
/*						*/
/************************************************/
char *dsp_advicemsg(DSPTYP *dsp)

{
char *advise;

(void) asprintf(&advise,"%s %s",MYADV,dsp->session);
if (dsp->rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;dsp->rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *ptr;

    ptr=dsp->rcpts[i];
    if (ptr->smtpstatus==RETADV) {
      (void) clm_fasprintf(&advise,"\n%s %d %d %s",SCHK,i,ptr->status,ptr->note);
      }
    }
  }
return advise;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a file name for	*/
/*	email to be store in quarantine.	*/
/*						*/
/************************************************/
char *eml_getquarfilename(char *session,char *recipient)

{
char *quarfilename;

quarfilename=(char *)0;
if ((session!=(char *)0)&&(recipient!=(char *)0)) {
  char *user;
  char *domain;

  user=strdup(recipient);
  domain=eml_splitdomain(user);
  (void) asprintf(&quarfilename,"%s/%s/%s/%s/%s",clm_appdir(d_mails),domain,user,QUARDIR,session);
  (void) free(user);
  }
return quarfilename;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to attache a file within	*/
/*	an Email file structure.		*/
/*						*/
/************************************************/
int eml_attache(char *infile,char *outfile,char *session,
	        char *from,char *to,char *subject,char **explain)

{
#define	OPAT	"unieml.c:eml_attache,"
#define CNT	"Content-"
/*appending only 2000 charateres		*/
#define	MAXCOUNT	2000

int success;
FILE *fout;
char *marker;
int phase;
int proceed;

success=false;
fout=(FILE *)0;
marker=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*opening file		*/
      if ((fout=fopen(outfile,"w"))==(FILE *)0) {
        (void) clm_alert(0,"%s unable to open enclosing file <%s> (error=<%s>)",
			    OPAT,outfile,strerror(errno));
	proceed=false;
	}
      break;
    case  1	:	/*setting marker	*/
      (void) asprintf(&marker,"--%s",session);
      break;
    case  2	:	/*setting header	*/
      (void) fprintf(fout,"Return-path: <%s>\n","");
      if (strlen(from)==0)
	from=POSTM;
      (void) fprintf(fout,"From: %s <%s>\n","noreply",from);
      (void) fprintf(fout,"To: <%s>\n",to);
      (void) fprintf(fout,"Date: %s\n",clm_ascsysstamp(time((time_t *)0)));
      (void) fprintf(fout,"Message-Id: <%s-QUAR>\n",session);
      (void) fprintf(fout,"Subject: %s\n",subject);
      (void) fprintf(fout,"%sID: <%s>\n",XCLEM,session);
      (void) fprintf(fout,"MIME-Version: 1.0\n");
      (void) fprintf(fout,"%sType: multipart/mixed; boundary=\"%s\"\n",
		           CNT,marker);
      (void) fprintf(fout,"\nThis is a multi-part message in MIME format.\n\n");
      break;
    case  3	:	/*adding attachment comment	*/
      (void) fprintf(fout,"--%s\n",marker);
      (void) fprintf(fout,"%sType: text/plain; charset=\"iso-8859-1\"\n",CNT);
      (void) fprintf(fout,"%sTransfer-Encoding: 8bit\n\n",CNT);
      if (explain!=(char **)0) {
	register int i;

	for (i=0;explain[i]!=(char *)0;i++)
          (void) fprintf(fout,"%s\n",explain[i]);
	}
      else
        (void) fprintf(fout,"No comment\n");
      (void) fprintf(fout,"\n");
      break;
    case  4	:	/*flag MIME attachement	*/
      (void) fprintf(fout,"--%s\n",marker);
      (void) fprintf(fout,"%sType: message/rfc822;\n\tname=\"%s\"\n",
			   CNT,session);
      (void) fprintf(fout,"%sTransfer-Encoding: 8bit\n",CNT);
      (void) fprintf(fout,"%sDisposition: attachment;\n\tfilename=\"%s\"\n\n",
		          CNT,session);
      break;
    case  5	:	/*appening file		*/
      (void) clm_appenfile(fout,infile,MAXCOUNT);
      break;
    case  6	:	/*set MIME end marker	*/
      (void) fprintf(fout,"\n--%s--\n\n",marker);
      success=true;
      break;
    default	:	/*SAFE guard		*/
      (void) free(marker);
      (void) fclose(fout);
      proceed=false;
      break;
    }
  phase++;
  }
return success;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to find a MENTYP status	*/
/*	according a transmisson status		*/
/*						*/
/************************************************/
MENTYP eml_getcomstat(int status,MENTYP orgstatus)

{
#define	OPGE	"unieml.c:eml_getcomstat,"

int comstat;
int phase;
int proceed;

comstat=m_undefined;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase)  {
    case 0	:	/*specific value	*/
      proceed=false;	/*assume simple case	*/
      switch (status) {
	case CMDOK	:
	  switch (orgstatus) {
	    case m_toforward	:
	      comstat=m_forwarded;
	      break;
	    default		:
	      comstat=m_delivered;
	      break;
	    }
	  break; 
	case NOANSWR	:
	case BOUNCE	:
	case RETRY	:
	case INRELAY	:
	  comstat=m_delayed;
	  break; 
	case FAILED	:
	  comstat=m_discarded;
	  break; 
        default		:
	  comstat=m_delayed;
	  proceed=true;
	  break; 
        }
      break;
    case 1	:	/*temp trouble status	*/
      if ((status>=TMPERR)&&(status<SYNERR)) {
			/*delayed already set	*/
	phase=999;	/*comstat found		*/
	}
      break;
    case 2	:	/*big trouble status	*/
      if (status>=SYNERR) {
	comstat=m_trouble;
	phase=999;	/*comstat found		*/
	}
      break;
    case 3	:	/*very unexpected status*/
      (void) clm_alert(0,"%s Unexpected status '%d' to be analysed",OPGE,status);
      comstat=m_delayed;/*lets try again	*/
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return comstat;
}
