/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIJRL:					*/
/*	UNIt level to take care of all		*/
/*	JouRnaLling functions.			*/
/*						*/
/************************************************/
#include	<time.h>
#include	<stdio.h>
#include	<malloc.h>
#include	<string.h>
#include	<stdarg.h>
#include	<sys/file.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unijrl.h"

typedef	struct	{/*journalling information	*/
	time_t stamp;	/*press stamp		*/
	char *news;	/*press information	*/
	}PRESSTYP;

typedef struct	{
	time_t today;	/*expiration date	*/
	char *jrldir;	/*jrl printing directory*/
	char *edition;	/*jrl edition		*/
	char *jrlname;	/*jrl fullname		*/
	VPTR *pbk;	/*news vector list	*/
	}JRLTYP;

/*area to store journal information		*/
static	JRLTYP	jrl[MXJRL];
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a new journal	*/
/*	edition.				*/
/*						*/
/************************************************/
static void newedition(JRLTYP *jr)

{
#define	DFRM "%Y%m%d"
#define	JFRM "%s/%s-%s.jrl"

time_t curtime;
struct tm *ttime;
char strdate[50];

curtime=time((time_t *)0);
ttime=localtime(&curtime);
(void) strftime(strdate,sizeof(strdate),DFRM,ttime);
if (jr->jrlname!=(char *)0)
  (void) free(jr->jrlname);
if ((jr->jrldir==(char *)0)||(jr->edition==(char *)0)) {
  (void) clm_core_dump("%s jrldir <%s> or edition <%s> not set",
		       "unijrl.c:newedition,",jr->jrldir,jr->edition);
  }
(void) asprintf(&(jr->jrlname),JFRM,jr->jrldir,jr->edition,strdate);
jr->today=ttime->tm_mday;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	press book information			*/
/*						*/
/************************************************/
static PRESSTYP *freepress(PRESSTYP *bkp)

{
if (bkp!=(PRESSTYP *)0) {
  if (bkp->news!=(char *)0)
    (void) free(bkp->news);
  (void) free(bkp);
  bkp=(PRESSTYP *)0;
  }
return bkp;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to format journal contents	*/
/*						*/
/************************************************/
static char *getedition(JRLTYP *jr)

{
#define	PLSSEC	"+%02d"
#define	PLSMIN	"+%02d:%02d"
#define	PLSHOUR	"+%d:%02d:%02d"

register char *edition;

edition=(char *)0;
if (jr->pbk!=(VPTR *)0) {
  time_t start;
  VPTR *vptr;
  VECLST *vec;

  register unsigned int i;

  start=(time_t)0;
  vptr=jr->pbk;
  for (i=0,vec=vptr->vcl;i<vptr->numvec;i++,vec=vec->nxt) {
    register u_int dly;
    register PRESSTYP *news;
    char *article;
    char logtime[40];

    dly=(u_int)0;
    news=(PRESSTYP *)vec->payload;
    article=(char *)0;
    (void) strcpy(logtime,"");
    if (start==(time_t)0) {
      start=news->stamp;
      (void) strftime(logtime,sizeof(logtime),"%H:%M:%S",localtime(&start));
      }
    else {
      char strtime[40];
      register int j;

      dly=(u_int)(news->stamp-start);
      strtime[0]='\000';
      for (j=0;j<3;j++) {
        switch (j) {
          case  0	:	/*less than one minute	*/
            if (dly<60)
              (void) sprintf(strtime,PLSSEC,dly);
            break;
          case  1	:	/*less than one hour	*/
            if (dly<3600)
              (void) sprintf(strtime,PLSMIN,(dly/60)%60,dly%60);
            break;
          case  2	:	/*one hour and more	*/
            (void) sprintf(strtime,PLSHOUR,dly/3600,(dly/60)%60,dly%60);
            break;
          }
        if (strtime[0]!='\000') {
          (void) sprintf(logtime,"%8s",strtime);
          break;
          }
        }
      }
    (void) asprintf(&article,"%-10s%s\n",logtime,news->news);
    edition=clm_astrcat(edition,article); 
    (void) free(article);
    }
  }
return edition;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to flush a pressbook contents	*/
/*						*/
/************************************************/
static void flushjournal(JRLTYP *jr)

{
#define	PLSSEC	"+%02d"
#define	PLSMIN	"+%02d:%02d"
#define	PLSHOUR	"+%d:%02d:%02d"



if (jr->pbk!=(VPTR *)0) {
  time_t curtime;
  register FILE *journal;
  register struct tm *ttime;

  curtime=time((time_t *)0);
  ttime=localtime(&curtime);
  if (ttime->tm_mday!=jr->today)
    (void) newedition(jr);
  if ((journal=fopen(jr->jrlname,"a+"))!=(FILE *)0) {
    char *edition;

    if ((edition=getedition(jr))!=(char *)0) {
      (void) flock(fileno(journal),LOCK_EX);
      (void) fprintf(journal,"%s",edition);
      (void) fflush(journal);
      (void) flock(fileno(journal),LOCK_UN);
      (void) free(edition);
      }
    (void) fclose(journal);
    }
  jr->pbk=vec_freeveclst(jr->pbk,(void *(*)(void *))freepress);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to close one journal		*/
/*						*/
/************************************************/
static void closeonejrl(JRLTYP *jr)

{
if (jr->edition!=(char *)0) {
  (void) flushjournal(jr);
  (void) free(jr->jrlname);
  (void) free(jr->edition);
  (void) free(jr->jrldir);
  (void) memset(jr,'\000',sizeof(JRLTYP));
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to feed the journal contents	*/
/*						*/
/************************************************/
static void newsfeed(int flush,int numjrl,const char *frm,va_list args)

{
JRLTYP *jr;

jr=jrl+numjrl;
if ((frm!=(char *)0)&&(strlen(frm)>0)) {
  PRESSTYP *prs;

  prs=calloc(1,sizeof(PRESSTYP));
  prs->stamp=time((long *)0);
  (void) vasprintf(&(prs->news),frm,args);
  jr->pbk=vec_addveclst(jr->pbk,(void *)prs);
  }
if (flush==true) {
  (void) flushjournal(jr);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close all journal, flushing*/
/*	all backlog news.			*/
/*						*/
/************************************************/
void jrl_closejournals()

{
int i;

for (i=0;i<MXJRL;i++) {
  (void) closeonejrl(jrl+i);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the full journal	*/
/*	name.					*/
/*						*/
/************************************************/
const char *jrl_getjourname(int numjrl)

{
return (const char *)(jrl+numjrl)->jrlname;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to flush/forget all message	*/
/*	already in journal edition and close	*/
/*	the journal.				*/
/*						*/
/************************************************/
void jrl_stoppresse(int numjrl)

{
if (numjrl<MXJRL) {
  JRLTYP *jr;

  jr=jrl+numjrl;
  jr->pbk=vec_freeveclst(jr->pbk,(void *(*)(void *))freepress);
  (void) closeonejrl(jr);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to set the new journal edition*/
/*	return 1 if successfull 0 otherwise.	*/
/*						*/
/************************************************/
int jrl_setjournal(int numjrl,char *jrldir,char *edition)

{
int status;

status=0;
if (numjrl<MXJRL) {
  JRLTYP *jr;

  jr=jrl+numjrl;
  (void) closeonejrl(jr);
  jr->jrldir=strdup(jrldir);
  jr->edition=strdup(edition);
  (void) newedition(jr);
  status=1;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to store event in the logs	*/
/*	book.					*/
/*						*/
/************************************************/
void jrl_journalling(int flush,int numjrl,const char *frm,...)

{
va_list args;

va_start(args,frm);
(void) newsfeed(flush,numjrl,frm,args);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	procedure to store multilines logs	*/
/*	within journal.				*/
/*						*/
/************************************************/
void jrl_multijrl(int numjrl,char *marge,char *str)

{
register char **split;

if ((split=clm_splitstring(str,'\n'))!=(char **)0) {
  char *rpt;
  register int s;

  rpt=(char *)0;	
  (void) clm_fasprintf(&rpt,"%s",split[0]);
  for (s=1;split[s]!=(char *)0;s++)
    (void) clm_fasprintf(&rpt,"\n\t%s%s%s","  ",marge,split[s]);
  (void) jrl_journalling(false,numjrl,"%s%s",marge,rpt);
  rpt=clm_freestr(rpt);
  split=(char **)vec_freelstlst((void **)split,VECFREE);
  }
else
  (void) jrl_journalling(false,numjrl,"%s%s",marge,str);
}
/*

*/
/************************************************/
/*						*/
/*	procedure to retreive the current	*/
/*	journal last edition.			*/
/*						*/
/************************************************/
char *jrl_getedition(int numjrl)

{
register char *edition;

edition=(char *)0;
if (numjrl<MXJRL)  {
  edition=getedition(jrl+numjrl);
  }
return edition;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make journalling if current*/
/*	debug level is greater or equal debug	*/
/*	limit.					*/
/*						*/
/************************************************/
void jrl_dbgjournalling(int dbglim,int flush,int numjrl,const char *frm,...)

{
if (debug>=dbglim) {
  va_list args;

  va_start(args,frm);
  (void) newsfeed(flush,numjrl,frm,args);
  va_end(args);
  }
}
