/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNILDA:					*/
/*	Take Care of all access to LDAP server	*/
/*						*/
/************************************************/
#include	<malloc.h>
#include	<stdio.h>
#include	<ldap.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unilda.h"

#define	LDAPDN	"dn"

typedef	enum	{
	ldap_email,	/*string to get email	*/
	ldap_otheremail,/*alternate email	*/
	ldap_forwdemail,/*forwarded email	*/
	ldap_delivery,	/*delivery mode		*/
	ldap_deliverynl,/*delivery not local	*/
	ldap_uid	/*to get uid		*/
	}LDASTR;
/*

*/
/************************************************/
/*						*/
/*	Procedure to return normal lDAP		*/
/*	parameter list;				*/
/*						*/
/************************************************/
static char *normpar(LDASTR num)

{
char *str;

str="";
switch (num) {
  case ldap_email	:
    if ((str=getenv("LDAP_EMAIL"))==(char *)0)
      str="mail";
    break;
  case ldap_otheremail	:
    if ((str=getenv("LDAP_ALTERNATE_EMAIL"))==(char *)0)
      str="mailAlternateAddress";
    break;
  case ldap_forwdemail	:
    if ((str=getenv("LDAP_FORWARDING_EMAIL"))==(char *)0)
	str="mailForwardingAddress";
    break;
  case ldap_uid		:
    if ((str=getenv("LDAP_UID"))==(char *)0)
      str="uid";
    break;
  case ldap_delivery	:
    if ((str=getenv("LDAP_DELIVERY"))==(char *)0)
      str="deliveryMode";
    break;
  case ldap_deliverynl	:
    if ((str=getenv("LDAP_DELIVERY_NOLOCAL"))==(char *)0)
      str="nolocal";
    break;
  default		:
    (void) clm_alert(0,"%s Unexpected ldap keyword (num='%d', bug?)","unilda.c:normpar",num);
    break;
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract email list from	*/
/*	lda result				*/
/*						*/
/************************************************/
static char **extractldadata(struct ldap *ldp,LDAPMessage *res)

{
char **ldadata;
int prvhook;
char *dn;
LDAPMessage *msg;

ldadata=(char **)0;
prvhook=gac_inhook(false);
if ((dn=ldap_get_dn(ldp,res))!=(char *)0) {
  char *ptr;

  (void) gac_inhook(prvhook);
  (void) asprintf(&ptr,"%s=%s",LDAPDN,dn);
  ldadata=(char **)vec_addlstlst((void **)0,(void *)ptr);
  (void) gac_inhook(false);
  (void) ldap_memfree(dn);
  }
msg=ldap_first_entry(ldp,res);
while (msg!=(LDAPMessage *)0) {
  char *attr;
  BerElement *berptr;

  attr=ldap_first_attribute(ldp,msg,&berptr);
  while (attr!=(char *)0) {
    struct berval **data;

    if ((data=ldap_get_values_len(ldp,msg,attr))!=(struct berval **)0) {
      int attrsize;
      int i;

      attrsize=strlen(attr)+1;
      (void) gac_inhook(prvhook);	/* to keep trace of allocated memory	*/
      for (i=0;data[i]!=(struct berval *)0;i++) {
	char *look;

	look=calloc(attrsize+data[i]->bv_len+1,sizeof(char));
	(void) strcpy(look,attr);
	(void) strcat(look,"=");
	(void) strncat(look,data[i]->bv_val,data[i]->bv_len);
	ldadata=(char **)vec_addlstlst((void **)ldadata,(void *)look);
	}
      (void) gac_inhook(false);
      (void) ldap_value_free_len(data);
      } 
    (void) ldap_memfree(attr);
    attr=ldap_next_attribute(ldp,msg,berptr);
    }
  (void) ber_free(berptr,0);
  msg=ldap_next_entry(ldp,msg);
  }
(void) gac_inhook(prvhook);
return ldadata;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the acces LDAP_UID	*/
/*	validated by an LDAP server.		*/
/*						*/
/************************************************/
static int getldapparams(int crypted,char *hostname,char *port,char **ldapuri)

{
#define	OPGE	"unilda.c:getldapparams,"

int done;
int phase;
int proceed;

done=false;
*ldapuri=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*input checking	*/
      if (hostname==(char *)0) {
	(void) clm_alert(0,"%s missing hostname=<%s>",OPGE,hostname);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*making the ldapuri	*/
      if (hostname!=(char *)0) {/*always	*/
	char *suffix;

	suffix="";
	if (crypted==true)
  	  suffix="s";
        (void) asprintf(ldapuri,"ldap%s://%s:%s/",suffix,hostname,port);
	done=true;
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract data related to	*/
/*	uid and return it			*/
/*						*/
/************************************************/
static int getldapdata(char **data,char *mainattr,char **ldadata)

{
int found;

found=0;
if (ldadata!=(char **)0) {
  int i;

  for (i=0;ldadata[i]!=(char *)0;i++) {
    char *ptr;

    if ((ptr=strchr(ldadata[i],'='))!=(char *)0) {
      *ptr='\000';
      ptr++;
      if (strcmp(ldadata[i],mainattr)==0) {
	*data=clm_freestr(*data);
	*data=strdup(ptr);
  	found=true;
	}
      }
    }
  }
return found;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect attributs from an	*/
/*	ldap server.				*/
/*						*/
/************************************************/
static char **searchldadata(char *ldapuri,char *ldapbase,char *filter,char **attrs)

{
#define	OPSE	"unilda.c:searchldadata,"

char **ldadata;
int ldapstatus;
LDAPControl **ldapctrl;
LDAPMessage *result;
struct timeval timeout;
struct ldap *ldp;
int phase;
int proceed;

ldadata=(char **)0;
ldapstatus=0;
ldapctrl=(LDAPControl **)0;
result=(LDAPMessage *)0;
timeout.tv_usec=0;;
timeout.tv_sec=10;	/*waiting 10 second MAX	*/
ldp=(struct ldap *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*connecting to server	*/
      if ((ldapstatus=ldap_initialize(&ldp,ldapuri))<0) {
	(void) clm_alert(0,"%s unable to initialize ldap connection with server <%s>  ldap_error=<%s>",
			    OPSE,ldapuri,ldap_err2string(ldapstatus));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*closing bind to ldap	*/
      ldapstatus=ldap_search_ext_s(ldp,ldapbase,LDAP_SCOPE_ONELEVEL,filter,attrs,0,
				   ldapctrl,ldapctrl,&timeout,LDAP_NO_LIMIT,&result);
      switch (ldapstatus) {
	case LDAP_SUCCESS	:	/*everything ok	*/
	  ldadata=extractldadata(ldp,result);
	  (void) ldap_msgfree(result);
	  break;
	default			:
	  (void) clm_alert(0,"%s unable to  connection with server <%s> search=<%s>  unexpected ldap_error '%#04x'=<%s>",
				OPSE,ldapuri,ldapbase,ldapstatus,ldap_err2string(ldapstatus));
	  break;
	}
      break;
    case 2	:	/*closing bind to ldap	*/
      if ((ldapstatus=ldap_unbind_ext_s(ldp,(LDAPControl **)0,(LDAPControl **)0))<0) {
	(void) clm_alert(0,"%s unable to close ldap connection with server <%s>  ldap_error=<%s>",
			    OPSE,ldapuri,ldap_err2string(ldapstatus));
	phase=999;	/*trouble trouble	*/
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return ldadata;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the bind parameters*/
/*	to bind to user ID (and check password)	*/
/*						*/
/************************************************/
static int getbindparams(char *ldapuri,char *ldapbase,char *uid,char **binduid)

{
#define	OPBI	"unilda.c:getbindparams,"

int done;
char **ldadata;
char *attrs[2];
int phase;
int proceed;

done=false;
ldadata=(char **)0;
attrs[0]=LDAP_NO_ATTRS;
attrs[1]=(char *)0;
*binduid=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting the full UID	*/
      if (uid!=(char *)0) {	/*always	*/
	char *filter;

        (void) asprintf(&filter,"(%s=%s)",normpar(ldap_uid),uid);
        if ((ldadata=searchldadata(ldapuri,ldapbase,filter,attrs))==(char **)0) {
          (void) clm_alert(0,"%s No data found for uid=<%s>",OPBI,uid);
          phase=999;      /*trouble ?             */
          }
	(void) free(filter);
	}
      break;
    case 1      :       /*finding the full uid	*/
      if (getldapdata(binduid,LDAPDN,ldadata)!=true) {
	(void) clm_alert(0,"%s Unable to get full uid for <%s>",OPBI,uid);
	phase=999;      /*trouble ?             */
	}
      ldadata=(char **)vec_freelstlst((void **)ldadata,VECFREE);
      break;
    case 2      :       /*so we got the binduid	*/
      done=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if credential is	*/
/*	validated by an LDAP server.		*/
/*						*/
/************************************************/
int lda_goodcredential(int crypted,char *hostname,char *port,char *ldapbase,char *uid,char *pass,int *ldapstatus)

{
#define	OPGO	"unilda.c:lda_goodcredential,"

int goodpass;
char *ldapuri;
char *binduid;
LDAPControl **ldapctrl;
struct ldap *ldp;
struct timeval timeout;
long starting;
int phase;
int proceed;

goodpass=false;
ldapuri=(char *)0;
binduid=(char *)0;
ldapctrl=(LDAPControl **)0;
ldp=(struct ldap *)0;
timeout.tv_sec=10;	/*waiting 10 second MAX	*/
ldp=(struct ldap *)0;
starting=mallinfo().uordblks;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting access timout	*/
      if (ldap_set_option((struct ldap *)0,LDAP_OPT_NETWORK_TIMEOUT,&timeout)<0) {
	(void) clm_alert(0,"%s unable to initialize ldap timeout with server <%s>  ldap_error=<%s>",
			    OPGO,ldapuri,ldap_err2string(*ldapstatus));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*extracting uri	*/
      if (getldapparams(crypted,hostname,port,&ldapuri)==false) {
	(void) clm_alert(0,"%s unable to ldap parameters",OPSE);
	phase=999;	/*trouble		*/
	}
      break;
    case 2	:	/*extracting binduid*/
      if (getbindparams(ldapuri,ldapbase,uid,&binduid)==false) {
	(void) clm_alert(0,"%s unable to ldap bind parameters",OPGO);
	phase=999;	/*trouble		*/
	}
      break;
    case 3	:	/*connecting to server	*/
      if ((*ldapstatus=ldap_initialize(&ldp,ldapuri))<0) {
	(void) clm_alert(0,"%s unable to initialize ldap connection with server <%s>  ldap_error=<%s>",
			    OPGO,ldapuri,ldap_err2string(*ldapstatus));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*binding to ldap	*/
      if (pass!=(char *)0) {	/*always	*/
	struct berval cred;

	cred.bv_val=pass;
	cred.bv_len=strlen(cred.bv_val);
	*ldapstatus=ldap_sasl_bind_s(ldp,binduid,LDAP_SASL_SIMPLE,&cred,ldapctrl,ldapctrl,(struct berval **)0);
	switch (*ldapstatus) {
	  case LDAP_SUCCESS	:	/*everything ok	*/
	    goodpass=true;
	    break;
	  case LDAP_INVALID_CREDENTIALS	:
	    *ldapstatus=LDAP_SUCCESS;	/*a no trouble	*/
	    break;
	  default		:	/*report trouble*/
	    (void) clm_alert(0,"%s unable to bind ldap connection with server <%s> search=<%s>  unexpected ldap_error '%#04x'=<%s>",
				OPGO,ldapuri,binduid,*ldapstatus,ldap_err2string(*ldapstatus));
	    break;
	  }
	}
      break;
    case 5	:	/*closing bind to ldap	*/
      if (ldp!=(struct ldap *)0) {
	int status;

        if ((status=ldap_unbind_ext_s(ldp,(LDAPControl **)0,(LDAPControl **)0))<0) {
	  (void) clm_alert(0,"%s unable to close ldap connection with server <%s>  ldap_error=<%s>",
			      OPGO,ldapuri,ldap_err2string(status));
	  phase=999;	/*trouble trouble	*/
	  }
	}
      break;
    default	:	/*SAFE Guard		*/
      ldapuri=clm_freestr(ldapuri);
      binduid=clm_freestr(binduid);
      proceed=false;
      break;
    }
  phase++;
  }
if ((debug>2)&&(starting!=mallinfo().uordblks)) 
  (void) clm_alert(0,"%s Potential memory leak (about %ld bytes) (Bug?)",OPGO,labs(mallinfo().uordblks-starting));
return goodpass;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure check provided email is 	*/
/*	related to a LDAP UID.			*/
/*						*/
/************************************************/
int lda_checkemail(int crypted,char *hostname,char *port,char *ldapbase,char *uid,char **email)

{
#define	OPCH	"unilda.c:lda_checkemail,"

int found;
char **ldadata;
char *ldapuri;
int phase;
int proceed;

found=false;
ldadata=(char **)0;
ldapuri=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting attribut	*/
      if (getldapparams(crypted,hostname,port,&ldapuri)==false) {
	(void) clm_alert(0,"%s unable to ldap parameters",OPCH);
	proceed=false;	/*trouble trouble		*/
	}
      break;
    case 1	:	/*extracting uri and uid*/
      if (ldapuri!=(char *)0) {	/*alway			*/
	char *filter;
	char *attrs[3];

      	(void) asprintf(&filter,"(%s=%s)",normpar(ldap_uid),uid);
	attrs[0]=normpar(ldap_email);
	attrs[1]=normpar(ldap_otheremail);
	attrs[2]=(char *)0;
        if ((ldadata=searchldadata(ldapuri,ldapbase,filter,attrs))==(char **)0) {
	  (void) clm_alert(0,"%s No data found for uid=<%s>",OPCH,filter);
	  phase=999;	/*trouble ?		*/
	  }
	else {
          found=getldapdata(email,attrs[0],ldadata);
          ldadata=(char **)vec_freelstlst((void **)ldadata,VECFREE);
	  }
        (void) free(filter);
        (void) free(ldapuri);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return found;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure find if an email is know by an*/
/*	ldap data-base. if successful return the*/
/*	official email address.			*/
/*						*/
/************************************************/
int lda_emailknown(int crypted,char *hostname,char *port,char *ldapbase,char **email)

{
#define	OPLK	"unilda.c:lda_emailknown,"

int found;
char *ldapuri;
char **ldadata;
char *search[3];
char *attrs[2];
char *filter;
int phase;
int proceed;

found=false;
ldapuri=(char *)0;
ldadata=(char **)0;
(void) memset(search,'\000',sizeof(search));
(void) memset(attrs,'\000',sizeof(attrs));
filter=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting attribut	*/
      if (getldapparams(crypted,hostname,port,&ldapuri)==false) {
	(void) clm_alert(0,"%s unable to ldap parameters",OPCH);
	proceed=false;	/*trouble trouble		*/
	}
      break;
    case 1	:	/*extracting uri and uid*/
      search[0]=normpar(ldap_email);
      search[1]=normpar(ldap_otheremail);
      attrs[0]=search[0];
      (void) asprintf(&filter,"(|(%s=%s)(%s=%s))",search[0],*email,search[1],*email);
      if ((ldadata=searchldadata(ldapuri,ldapbase,filter,attrs))==(char **)0) {
	(void) clm_alert(0,"%s LDAP unknown email=<%s>",OPLK,*email);
	phase=999;	/*trouble ?		*/
	}
      (void) free(ldapuri);
      (void) free(filter);
      break;
    case 2	:	/*connecting to server	*/
      found=getldapdata(email,attrs[0],ldadata);
      ldadata=(char **)vec_freelstlst((void **)ldadata,VECFREE);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return found;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect all aliases	*/
/*	(distribution list) related to one	*/
/*	email.					*/
/*						*/
/************************************************/
char **lda_emailaliases(int crypted,char *hostname,char *port,char *ldapbase,char *email)

{
#define	OPAL	"unilda.c:lda_emailaliases,"

char *ldapuri;
char **lists;
char **ldadata;
char *attrs[3];
char *filter;
char *official;
int phase;
int proceed;

ldapuri=(char *)0;
lists=(char **)0;
ldadata=(char **)0;
(void) memset(attrs,'\000',sizeof(attrs));
filter=(char *)0;
official=strdup(email);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting attribut	*/
      if (getldapparams(crypted,hostname,port,&ldapuri)==false) {
	(void) clm_alert(0,"%s unable to ldap parameters",OPCH);
	phase=999;	/*trouble trouble		*/
	}
      break;
    case 1	:	/*collecting alias list		*/
      (void) memset(attrs,'\000',sizeof(attrs));
      attrs[0]=normpar(ldap_forwdemail);
      (void) asprintf(&filter,"(%s=%s)",normpar(ldap_email),email);
      if ((ldadata=searchldadata(ldapuri,ldapbase,filter,attrs))==(char **)0) {
	(void) clm_alert(0,"%s LDAP unknown email=<%s> (unlikely, bug!)",OPAL,email);
	phase=999;	/*trouble ?			*/
	}
      break;
    case 2	:	/*extracting alias list		*/
      if ((ldadata!=(char **)0)&&(ldadata[0]!=(char *)0)) {/*always*/
	int i;

	for (i=1;ldadata[i]!=(char *)0;i++) {
	  char *ptr;
	  if ((ptr=strchr(ldadata[i],'='))!=(char *)0) 
	    lists=(char **)vec_addlstlst((void **)lists,(void *)strdup(ptr+1));
	  }
        ldadata=(char **)vec_freelstlst((void **)ldadata,VECFREE);
	}
      break;
    default	:	/*SAFE Guard			*/
      filter=clm_freestr(filter);
      ldapuri=clm_freestr(ldapuri);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(official);
return lists;
}
