/************************************************/
/*						*/
/*	Procedure implementation to manage	*/
/*	SQL database.				*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<signal.h>
#include	<syslog.h>
#include	<string.h>
#include	<stdio.h>
#include	<setjmp.h>
#include	<sys/time.h>
#include	<sys/wait.h>
#include	<sys/stat.h>
/*library from clement	*/
#
#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"basafn.h"
#include	"bascnv.h"
#include	"basndb.h"
#include	"baspos.h"
#include	"basmys.h"
#include	"unisql.h"

/*ENV variable to get the Mentor DB type		*/
#define	CLEMENT_DB_TYPE	"DB_TYPE"
#define	CLEMENT_DB_HOST	"DB_HOST"
#define	CLEMENT_DB_PORT	"DB_PORT"
#define	CLEMENT_DB_NAME	"DB_NAME"

/*DB access default port				*/
#define	DBPORT	"2100"		

/*level to monitor SQL delay				*/
#define	DBGMXDUR	10	

typedef enum {
        db_postgresql,
        db_mysql,
        db_unknown
        }SQLDBTYP;

static struct	{	/*data base description		*/
	char *sqlname;	/*data-base name		*/
	SQLDBTYP sql;	/*data-base type		*/
	}sqldb[]= {
	    	{POSTGRESQL,db_postgresql},
	    	{MYSQL,db_mysql},
	    	{(char *)0,db_unknown},
	        };

typedef	struct	{
	SQLDBTYP sql;	/*database type			*/
	union	{
	  POSFILE *psql;/*postgreSQL			*/
	  MYSFILE *msql;/*MySQL				*/
	  }db;
	}SQLTYP;

typedef	struct	{
	pid_t master;	/*data-base master daemon	*/
	pid_t vaccum;	/*autovacuum master daemon	*/
	}DAEMTYP;
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the data-base type	*/
/*						*/
/************************************************/
static SQLDBTYP getsqltype()

{
char *dbtype;
SQLDBTYP sqltype;

sqltype=db_postgresql;
if ((dbtype=getenv(CLEMENT_DB_TYPE))!=(char *)0) {
  int i;

  for (i=0;sqldb[i].sqlname;i++) {
    if (strcasecmp(sqldb[i].sqlname,dbtype)==0) {
      sqltype=sqldb[i].sql;
      break;
      }
    }
  }
else {
  (void) clm_alert(1,"unisql.c,getsqltype: variable '%s' missing from env "
		     "(using default '%s' (available values are '%s','%s'))",
		      CLEMENT_DB_TYPE,POSTGRESQL,POSTGRESQL,MYSQL);
  }
return sqltype;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a tupple from the data	*/
/*	from the data-base			*/
/*						*/
/************************************************/
static SQLRESULT *gettupple(SQLFILE *sqlfile,char *command)

{
SQLRESULT *sqlresult;
SQLTYP *sql;
struct timeval start_tv;
struct timezone start_tz;
struct timeval stop_tv;
struct timezone stop_tz;
long duration;

sqlresult=(SQLRESULT *)0;
sql=(SQLTYP *)sqlfile;
if (debug>DBGMXDUR)
  (void) gettimeofday(&start_tv,&start_tz);
switch(sql->sql) {
  case (db_postgresql)	:
    sqlresult=(SQLRESULT *)pos_tupple(sql->db.psql,command);
    break;
  case (db_mysql)	:
    sqlresult=(SQLRESULT *)mys_tupple(sql->db.msql,command);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,gettupple: Unknown SQL daemon (type='%d')",
		        sql->sql);
    break;
  }
if (debug>DBGMXDUR) {
  (void) gettimeofday(&stop_tv,&stop_tz);
  duration=((stop_tv.tv_sec-start_tv.tv_sec)*1000000)+(stop_tv.tv_usec-start_tv.tv_usec);
  (void) clm_alert(4,"'%05d' sql_tupple=<%s>",getpid(),command);
  if (duration>2000000) 
    /*elapse time greater than 2 seconds...	*/
    (void) clm_alert(4,"\t\tduration='%7.5f'",(float)duration/1000000);
  }
return sqlresult;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to store/update data in the	*/
/*	data-base.				*/
/*	The command is "format ready"		*/
/*						*/
/************************************************/
static int sqldirection(SQLTYP *sql,char *command)

{
#define	OPSQ	"unisql.c:sqldirection,"

int status;
struct timeval start;
struct timezone start_tz;
struct timeval stop;
struct timezone stop_tz;
long duration;

status=-1;
if (debug>DBGMXDUR) {
  (void) gettimeofday(&start,&start_tz);
  }
switch(sql->sql) {
  case (db_postgresql)	:
    status=pos_action(sql->db.psql,command);
    break;
  case (db_mysql)	:
    status=mys_action(sql->db.msql,command);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c:sqldirection Unknown SQL daemon (type='%d')",
		        sql->sql);
    break;
  }
if (debug>DBGMXDUR) {
  (void) gettimeofday(&stop,&stop_tz);
  duration=((stop.tv_sec-start.tv_sec)*1000000)+(stop.tv_usec-start.tv_usec);
  (void) clm_alert(4,"'%05d' sql_command=<%s>",getpid(),command);
  if (duration>2000000) {	
    /*elapse time greater than 2 seconds...	*/
    (void) clm_alert(4,"\t\tduration='%7.5f'",(float)duration/1000000);
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the data-base host	*/
/*						*/
/************************************************/
const char *sql_getdbhost(const char *dbhost)

{
if (dbhost==(const char *)0) { 
  dbhost=(const char *)getenv(CLEMENT_DB_HOST);
  }
return dbhost;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the data-base dbport	*/
/*						*/
/************************************************/
const char *sql_getdbport(const char *dbport)

{
if ((dbport==(const char *)0)&&
   ((dbport=getenv(CLEMENT_DB_PORT))==(const char *)0)) {
  switch(getsqltype()) {
    case (db_postgresql):
      dbport=pos_getdfltport();
      break;
    case (db_mysql)	:
      dbport=mys_getdfltport();
      break;
    case (db_unknown)	:
    default		:
      (void) clm_alert(0,"unisql.c:sql_getdbport, Unknown SQL daemon (type='%d')",
		          getsqltype());
      break;
    }
  if (dbport==(const char *)0) {
    dbport=DBPORT;
    }
  }
return dbport;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract the data-base type	*/
/*						*/
/************************************************/
const char *sql_getdbname(const char *dbname)

{
if (dbname==(const char *)0) { 
  dbname=(const char *)getenv(CLEMENT_DB_NAME);
  if (dbname==(const char *)0)
    dbname=appname;
  }
return dbname;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to massage string to be	*/
/*	submitted to database storage		*/
/*						*/
/************************************************/
char *sql_setdbstr(char *str)

{
char *dbstr;

dbstr=(char *)0;
if (str==(char *)0)
  dbstr=strdup("NULL");
else
  (void) asprintf(&dbstr,"'%s'",str);
return dbstr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if the selected	*/
/*	data-base daemon is up.			*/
/*						*/
/************************************************/
int sql_isdaemonup(const char *dbhost,const char *dbport,const char *dbname)

{
#define TTATIV 60	/*lets try for 1 minute*/

int running;

running=false;
dbhost=sql_getdbhost(dbhost);
dbport=sql_getdbport(dbport);
dbname=sql_getdbname(dbname);
switch(getsqltype()) {
  case (db_postgresql)	:
    running=pos_isdaemonup(dbhost,dbport,dbname,TTATIV);
    break;
  case (db_mysql)	:
    running=mys_isdaemonup(dbhost,dbport,dbname,TTATIV);
    break;
  case (db_unknown)	:
  default			:
    (void) clm_alert(0,"unisql.c:sql_isdaemonup, Unknown SQL daemon (type='%d')",
		        getsqltype());
    break;
  }
return running;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make a connection to the	*/
/*	data-base daemon.			*/
/*						*/
/************************************************/
SQLFILE *sql_open(const char *dbhost,const char *dbport,const char *dbname)

{
SQLTYP *sql;

sql=(SQLTYP *)calloc(1,sizeof(SQLTYP));
sql->sql=getsqltype();
dbhost=sql_getdbhost(dbhost);
dbport=sql_getdbport(dbport);
dbname=sql_getdbname(dbname);
switch(sql->sql) {
  case (db_postgresql)	:
    if ((sql->db.psql=pos_open(dbhost,dbport,dbname))==(POSFILE *)0) {
      (void) free(sql);
      sql=(SQLTYP *)0;
      }
    break;
  case (db_mysql)	:
    if ((sql->db.msql=mys_open(dbhost,dbport,dbname))==(MYSFILE *)0) {
      (void) free(sql);
      sql=(SQLTYP *)0;
      }
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c:sql_open, Unknown SQL daemon (type='%d')",
		        sql->sql);
    (void) free(sql);
    sql=(SQLTYP *)0;
    break;
  }
if (sql!=(SQLTYP *)0) {
  (void) clm_alert(11,"unisql.c,sql_open: now connected to <%s> "
		      "data-base (port=<%s>)",dbname,dbport);
  }
return (SQLFILE *)sql;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to break a connection to the	*/
/*	data-base daemon.			*/
/*						*/
/************************************************/
SQLFILE *sql_close(SQLFILE *sqlfile)

{
SQLTYP *sql;

sql=(SQLTYP *)sqlfile;
if (sql!=(SQLTYP *)0) {
  switch(sql->sql) {
    case (db_postgresql)	:
      sql->db.psql=pos_close(sql->db.psql);
      break;
    case (db_mysql)	:
      sql->db.msql=mys_close(sql->db.msql);
      break;
    case (db_unknown)	:
    default		:
      (void) clm_alert(0,"unisql.c,sql_close: Unknown SQL daemon (type='%d')",
		         sql->sql);
      break;
    }
  (void) free(sql);
  sql=(SQLTYP *)0;
  (void) clm_alert(11,"unisql.c,sql_close: disconnected from SQL data-base");
  }
return (SQLFILE *)sql;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request a lock to access 	*/
/*	set of table.				*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int sql_getlock(SQLFILE *sqlfile,LCKTYP **lcklst)

{
SQLTYP *sql;
int status;

sql=(SQLTYP *)sqlfile;
status=-1;
switch(sql->sql) {
  case (db_postgresql)	:
    status=pos_getlock(sql->db.psql,lcklst);
    break;
  case (db_mysql)	:
    status=mys_getlock(sql->db.msql,lcklst);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_getlock: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a BEGIN to database	*/
/*	and lock action result until a commit	*/
/*	is done.				*/
/*						*/
/************************************************/
int sql_lock_action(SQLFILE *sqlfile)

{
SQLTYP *sql;
int status;

sql=(SQLTYP *)sqlfile;
status=-1;
switch(sql->sql) {
  case (db_postgresql)	:
    status=pos_lock(sql->db.psql);
    break;
  case (db_mysql)	:
    status=mys_lock(sql->db.msql);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c:sql_lock_action: Unknown SQL daemon (type='%d')",
		       sql->sql);
    (void) clm_core_dump("unisql.c:sql_lock_action, Unknown SQL daemon (type='%d')",sql->sql);
    break;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to release pending lock  	*/
/*	set on data-base tables while commiting	*/
/*	result.					*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int sql_unlock_commit(SQLFILE *sqlfile)

{
SQLTYP *sql;
int status;

sql=(SQLTYP *)sqlfile;
status=-1;
switch(sql->sql) {
  case (db_postgresql)	:
    status=pos_unlock(sql->db.psql,true);
    break;
  case (db_mysql)	:
    status=mys_unlock(sql->db.msql,true);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_unlock_commit: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to release pending lock  	*/
/*	set on data-base tables while rolling	*/
/*	back all transcation (used in case of	*/
/*	trouble).				*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int sql_unlock_rollback(SQLFILE *sqlfile)

{
SQLTYP *sql;
int status;

sql=(SQLTYP *)sqlfile;
status=-1;
switch(sql->sql) {
  case (db_postgresql)	:
    status=pos_unlock(sql->db.psql,false);
    break;
  case (db_mysql)	:
    status=mys_unlock(sql->db.msql,false);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_unlock_rollback: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to store/update data in the	*/
/*	data-base.				*/
/*	Return 0, if everything fine, -1	*/
/*	otherwise.				*/
/*						*/
/************************************************/
int sql_action(SQLFILE *sqlfile,const char *fmt,...)

{
va_list args;
char *cmd;
int status;

status=-1;
cmd=(char *)0;
va_start(args,fmt);
if ((vasprintf(&cmd,fmt,args))>0) {
  status=sqldirection((SQLTYP *)sqlfile,cmd);
  (void) free(cmd);
  }
va_end(args);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to retrieve data from the	*/
/*	data-base.				*/
/*	Return a RGresult (NULL if not 		*/
/*	successful).				*/
/*						*/
/************************************************/
SQLRESULT *sql_tupple(SQLFILE *sqlfile,const char *fmt,...)

{
SQLRESULT *rs;
va_list args;
char *cmd;

rs=(SQLRESULT *)0;
va_start(args,fmt);
if ((vasprintf(&cmd,fmt,args))>0) {
  rs=gettupple(sqlfile,cmd);
  (void) free(cmd);
  }
va_end(args);
return rs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the number of tupple*/
/*	available from a previous search in the	*/
/*	data-base.				*/
/*						*/
/************************************************/
int sql_nbrtupple(SQLFILE *sqlfile,SQLRESULT *sqlresult)

{
int nbr;
SQLTYP *sql;

nbr=0;
sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    nbr=pos_nbrtupple((POSRESULT *)sqlresult);
    break;
  case (db_mysql)	:
    nbr=mys_nbrtupple((MYSRESULT *)sqlresult);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_nbrtupple: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return nbr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to clean the result area	*/
/*	allocated in a previous sql_tupple	*/
/*						*/
/************************************************/
void sql_dropresult(SQLFILE *sqlfile,SQLRESULT *sqlresult)

{
SQLTYP *sql;

sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    (void) pos_dropresult((POSRESULT *)sqlresult);
    break;
  case (db_mysql)	:
    (void) mys_dropresult((MYSRESULT *)sqlresult);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_dropresult: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the ASCII form of field*/
/*	located in a specific row in sqlresult	*/
/*						*/
/************************************************/
char *sql_getvalue(SQLFILE *sqlfile,SQLRESULT *sqlresult,
	           int row,char *fieldname)

{
char *value;
SQLTYP *sql;

value=(char *)0;
sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    value=pos_getvalue((POSRESULT *)sqlresult,row,fieldname);
    break;
  case (db_mysql)	:
    value=mys_getvalue((MYSRESULT *)sqlresult,row,fieldname);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_getvalue: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return value;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the ASCII form of field*/
/*	located in a specific row and column in	*/
/*	sqlresult.				*/
/*						*/
/************************************************/
char *sql_getfield(SQLFILE *sqlfile,SQLRESULT *sqlresult,int row,int column)

{
char *value;
SQLTYP *sql;

value=(char *)0;
sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    value=pos_getfield((POSRESULT *)sqlresult,row,column);
    break;
  case (db_mysql)	:
    value=mys_getfield((MYSRESULT *)sqlresult,row,column);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_getfield: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return value;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the ASCII form for a	*/
/*	unix time.				*/
/*						*/
/************************************************/
const char *sql_fromunixtime(SQLFILE *sqlfile,time_t timestamp)

{
const char *asct;
SQLTYP *sql;

asct="";
sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    asct=pos_fromunixtime(timestamp);
    break;
  case (db_mysql)	:
    asct=mys_fromunixtime(timestamp);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_fromunixtime: Unknown SQL daemon (type='%d')",
		       sql->sql);
    break;
  }
return asct;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get the ASCII form for a	*/
/*	unix time.				*/
/*						*/
/************************************************/
time_t sql_tounixtime(SQLFILE *sqlfile,char *date)

{
time_t datetime;

datetime=(time_t)0;
if (date!=(char *)0) {
  SQLTYP *sql;

  sql=(SQLTYP *)sqlfile;
  switch(sql->sql) {
    case (db_postgresql)	:
      datetime=pos_tounixtime(date);
      break;
    case (db_mysql)	:
      datetime=mys_tounixtime(date);
      break;
    case (db_unknown)	:
    default		:
      (void) clm_alert(0,"unisql.c,sql_tounixtime: Unknown SQL daemon (type='%d')",
		         sql->sql);
      break;
    }
  }
return datetime;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert a date in format	*/
/*	YYYY-MM-DD to a long integer format.	*/
/*						*/
/************************************************/
u_long sql_date(SQLFILE *sqlfile,char *strdate)

{
#define	OPDA	"unisql.c:sql_date,"
u_long date;
SQLTYP *sql;

date=(u_long)0;
sql=(SQLTYP *)sqlfile;
switch (sql->sql) {
  case (db_postgresql)	:
  case (db_mysql)	:
    if ((strdate!=(char *)0)&&(strlen(strdate)>0)) {
      int annees;
      int mois;
      int jours;

      if (sscanf(strdate,"%d-%d-%d",&annees,&mois,&jours)==3)
        date=((u_long)annees*10000)+(mois*100)+jours;
      else
        (void) clm_alert(1,"%s Unable to convert date like '%s'",
		            OPDA,strdate);
      }
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_date: Unknown SQL daemon (type='%d')",
		        sql->sql);
    break;
  }
return date;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to duplicate a string		*/
/*	extracted from the data-base.		*/
/*						*/
/************************************************/
char *sql_strdup(const char *dbstr)

{
register char *str;

str=(char *)0;
if (dbstr!=(const char *)0) {
  str=strdup(dbstr);
  }
return str;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to convert a data-base string	*/
/*	to a long if possible.			*/
/*						*/
/************************************************/
long sql_atol(char *dbstr)

{
long result;

result=(long)0;
if (dbstr!=(char *)0) {
  result=atol(dbstr);
  }
return result;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to make sure a string to store*/
/*	within the data-base is clean. (no 	*/
/*	character to be misunderstood by	*/
/*	data-base).				*/
/*	String can be added directly to the	*/
/*	data-base.				*/
/*						*/
/************************************************/
char *sql_cleanstr(SQLFILE *sqlfile,char *str)

{
char *cleanstr;

cleanstr=(char *)0;
if (str!=(char *)0) {
  SQLTYP *sql;

  sql=(SQLTYP *)sqlfile;
  (void) cnv_checkencoding(str);
  switch(sql->sql) {
    case (db_postgresql)	:
      cleanstr=pos_cleanquote(str);
      break;
    case (db_mysql)	:
      cleanstr=mys_cleanquote(str);
      break;
    case (db_unknown)	:
    default		:
      (void) clm_alert(0,"unisql.c,sql_cleanstr: Unknown SQL daemon (type='%d')",
		         sql->sql);
      break;
    }
  }
else
  cleanstr=strdup("NULL");
return cleanstr;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get a current sequence	*/
/*	number					*/
/*						*/
/************************************************/
u_vlong sql_getseq(SQLFILE *sqlfile,char *seqname)

{
u_vlong seqnum;
SQLTYP *sql;

seqnum=(u_vlong)0;
sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    seqnum=pos_getseq(sql->db.psql,seqname);
    break;
  case (db_mysql)	:
    seqnum=mys_getseq(sql->db.msql,seqname);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_getseq: Unknown SQL daemon (type='%d')",
		        sql->sql);
    break;
  }
return seqnum;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to increment a sequence	*/
/*	number					*/
/*						*/
/************************************************/
u_vlong sql_incseq(SQLFILE *sqlfile,char *seqname)

{
u_vlong seqnum;
SQLTYP *sql;

seqnum=(u_vlong)0;
sql=(SQLTYP *)sqlfile;
switch(sql->sql) {
  case (db_postgresql)	:
    seqnum=pos_incseq(sql->db.psql,seqname);
    break;
  case (db_mysql)	:
    seqnum=mys_incseq(sql->db.msql,seqname);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_incseq: Unknown SQL daemon (type='%d')",
		        sql->sql);
    break;
  }
return seqnum;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to format an IP number in	*/
/*	binary format according data-base type	*/
/*	IP in binary format are needed if you	*/
/*	search for IP within range.		*/
/*						*/
/************************************************/
char *sql_toipbin(SQLFILE *sqlfile,char *ipnum)

{
char *ipbin;
SQLTYP *sql;

ipbin=(char *)0;
sql=(SQLTYP *)sqlfile;
if (strstr(ipnum,STRIPV6)==ipnum)
  ipnum+=strlen(STRIPV6);
switch(sql->sql) {
  case (db_postgresql)	:
    ipbin=pos_toipbin(ipnum);
    break;
  case (db_mysql)	:
    if (strchr(ipnum,':')!=(char *)0) {
      /*this code can be removed when net_aton	*/
      /*will be working too for IPV6 (Mysql 5.3)*/
      AFNTYP *afnnum;
      char *ipnorm;
	
      afnnum=afn_getipnum(ipnum);
      ipnorm=afn_ipnormal(afnnum);
      ipbin=mys_toipbin(ipnorm);
      ipnorm=clm_freestr(ipnorm);
      afnnum=afn_freeipnum(afnnum);
      }
    else
      ipbin=mys_toipbin(ipnum);
    break;
  case (db_unknown)	:
  default		:
    (void) clm_alert(0,"unisql.c,sql_toipbin: Unknown SQL daemon (type='%d')",
		        sql->sql);
    ipbin=strdup("");
    break;
  }
return ipbin;
}
